/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF
 *  Copyright (C) 2006 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 *
 */
#import "Defines.h"

#import "EOAssociation+PopUpComboSupport.h"

#import "EODisplayGroup.h"
#import "EODisplayGroup+Objects.h"
#import "EODisplayGroup+Selection.h"
#import "_EOInterfaceUniqueObjectsAndTags.h"


@implementation EOAssociation ( PopUpComboSupport)


/*
 *  SOME SUPPORT CODE. MAINLY FOR EOPOPUPASSOCIATION AND
 *  EOCOMBOBOXASSOCIATION
 */
//
// usually this will check a dispayGroup for some "name" aspect of the selected 
// objects. But we also support, that the displayGroup has "one selected" object
// and the keypath, goes through a toMany relationship and gets its values from
// there...
//
typedef struct
{
   unsigned int   count;
   BOOL           isUnknown;
   BOOL           isToMany;
   NSArray        *components;
   NSArray        *leadingComponents;
   NSString       *tailAttributeKey;
   
   EODisplayGroup *group;
   NSString       *key;
} EOAssociationKeyPathInfo;


- (void) setKeyPathInfo:(EOAssociationKeyPathInfo *) info
              forAspect:(NSString *) aspect
{
   unsigned int         i, n;
   NSString             *name;
   EOClassDescription   *desc;
   
   info->group  = [self displayGroupForAspect:aspect];
   info->key    = [self displayGroupKeyForAspect:aspect];
   
   info->components = [info->key componentsSeparatedByString:@"."];
   info->count      = [info->components count];
   info->isToMany   = NO;
   info->isUnknown  = NO;
   
   info->tailAttributeKey  = nil;
   info->leadingComponents = (id) [NSMutableArray array];
   
   // it would be better if we could ask "EnterpriseObjects" themselves
   // or may their class
#warning (nat) this is the only place EOInterface currently relies on classDescription   
   desc = [[info->group dataSource] classDescriptionForObjects];
   n    = info->count;
   for( i = 0; i < n; i++)
   {
      name = [info->components objectAtIndex:i];
      if( [[desc attributeKeys] containsObject:name])
      {
         info->tailAttributeKey = name;
         NSCParameterAssert( i == n - 1);
         return;
      }
      
      [(id) info->leadingComponents addObject:name];
      if( [[desc toManyRelationshipKeys] containsObject:name])
      {
         desc = [desc classDescriptionForDestinationKey:name];
         info->isToMany = YES;
      }
      else
         if( [[desc toOneRelationshipKeys] containsObject:name])
            desc = [desc classDescriptionForDestinationKey:name];
         else
            desc = nil;
      
      //
      // hmm, going through a non-class description function here
      //
      if( ! desc)
      {
         // fix it up to some default, if this doesn't work
         // use EOClassDescription, like you should...
         info->isUnknown = YES;
         info->leadingComponents = [info->components subarrayWithRange:NSMakeRange( 0, n - 1)];
         info->tailAttributeKey  = [info->components lastObject];
         info->isToMany          = info->isToMany ?: [name hasSuffix:@"s"];
         return;
      }
   }
}


//
// well even if the result could point to a duplicate
// th
- (NSString *) uniquedStringForObject:(id) obj
                               aspect:(NSString *) aspect
{
   NSString   *key;
   NSString   *s;
   id         value;
   
   key = [self displayGroupKeyForAspect:aspect];
   if( ! key)
   {
#if DEBUG      
      NSLog( @"%@: aspect %@ is not connected as expected", self, aspect);
#endif      
      return( nil);
   }
   
   value = [obj valueForKeyPath:key];
   s     = [value description];
   
   return( s);
}


/*
 * this is useful for popup boxes, where you want the titles from
 * a displayGroup or a relation, but you usually also want the
 * objects that lead up to the titles. So you collect the objects
 * here and later apply the final path component (if it is
 * an attribute)                                             
 */
- (NSArray *) collectObjectsSmartlyForAspect:(NSString *) aspect
                                 keyPathInfo:(EOAssociationKeyPathInfo *) info
{
   id                         obj;
   EOAssociationKeyPathInfo   dummy;
   NSString                   *path;
   NSArray                    *array;
   
   if( ! info)
      info = &dummy;
   
   [self setKeyPathInfo:info
              forAspect:aspect];
   
   if( ! info->count)
      return( [info->group displayedObjects]);

//   if( info->isUnknown)
//      return( [[info->group displayedObjects] valueForKeyPath:info->key]);
   
   if( info->count == 1)
   {
      if( info->tailAttributeKey)
         return( [info->group displayedObjects]);
      if( ! info->isToMany)
         return( [[info->group displayedObjects] valueForKeyPath:info->key]);
      
      //
      // hmm something like "bars": toMany -> objects 
      // use the selectedObjected for this 
      //
   }
   
   if( ! info->isToMany)
   {
      if( info->tailAttributeKey)
         return( [info->group displayedObjects]);

      return( [[info->group displayedObjects] valueForKeyPath:info->key]);
   }
   
   array = [info->group selectedObjects];
   if( [array count] != 1)
      return( [NSArray array]);

   obj = [array lastObject];
   if( info->tailAttributeKey)
   {
      path = [info->leadingComponents componentsJoinedByString:@"."];
      array = [obj valueForKeyPath:path];
   }
   else
      array = [obj valueForKeyPath:info->key];
   
   // mung hack hack hack
   if( ! [array isKindOfClass:[NSArray class]])
   {
#if DEBUG
      NSLog( @"You should have better class descriptions for %@ from %@", info->key, info->group);
#endif
      array = [NSArray arrayWithObject:array];
   }
   return( array);
}


- (NSArray *) originatingObjectsOfUniquedStringsForAspect:(NSString *) aspect
{
   NSArray   *array;
   
   array = [self collectObjectsSmartlyForAspect:aspect
                                    keyPathInfo:NULL];
   return( array);
   
}


- (_EOInterfaceUniqueObjectsAndTags *) uniquedStringsForAspect:(NSString *) aspect
{
   _EOInterfaceUniqueObjectsAndTags   *objectsAndTags;
   EOAssociationKeyPathInfo           info;
   NSArray                            *array;
   id                                 object;
   unsigned int                       i, n;

   array = [self collectObjectsSmartlyForAspect:aspect
                                    keyPathInfo:&info];
   
   //
   // remove duplicates, keep index in tag
   //
   n = [array count];
   
   objectsAndTags = [[[_EOInterfaceUniqueObjectsAndTags alloc] initWithCapacity:n] autorelease];

   //
   // ugh we have to sort arrays, but keep the indexes 
   // correct, that's hard. We do this with
   //
   for( i = 0; i < n; i++)
   {
      // valueForKey on NSArray makes nil -> @"<null>", which is bad
      object = [array objectAtIndex:i];
      if( info.tailAttributeKey)
         object = [object valueForKeyPath:info.key];
      
      if( ! object)
         continue;
      [objectsAndTags addObject:object
                        withTag:i];  // index of source array
   }
   
   //
   // this is usally an error, not a feature...
   //
   if( n -= [objectsAndTags count])
   {
      NSMutableSet   *set;
      
      set = [NSMutableSet setWithArray:array];
      [set minusSet:[NSSet setWithArray:[objectsAndTags allObjects]]];
      
      NSLog( @"%@ : %d duplicate strings (%@) have been filtered from aspect %@ (%@)", self, n, [set allObjects], aspect, info.key);
   }
   
   return( objectsAndTags);
}

@end
