/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF
 *  Copyright (C) 2006 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 *
 */
#ifndef _EOASSOCIATION_H
#define _EOASSOCIATION_H


#import "MulleEOInterfaceParentIncludes.h"


@class EODisplayGroup;

typedef enum
{
   EOAssociationObjectStateKeepState = -1,
   EOAssociationObjectStateDisabled  = 0,
   EOAssociationObjectStateEnabled   = 1
} EOAssociationObjectState;


typedef struct
{
   NSMapTable            *_displayGroupBindings;
   NSMutableDictionary   *_keyBindings;
} eo_association_extra;


@interface EOAssociation : EODelayedObserver <NSCoding>
{
    id   _object;

    unsigned   _refs:7;       // unused
    unsigned   _copyObjects:1;
    unsigned   _isConnected:1;
    unsigned   _hasBeenConnectedOnce:1;
    unsigned   _wasEnabled:1;
    unsigned   _editedYet:1;  // used by some associations, but not al
    unsigned   _extras:4;     
    unsigned   subclassFlags:16;
}

+ (NSString *) displayName;
+ (BOOL) isUsableWithObject:(id)object;
+ (NSArray *) aspects;
+ (NSArray *) aspectSignatures;
+ (NSString *) primaryAspect;
+ (NSArray *) objectKeysTaken;
+ (NSArray *) associationClassesSuperseded;
- (id) initWithObject:(id) object;
- (id) initWithCoder:(NSCoder *)coder;
- (void) encodeWithCoder:(NSCoder *)coder;
- (void) establishConnection;
- (void) dealloc;
- (void) breakConnection;
- (id) object;
- (void) bindAspect:(NSString *) name
       displayGroup:(EODisplayGroup *) dest
                key:(NSString *) key;
- (EODisplayGroup *) displayGroupForAspect:(NSString *) name;
- (NSString *) displayGroupKeyForAspect:(NSString *) name;
- (BOOL) canBindAspect:(NSString *) name
          displayGroup:(EODisplayGroup *) displayGroup
                   key:(NSString *) key;
- (BOOL) endEditing;
- (void) subjectChanged;
+ (NSArray *) associationClassesForObject:(id) object;
- (BOOL) isConnected;

// addition:
//+ (id) associationWithObject:(id) object;  // convenience constructor
- (void) _breakWrongConnectionToObject:(id) other;
- (NSArray *) _allAspects;
- (BOOL) shouldRetainObjectWhileBeingEstablished;

+ (size_t) associationExtraSize;
- (void *) associationExtraStorage;
- (void) _readWithCoder:(NSCoder *) decoder;

- (void) unbindAspect:(NSString *) aspectName;

@end


@interface EOAssociation ( EOExtras)

- (id) valueForAspect:(NSString *) name;
- (BOOL) setValue:(id) value
        forAspect:(NSString *) name;
- (id) valueForAspect:(NSString *) name
              atIndex:(unsigned) index;
- (BOOL) setValue:(id) value
        forAspect:(NSString *) name
          atIndex:(unsigned) index;
- (BOOL) shouldEndEditingForAspect:(NSString *) aspectName
                      invalidInput:(NSString *) string
                  errorDescription:(NSString *) errorDescription;
- (BOOL) shouldEndEditingForAspect:(NSString *) aspectName
                      invalidInput:(NSString *) string
                  errorDescription:(NSString *) errorDescription
                             index:(unsigned) index;
- (void) copyMatchingBindingsFromAssociation:(EOAssociation *) other;

// addition!
- (BOOL) boolForValue:(id) value;
- (BOOL) boolForAspect:(NSString *) aspect;  

// bool for Enabled is totally different
- (EOAssociationObjectState) objectStateForEnabledAspect;
- (EOAssociationObjectState) objectStateForAspect:(NSString *) aspect; // quite different logic!!
- (EOAssociationObjectState) objectStateForAspect:(NSString *) aspect
                             editingDisplayGroup:(EODisplayGroup *) editingGroup;
- (NSException *) validateValue:(id *) value
                      forAspect:(NSString *) name;
- (id) nibInstantiate;
- (id) mulleSelectedValueObject;

@end

#endif
