/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.2  2002/04/22 21:56:28  nat
 * EODisplayGroup mit code zugemuellt
 *
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOAssociation.h"
#import "AspectsAndKeys.h"

#import "EODisplayGroup.h"
#import "EODisplayGroup+Associations.h"
#import "EODisplayGroup+InsertUpdateDelete.h"
#import "EODisplayGroup+Objects.h"
#import "EODisplayGroup+Selection.h"
#ifdef GNUSTEP
#else
# import <objc/objc-runtime.h>
#endif

#define LATE_OBJECT_RELEASE  0  // a stoopid hack for testing

#define ENCODING_VERSION   229

/*
2002-05-09 00:16:02.843 listassociationnames[1024] EOGenericControlAssociation EOGenericControlAssociation
2002-05-09 00:16:02.853 listassociationnames[1024] EOControlAssociation EOControlAssoc
2002-05-09 00:16:02.863 listassociationnames[1024] EOActionAssociation EOActionAssociation
2002-05-09 00:16:02.872 listassociationnames[1024] EOComboBoxAssociation EOComboAssoc
2002-05-09 00:16:02.881 listassociationnames[1024] EOMasterPeerAssociation EOMasterPeerAssoc
2002-05-09 00:16:02.891 listassociationnames[1024] EOTableViewAssociation EOTableViewAssociation
2002-05-09 00:16:02.900 listassociationnames[1024] EOActionInsertionAssociation EOActionInsertion
2002-05-09 00:16:02.909 listassociationnames[1024] EOActionCellAssociation EOCellAssoc
2002-05-09 00:16:02.918 listassociationnames[1024] EOPopUpAssociation EOPopupAssoc
2002-05-09 00:16:02.927 listassociationnames[1024] EOColumnAssociation EOColumnAssoc
2002-05-09 00:16:02.936 listassociationnames[1024] EOTextAssociation EOTextAssociation
2002-05-09 00:16:02.945 listassociationnames[1024] EOAssociation EOAssociation
2002-05-09 00:16:02.954 listassociationnames[1024] EORecursiveBrowserAssociation EORecBrowser
2002-05-09 00:16:02.955 listassociationnames[1024] EODetailSelectionAssociation EODetailSelection
2002-05-09 00:16:02.955 listassociationnames[1024] EOMasterDetailAssociation EOMasterDetailAssoc
2002-05-09 00:16:02.955 listassociationnames[1024] EOQuickTimeAssociation EOQuickTimeAssoc
2002-05-09 00:16:02.955 listassociationnames[1024] EOMasterCopyAssociation EOMasterCopyAssoc
2002-05-09 00:16:02.956 listassociationnames[1024] EORadioMatrixAssociation EORadioAssoc
2002-05-09 00:16:02.956 listassociationnames[1024] EOMatrixAssociation EOMatrixAssoc
2002-05-09 00:16:02.956 listassociationnames[1024] EOPickTextAssociation EOPickTextAssociation
*/

@implementation EOAssociation

+ (void) initialize
{
   [EOAssociation setVersion:ENCODING_VERSION];
}

static inline BOOL   isKindOfClass( struct objc_class *child, struct objc_class *parent)
{
   do
      if( child == parent)
         return( YES);
   while( child = child->super_class);
   return( NO);
}


/* apparently not needed in GnuStep as this function
 * does nothing now. IB Support only apparently.
 */
static NSMutableArray *loadedSubclassesOf( Class masterClass)
{
   Class              aClass;
   NSMutableArray     *array;
   int                numClasses = 0;
   int                newNumClasses;
#ifndef GNUSTEP
   Class              *classes = NULL;
#endif
   int                i;

   //
   // check that masterClass instances can be added to NSArray
   //
   aClass = masterClass;
   while( aClass != [NSObject class])
      if( ! (aClass = [aClass superclass]))
         return( nil);
   array = [NSMutableArray array];

#ifndef GNUSTEP
   newNumClasses = objc_getClassList( NULL, 0);
   do
   {
      numClasses = newNumClasses;
      classes = realloc( classes, sizeof(Class) * numClasses);
      newNumClasses = objc_getClassList(classes, numClasses);
   }
   while( numClasses < newNumClasses);  // hein ??

   for( i = 0; i < numClasses; i++)
   {
      aClass = classes[ i];
      if( isKindOfClass( aClass, masterClass))
      {
         [array addObject:aClass];
         continue;
      }
   }
   free( classes);
#endif   
   return( array);
}


static inline Class   _isa( NSObject *obj)
{
   return( ((struct { @defs( NSObject) } *) obj)->isa);
}


#if DEBUG
unsigned int   NumberOfAssociationsAlive;
NSHashTable    *aliveAssociations;


static void    addAliveAssociation( EOAssociation *assoc)
{
   if( ! aliveAssociations)
      aliveAssociations = NSCreateHashTable( NSNonRetainedObjectHashCallBacks, 10);
   NSHashInsertKnownAbsent( aliveAssociations, assoc);
}


static void    removeAliveAssociation( EOAssociation *assoc)
{
   if( ! aliveAssociations)
      return;
   
   NSHashRemove( aliveAssociations, assoc);
   if( ! NSCountHashTable( aliveAssociations))
   {
      NSFreeHashTable( aliveAssociations);
      aliveAssociations = NULL;
   }
}


//
// convenient sometimes...
//
void    clearAliveAssociationsMemory()
{
   if( aliveAssociations)
   {
      NSFreeHashTable( aliveAssociations);
      aliveAssociations = NULL;
   }
}


static void   dumpAliveAssociations()
{
   NSHashEnumerator   rover;
   EOAssociation      *assoc;
   
   if( ! aliveAssociations)
      return;
   
   rover = NSEnumerateHashTable( aliveAssociations);
   while( assoc = NSNextHashEnumeratorItem( &rover))
      printf( "%s %p +%u\n", _isa( assoc)->name, assoc, [assoc retainCount]);
   NSEndHashTableEnumeration( &rover);
}

#endif

+ (size_t) associationExtraSize
{
   return( sizeof( eo_association_extra));
}


static inline eo_association_extra   *associationExtra( EOAssociation *self)
{
   return( (eo_association_extra *) &((char *) self)[ ((struct objc_class *) self->isa)->instance_size]);
}


- (void *) associationExtraStorage
{
   return( associationExtra( self)); 
}


static inline NSMutableDictionary   *keyBindings( EOAssociation *self)
{
   return( associationExtra( self)->_keyBindings);
}


static inline void  setKeyBindings( EOAssociation *self, NSMutableDictionary *dict)
{
   associationExtra( self)->_keyBindings = dict;
}


static inline NSMapTable   *displayGroupBindings( EOAssociation *self)
{
   return( associationExtra( self)->_displayGroupBindings);
}


static inline void  setDisplayGroupBindings( EOAssociation *self, NSMapTable *table)
{
   associationExtra( self)->_displayGroupBindings = table;
}


+ (id) alloc
{
   id   p;
   
   p = NSAllocateObject( self, [self associationExtraSize], NULL);;
#if DEBUG   
   NumberOfAssociationsAlive++;
   addAliveAssociation( p);
#endif   
   return( p);
}


+ (id) allocWithZone:(NSZone *) zone
{
   id   p;

   p = NSAllocateObject( self, [self associationExtraSize], zone);;
#if DEBUG   
   NumberOfAssociationsAlive++;
   addAliveAssociation( p);
#endif   
   return( p);
}


- (id) initWithObject:(id) object
{
   [super init];

#if DEBUG_CODER
   NSLog( @"%@ %s %p",  isa, __PRETTY_FUNCTION__, self);
#endif

   if( ! object)
   {
#if DEBUG_HARDCORE
      abort();
#endif   
      [self release];
      return( nil);
   }

   //
   // laymers do the darndest things. This is for older NIBs
   //
   if( ! [isa isUsableWithObject:object])
   {
      NSLog( @"Association %@ does not support objects of type %@, but it will be setup this way", 
                isa, [object class]);
#if DEBUG_HARDCORE
         abort();
#endif   
   }
   
   /*
    * this is bullshit, but currently apparently neccessary or there is 
    * something I don't understand. initWithObject: is called by initWithCoder:
    * at the time we are called and we are EOColumnAssociation then object
    * will be a NSTableColumn. The NSTableView apparently has not been
    * deserialized yet, so the NSTableColumn has a retainCount of 1
    * autoreleaseing won't protect it long enough to have the NSTableView
    * deserialized and retaining the NSTableColumn. The effect will be a 
    * released NSTableColumn and a stale _objec pointer. So what we do is
    * keep it around, until establishConnection and then release it once.
    */
   _object = object; // [object retain];  // temporary! retain
#if LATE_OBJECT_RELEASE   
   [object retain];
#endif   
   setDisplayGroupBindings( self, NSCreateMapTable( NSObjectMapKeyCallBacks, NSNonRetainedObjectMapValueCallBacks, 4));
   setKeyBindings( self, [NSMutableDictionary new]);

   return( self);
}


- (void) dealloc
{
#if DEBUG_VERBOSE
   fprintf( stderr, "%s %p\n", __PRETTY_FUNCTION__, self);
#endif
   NSAssert( ! _isConnected, @"Why is it connected, yet reaped ?. There should be a pending release via breakConnection");
#if DEBUG
#ifdef MULLE_EO_CONTROL
   NSAssert( [[self observerQueue] _priorityOfObserver:self] == (unsigned int) -1, @"Hey we broke the connection, yet we are still queued up ?");
   NSAssert1( [[EOObserverCenter _objectsObservedByObserver:self] count] == 0, @"Why are we still observing objects %@", [EOObserverCenter _objectsObservedByObserver:self]);
#endif   
#endif   
#if LATE_OBJECT_RELEASE   
   if( ! _hasBeenConnectedOnce)
      [_object release];
#endif   
   if( displayGroupBindings( self))
      NSFreeMapTable( displayGroupBindings( self));
   [keyBindings( self) release];

#if DEBUG
   removeAliveAssociation( self);
   NumberOfAssociationsAlive--;
#endif   
   [super dealloc];
}


- (NSArray *) _allAspects
{
   NSMutableArray    *array;
   NSMapEnumerator   rover;
   NSString          *aspect;
   id                unused;
   
   array = [NSMutableArray array];
   rover = NSEnumerateMapTable( displayGroupBindings( self));
   while( NSNextMapEnumeratorPair( &rover, (void *) &aspect, (void *) &unused))
      [array addObject:aspect];  // can keyBindings( self) contain more aspects ? 
   NSEndMapTableEnumeration( &rover);
   return( array);
}


- (NSArray *) _allDisplayGroups
{
   NSMutableSet      *set;
   NSMapEnumerator   rover;
   EODisplayGroup    *group;
   id                unused;
   
   set   = [[NSMutableSet new] autorelease];  // GNUSTEP doesn't know +set ?
   rover = NSEnumerateMapTable( displayGroupBindings( self));
   while( NSNextMapEnumeratorPair( &rover, (void *) &unused, (void *) &group))
      [set addObject:group];  // can keyBindings( self) contain more aspects ? 
   NSEndMapTableEnumeration( &rover);
   return( [set allObjects]);
}


- (void) _readWithCoder:(NSCoder *) decoder
{
   NSString         *key;
   int              i, n;
   id               object;
   NSString         *aspect;
   EODisplayGroup   *group;

   [decoder decodeValuesOfObjCTypes:"@i", &object, &n];
#if DEBUG_CODER   
   NSLog( @"Expect %d aspects", n);
#endif
   
   if( ! object)
   {
      NSLog( @"%s %@ %p: has no object assigned to it, NIB corrupt.", 
             __PRETTY_FUNCTION__, isa, self);
#if DEBUG      
      abort();
#endif      
      //  [self autorelease];  // isn't that a trifle surprising ???
      return;
   }
   
   [self initWithObject:object];
   [object autorelease];  // we need to autorelase it because it's not ours
   
   for( i = 0; i < n; i++)
   {
      [decoder decodeValuesOfObjCTypes:"@@@", &aspect, &group, &key];
#if DEBUG_CODER // | 1  
      NSLog( @"<%@ %p>: read aspect \"%@\" DisplayGroup:%@ key:%@", isa, self, aspect, group, key);
#endif    
      
      // it's not very good to "assert" in initWithCoder. rather let the
      // NIB load and warn in the console...
      if( group)
      {
#if DEBUG         
         if( ! [group isKindOfClass:[EODisplayGroup class]])
         {
            NSLog( @"%@: expected a EODisplayGroup but read a %@ classed object for aspect \"%@\" and key \"%@\"", 
                   isa, [group class], aspect, key);
         }
         else
            NSMapInsert( displayGroupBindings( self), aspect, group);
#else         
         NSMapInsert( displayGroupBindings( self), aspect, group);
#endif         
         
         if( key)
            [keyBindings( self) setObject:key
                                   forKey:aspect];
         [group autorelease];
      }
      [aspect release];
      [key release];
   }
#if DEBUG_CODER   
   NSLog( @"Leave %@ %s %p (g:%@ k:%@)",
          isa, __PRETTY_FUNCTION__, self, 
          NSStringFromMapTable( displayGroupBindings( self)), 
          keyBindings( self));
#endif      
}


- (id) initWithCoder:(NSCoder *) decoder
{
   int   version;

   version = [decoder versionForClassName:@"EOAssociation"];
#if DEBUG_CODER   
   NSLog( @"Enter %@ %s %p",  isa, __PRETTY_FUNCTION__, self);
#endif

   [self _readWithCoder:decoder];
   
   return( self);
}


- (void) encodeWithCoder:(NSCoder *) encoder
{
   NSArray         *aspects;
   EODisplayGroup  *group;
   NSString        *aspect;
   NSString        *key;
   int             i, n;
   id              obj;
   
#if DEBUG_CODER   
   NSLog( @"Enter %@ %s 0x%lx",  isa, __PRETTY_FUNCTION__, (long) self);
#endif
   aspects = [self _allAspects];
   n       = [aspects count];

   obj = [self object];  // for EOColumnAssociation
   if( ! obj)
   {
      NSLog( @"%s %@ %p: has no object assigned to it, NIB corrupt. Committing suicide", 
             __PRETTY_FUNCTION__, isa, self);
#if DEBUG      
      abort();
#endif
      [NSException raise:NSInternalInconsistencyException
                  format:@"%@ %p has no object assigned to it", isa, self];
   }      
   [encoder encodeValuesOfObjCTypes:"@i", &obj, &n];

#if DEBUG_CODER   
   NSLog( @"Write %d aspects", n);
#endif
   for( i = 0; i < n; i++)
   {
      aspect = [aspects objectAtIndex:i];
      key    = [keyBindings( self) objectForKey:aspect];
      group  = NSMapGet( displayGroupBindings( self), aspect);

#if DEBUG_CODER   
      NSLog( @"Write aspect:%@ DisplayGroup:%@ key:%@", aspect, group, key);
#endif      
      [encoder encodeValuesOfObjCTypes:"@@@", &aspect, &group, &key];
   }
#if DEBUG_CODER   
   NSLog( @"Leave %@ %s 0x%lx", isa, __PRETTY_FUNCTION__, (long) self);
#endif   
}


#if DEBUG

- (id) retain
{
   return( [super retain]);
}


- (void) release
{
   [super release];
}


- (id) autorelease
{
   return( [super autorelease]);
}

#endif


+ (NSString *) displayName
{
   return( NSStringFromClass( self));
}


+ (BOOL) isUsableWithObject:(id)object
{
   return( NO);
}


+ (NSArray *) aspects
{
   return( [NSArray array]);
}


+ (NSArray *) aspectSignatures
{
   unsigned int     n;
   NSMutableArray   *array;
   unsigned int     i;
   
   n = [[self aspects] count];
   array = [NSMutableArray arrayWithCapacity:n];
   
   for( i = 0; i < n; i++)
      [array addObject:@"A1M"];
   return( array);
}


+ (NSString *) primaryAspect
{
   return( nil);
}


+ (NSArray *) objectKeysTaken
{
   return( [NSArray array]);
}


+ (NSArray *) associationClassesSuperseded
{
   return( nil);
}


- (id) nibInstantiate
{
   return( self);
}


- (BOOL) shouldRetainObjectWhileBeingEstablished
{
   return( YES);
}


- (void) establishConnection
{
   EODisplayGroup    *group;
   NSMapEnumerator   rover;
   id                unused;
   
#if DEBUG_VERBOSE
   fprintf( stderr, "%s %p (_isConnected=%s)\n", __PRETTY_FUNCTION__, self, _isConnected ? "YES" : "NO");
#endif
   if( _isConnected)
   {
      NSLog( @"%s: connection already established for %@", __PRETTY_FUNCTION__, self);
#if DEBUG
      abort();
#endif      
      return;
   }

   /* see initWithObject: */
#if LATE_OBJECT_RELEASE
   if( ! _hasBeenConnectedOnce)
      [_object autorelease];
#endif
   
   rover = NSEnumerateMapTable( displayGroupBindings( self));
   while( NSNextMapEnumeratorPair( &rover, (void *) &unused, (void *) &group))
   {
#if DEBUG_VERBOSE
      NSLog( @"%@ observing %@ for changes", self, group);
#endif      
      [EOObserverCenter addObserver:self
                          forObject:group];
      [self objectWillChange:group];  // force initial sync with group 
   }
   NSEndMapTableEnumeration( &rover);
      
   _isConnected         = YES;
   _hasBeenConnectedOnce = YES;
   [self retain];
   if( [self shouldRetainObjectWhileBeingEstablished])
      [_object retain];
}


- (void) _breakWrongConnectionToObject:(id) other
{
   if( [other isKindOfClass:[EOAssociation class]])
   {
      NSLog( @"*** NIB problem: Ghost assocation %@ shares our object %@, breaking that connection...", other, [self object]);
      [other breakConnection];
   }
}


//
// super classes should call [super breakConnection] LAST in their
// breakConnection methods
//
- (void) breakConnection
{
   NSMapEnumerator   rover;
   EODisplayGroup    *group;
   id                unused;
   
#if DEBUG_VERBOSE
   fprintf( stderr, "%s %p (_isConnected=%s)\n", __PRETTY_FUNCTION__, self, _isConnected ? "YES" : "NO");
#endif
   if( ! _isConnected)
   {
      NSLog( @"%s: no connection established for %@", __PRETTY_FUNCTION__, self);
      return;
   }
   
   //
   // dat macht das Original nicht
   // das ist aber besser so
   //
   rover = NSEnumerateMapTable( displayGroupBindings( self));
   while( NSNextMapEnumeratorPair( &rover, (void *) &unused, (void *) &group))
   {
#if DEBUG_VERBOSE
      NSLog( @"%@ stop observation of %@ for changes", self, group);
#endif      
      [EOObserverCenter removeObserver:self
                             forObject:group];
   }
   NSEndMapTableEnumeration( &rover);

   // make sure pending observations are ignored
   [[self observerQueue] dequeueObserver:self];
   
   _isConnected = 0;
   // this
   if( [self shouldRetainObjectWhileBeingEstablished])
   {
#if DEBUG
      [_object release];
#else 
      // this is safer and more forgiving
      [_object autorelease];
#endif   
   }
#if DEBUG
   [self release];
#else 
   // this is safer and more forgiving
   [self autorelease];
#endif   
}


- (id) object
{
   return( _object);
}


- (void) bindAspect:(NSString *) aspectName
       displayGroup:(EODisplayGroup *) dest
                key:(NSString *) key
{
   // check things here, it doesn't matter perf. wise
   NSAssert2( [aspectName isKindOfClass:[NSString class]], @"%s: name is a %@ - need a NSString", __PRETTY_FUNCTION__, [aspectName class]);
   NSAssert2( [dest isKindOfClass:[EODisplayGroup class]], @"%s: key is a %@ - need a EODisplayGroup", __PRETTY_FUNCTION__, [dest class]);
   NSAssert2( ! key || [key isKindOfClass:[NSString class]], @"%s: key is a %@ - need a NSString", __PRETTY_FUNCTION__, [key class]);
#if DEBUG   
   if( key && ! [key length])
      NSLog( @"warning: Would binding aspect \"%@\" to %@ with empty key, key ignored", aspectName, dest);
#endif
   
   NSMapInsert( displayGroupBindings( self), aspectName, dest);
   if( [key length])
      [keyBindings( self) setObject:key
                             forKey:aspectName];
}


- (void) unbindAspect:(NSString *) aspectName
{
   // check things here, it doesn't matter perf. wise
   NSAssert2( [aspectName isKindOfClass:[NSString class]], @"%s: name is a %@ - need a NSString", __PRETTY_FUNCTION__, [aspectName class]);

   NSMapRemove( displayGroupBindings( self), aspectName);
   [keyBindings( self) removeObjectForKey:aspectName];
}


- (EODisplayGroup *) displayGroupForAspect:(NSString *) aspectName
{
   NSParameterAssert( [aspectName isKindOfClass:[NSString class]]);
   
   return( NSMapGet( displayGroupBindings( self), aspectName));
}


- (NSString *) displayGroupKeyForAspect:(NSString *) aspectName
{
   NSParameterAssert( [aspectName isKindOfClass:[NSString class]]);

   return( [keyBindings( self) objectForKey:aspectName]);
}


- (BOOL) canBindAspect:(NSString *) aspectName
          displayGroup:(EODisplayGroup *) displayGroup
                   key:(NSString *) key
{
   NSParameterAssert( [aspectName isKindOfClass:[NSString class]]);
   NSParameterAssert( [displayGroup isKindOfClass:[EODisplayGroup class]]);
   NSParameterAssert( ! key || [key isKindOfClass:[NSString class]]);

   return( YES);
}


- (BOOL) endEditing
{
   return( YES);
}


- (EOObserverPriority) priority
{
   return( EOObserverPriorityFourth);  // changed from third, so I can wedge in TableViews!!
}


- (void) subjectChanged
{
#if DEBUG_VERBOSE
   NSLog( @"%s (%@)", __PRETTY_FUNCTION__, self);
#endif   
}


//
// i don't think associationClassesSuperseded comes into play
// here yet...
//
+ (NSArray *) associationClassesForObject:(id) object
{
   NSArray          *classes;
   NSEnumerator     *rover;
   Class            associationClass;
   NSMutableArray   *array;

   array   = [NSMutableArray array];
   classes = loadedSubclassesOf( [EOAssociation class]);
   rover   = [classes objectEnumerator];

   while( associationClass = [rover nextObject])
   {
      if( [associationClass isUsableWithObject:object])
         [array addObject:associationClass];
   }
   return( array);
}


// kinda nice for debugging sometimes
- (id) mulleSelectedValueObject
{
   EODisplayGroup   *group;
   id               obj;
   
   group = NSMapGet( displayGroupBindings( self), MulleEOInterfaceValueAspect);
   obj   = [group selectedObject];
   return( obj);
}


- (id) valueForAspect:(NSString *) aspectName
{
   EODisplayGroup   *group;
   NSString         *key;
   id               value;
   
   NSParameterAssert( [aspectName isKindOfClass:[NSString class]]);

   group = NSMapGet( displayGroupBindings( self), aspectName);
   key   = [keyBindings( self) objectForKey:aspectName];
   if( ! key)
      return( nil);
   
   value = [group selectedObjectValueForKey:key];
   return( value);
}


- (BOOL) setValue:(id) value
        forAspect:(NSString *) aspectName
{
   EODisplayGroup   *group;
   NSString         *key;

   NSParameterAssert( [aspectName isKindOfClass:[NSString class]]);
   
   group = NSMapGet( displayGroupBindings( self), aspectName);
   key   = [keyBindings( self) objectForKey:aspectName];

   if( [group editingAssociation] != self)
      if( [group endEditing] == NO)
         return( NO);
   
   return( [group setSelectedObjectValue:value
                                  forKey:key]);
}


- (NSException *) validateValue:(id *) value
                      forAspect:(NSString *) aspectName
{
   EODisplayGroup   *group;
   NSString         *key;

   NSParameterAssert( [aspectName isKindOfClass:[NSString class]]);

   group = NSMapGet( displayGroupBindings( self), aspectName);

   //
   // This should make sense...
   //
   if( ! [group validatesChangesImmediately] || [group inQueryMode])
      return( nil);
   
   key = [keyBindings( self) objectForKey:aspectName];
   
   return( [[group selectedObject] validateValue:value
                                          forKey:key]);
}


- (id) valueForAspect:(NSString *) aspectName
              atIndex:(unsigned) index
{
   EODisplayGroup   *group;
   NSString         *key;

   NSParameterAssert( [aspectName isKindOfClass:[NSString class]]);
   
   group = NSMapGet( displayGroupBindings( self), aspectName);
   key   = [keyBindings( self) objectForKey:aspectName];

   return( [group valueForObjectAtIndex:index
                                    key:key]);
}


- (BOOL) setValue:(id) value
        forAspect:(NSString *) aspectName
          atIndex:(unsigned) index
{
   EODisplayGroup   *group;
   NSString         *key;

   NSParameterAssert( [aspectName isKindOfClass:[NSString class]]);

   group = NSMapGet( displayGroupBindings( self), aspectName);
   key   = [keyBindings( self) objectForKey:aspectName];

   // this is important. if you don't do this user typed values will
   // be miraculously lost, because our change forces a change in 
   // the object, which will be noticed by the group, which in turn
   // will redisplay the editingAssociation, but the displayed value
   // will be the old value, overwriting the current edit
   if( [group editingAssociation] != self)
      if( [group endEditing] == NO)
         return( NO);

   return( [group setValue:value
          forObjectAtIndex:index
                       key:key]);
}


- (BOOL) shouldEndEditingForAspect:(NSString *) aspectName
                      invalidInput:(NSString *) string
                  errorDescription:(NSString *) errorDescription
{
   EODisplayGroup  *group;

   NSParameterAssert( [aspectName isKindOfClass:[NSString class]]);
   NSParameterAssert( ! string || [string isKindOfClass:[NSString class]]);
   NSParameterAssert( ! errorDescription || [errorDescription isKindOfClass:[NSString class]]);

   group = [self displayGroupForAspect:aspectName];
   return( [group association:self
        failedToValidateValue:string
                       forKey:[self displayGroupKeyForAspect:aspectName]
                       object:[group selectedObject]
             errorDescription:errorDescription]);
}



- (BOOL) shouldEndEditingForAspect:(NSString *) aspectName
                      invalidInput:(NSString *) string
                  errorDescription:(NSString *) errorDescription
                             index:(unsigned) index
{
   EODisplayGroup  *group;

   NSParameterAssert( [aspectName isKindOfClass:[NSString class]]);
   NSParameterAssert( ! string || [string isKindOfClass:[NSString class]]);
   NSParameterAssert( ! errorDescription || [errorDescription isKindOfClass:[NSString class]]);

   group = [self displayGroupForAspect:aspectName];
   return( [group association:self
        failedToValidateValue:string
                       forKey:[self displayGroupKeyForAspect:aspectName]
                       object:[[group displayedObjects] objectAtIndex:index]
             errorDescription:errorDescription]);
}


- (void) copyMatchingBindingsFromAssociation:(EOAssociation *) other
{
   NSMapTable           *otherDisplayGroups;
   NSMutableDictionary  *otherKeys;
   NSEnumerator         *rover;
   NSString             *aspectName;
   EODisplayGroup       *group;
   NSString             *key;
   
   NSParameterAssert( [other isKindOfClass:[EOAssociation class]]);
   
   otherDisplayGroups = displayGroupBindings( other);
   otherKeys          = keyBindings( other);
   
   rover = [[isa aspects] objectEnumerator];
   while( aspectName = [rover nextObject])
   {
      group = NSMapGet( otherDisplayGroups, aspectName);
      if( group)
      {
         key   = [otherKeys objectForKey:aspectName];
         [self bindAspect:aspectName
             displayGroup:group
                      key:key];
      }
   }
}


// default method for asking a value for a bool
- (BOOL) boolForValue:(id) value
{
   if( ! value)
      return( NO);
   
   if( [value respondsToSelector:@selector( boolValue)])
      return( [value boolValue]);
   
   if( [value respondsToSelector:@selector( count)])
      return( [value count] ? YES : NO);
   
   return( YES);
}


//
// this is default logic for pulling a bool value
// from an aspect
//
- (BOOL) boolForAspect:(NSString *) aspect
{
   EODisplayGroup   *group;

   NSParameterAssert( [aspect isKindOfClass:[NSString class]]);

   //
   // if enabled is not wired, then the default is YES
   //
   group = [self displayGroupForAspect:aspect];
   if( ! group)
      return( NO); 	// no aspect set, then NO

   //
   // if the group to edit - assuming enabled group if
   // editingGroup was nil - is in query mode, then
   // say NO regardless..
   //
   if( [group inQueryMode])
      return( NO);

   //
   // Try to read the value from the selected enabled
   // object. if there are multiple values just say yes
   // if there is none say no. If there is one, try to
   // extract a bool value from it
   //
   return( [self boolForValue:[self valueForAspect:aspect]]);
}


- (EOAssociationObjectState) objectStateForAspect:(NSString *) aspect
                             editingDisplayGroup:(EODisplayGroup *) editingGroup
{
   EODisplayGroup   *group;
   NSString         *key;
   id               value;
   
   NSParameterAssert( [aspect isKindOfClass:[NSString class]]);
   NSParameterAssert( ! editingGroup || [editingGroup isKindOfClass:[EODisplayGroup class]]);
   
   //
   // if enabled is not wired, then the default is YES
   //
   group = [self displayGroupForAspect:aspect];
   if( ! group)
      return( EOAssociationObjectStateKeepState); 	// no aspect set, then nothing

   //
   // if the group to *edit* has a multiple selection then
   // enabling does not make sense, since the control is
   // unlikely to be able to edit multiple objects
   //
   if( editingGroup)
   {
      if( [[editingGroup selectedObjects] count] != 1)
         return( EOAssociationObjectStateDisabled);
   }
   else
      editingGroup = group;

   //
   // if the group to edit - assuming enabled group if
   // editingGroup was nil - is in query mode, then
   // say YES regardless..
   //
   if( [editingGroup inQueryMode])
      return( EOAssociationObjectStateEnabled);

   //
   // Try to read the value from the selected enabled
   // object. if there are multiple values just say yes
   // if there is none say no. If there is one, try to
   // extract a bool value from it
   //
   key = [self displayGroupKeyForAspect:aspect];
   if( key)
   {
      if( [[group selectedObjects] count] == 1)
      {
         value = [group selectedObjectValueForKey:key];
         if( [value isKindOfClass:[NSNumber class]])
            return( (int) [value boolValue]);
      }
      value = [[group selectedObjects] lastObject];
   }
   else
      value = [group selectedObject];
   
   return( value != nil);
}


- (EOAssociationObjectState) objectStateForEnabledAspect
{
   return( [self objectStateForAspect:MulleEOInterfaceEnabledAspect
                editingDisplayGroup:nil]);
}


- (EOAssociationObjectState) objectStateForAspect:(NSString *) aspect
{
   return( [self objectStateForAspect:aspect
                  editingDisplayGroup:nil]);
}


- (BOOL) isConnected
{
   return( _isConnected);
}


@class NSButton;

static NSString   *objectDescription( id obj)
{
   NSString  *s;
   
   s = [obj description];
   if( [obj isKindOfClass:[NSButton class]])
      return( [NSString stringWithFormat:@"%@ \"%@\"", s, [obj title]]);
   return( s);
}


- (NSString *) debugDescription
{
   NSString   *key;
   NSString   *aspect;
   
   aspect = [isa primaryAspect];
   if( ! aspect)
      aspect = MulleEOInterfaceValueAspect;
   key = [self displayGroupKeyForAspect:aspect];
   if( ! key)
      key = @"N/A";
   return( [NSString stringWithFormat:@"<%@ %p %@=\"%@\"(%@ %@)>", isa, self, aspect, key, objectDescription( [self object]), [self mulleSelectedValueObject]]);
}


- (void) addBoundKeysForDisplayedObjectsToSet:(NSMutableSet *) set
{
   // does nothing
}

@end


