/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF
 *  Copyright (C) 2006 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 *
 */
#import "_EOInterfaceUniqueObjectsAndTags.h"


#if MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_3

static int   compare_object_with_tag(void *sel, const void *a, const void *b)
{
   return( (int) objc_msgSend( ((_EOInterfaceObjectWithTag *) a)->obj, 
                               sel, 
                               ((_EOInterfaceObjectWithTag *) b)->obj));
}

#else

static int   compare_object_with_tag(void *sel, const void *a, const void *b)
{
   return( (int) objc_msgSend( ((_EOInterfaceObjectWithTag *) a)->obj, 
                               @selector( compare:), 
                               ((_EOInterfaceObjectWithTag *) b)->obj));
}

#endif


@implementation _EOInterfaceUniqueObjectsAndTags

- initWithCapacity:(unsigned int) max
{
   [super init];
   
   size_ = max;
   storage_  = calloc( sizeof( _EOInterfaceObjectWithTag), max);
   if( ! storage_)
   {
      [self release];
      return( nil);
   }
   
   objects_ = NSCreateHashTable( NSNonRetainedObjectHashCallBacks, max);
   tags_    = NSCreateHashTable( NSIntHashCallBacks, max);
   
   return( self);
}


static void   releaseAll( unsigned int n,  _EOInterfaceObjectWithTag *p)
{
   while( n)
   {
      [p->obj release];
      ++p;
      --n;
   }
}


- (void) dealloc
{
   NSFreeHashTable( objects_);
   NSFreeHashTable( tags_);
   
   releaseAll( count_, storage_);
   free( storage_);
   [super dealloc];
}


static void   growIfNeeded( _EOInterfaceUniqueObjectsAndTags *self)
{
   if( self->count_ >= self->size_)
   {
      self->size_ += self->size_;
      if( self->size_ < 8)
         self->size_ = 8;
      self->storage_ = realloc( self->storage_, sizeof( _EOInterfaceObjectWithTag) * self->size_);
   }
}


// can't have a duplicate object or tag
- (BOOL) addObject:(id) obj
           withTag:(unsigned int) tag
{
   _EOInterfaceObjectWithTag   *p;
   
   NSParameterAssert( obj);
   NSParameterAssert( index != NSNotAnIntMapKey);
   
   if( NSHashInsertIfAbsent( self->objects_, obj))
      return( NO);
   if( tag)
   {
      if( NSHashInsertIfAbsent( self->tags_, (void *) tag))
      {
         NSHashRemove( self->objects_, obj);
         return( NO);
      }
   }
   else
   {
      if( haveZeroTag_)
         return( NO);
      haveZeroTag_ = YES;  // stoopid HashTable
   }
   
   growIfNeeded( self);

   p = &storage_[ count_++];
   p->obj = [obj retain];
   p->tag = tag;

   return( YES);
}


- (BOOL) insertObject:(id) obj
              withTag:(unsigned int) tag
              atIndex:(unsigned int) index
{
   _EOInterfaceObjectWithTag   *p;
   
   NSParameterAssert( index <= count_);
   NSParameterAssert( obj);
   NSParameterAssert( (void *) index != NSNotAnIntMapKey);
   
   if( NSHashGet( self->objects_, obj))
      return( NO);
   if( tag)
   {
      if( NSHashGet( self->tags_, (void *) tag))
         return( NO);
   }
   else
      if( haveZeroTag_)
         return( NO);
   
   growIfNeeded( self);
   
   if( index != count_)
      memcpy( &storage_[ index + 1], &storage_[ index], (count_ - index) * sizeof( _EOInterfaceObjectWithTag));
      
   count_++;

   p = &storage_[ index];
   p->obj = [obj retain];
   p->tag = tag;
   
   NSHashInsertKnownAbsent( objects_, obj);
   if( tag)
      NSHashInsertKnownAbsent( tags_, (void *) tag);
   else
      haveZeroTag_ = YES;
   
   return( YES);
}


- (unsigned int) count
{
   return( count_);
}


- (id) objectAtIndex:(unsigned int) i
{
   _EOInterfaceObjectWithTag   *p;
   
   NSParameterAssert( i < count_);
   
   p = &storage_[ i];
   return( p->obj);
}


- (unsigned int) tagAtIndex:(unsigned int) i
{
   _EOInterfaceObjectWithTag   *p;
   
   NSParameterAssert( i < count_);
   
   p = &storage_[ i];
   return( p->tag);
}


- (void) sortUsingSelector:(SEL) sel
{
#if MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_3
   qsort_r( storage_, count_, sizeof( _EOInterfaceObjectWithTag), sel, compare_object_with_tag);
#else
   NSAssert( sel == @selector( compare:), @"selector must be compare: when running under 10.2");
   qsort( storage_, count_, sizeof( _EOInterfaceObjectWithTag), compare_object_with_tag);
#endif
}


- (NSArray *) allIndexes
{
   NSMutableArray   *array;
   unsigned int     i;
   
   array = [NSMutableArray arrayWithCapacity:count_];
   for( i = 0; i < count_; i++)
      [array addObject:[NSNumber numberWithUnsignedInt:storage_[ i].tag]];
   return( array);
}


- (NSArray *) allObjects
{
   NSMutableArray   *array;
   unsigned int     i;
   
   array = [NSMutableArray arrayWithCapacity:count_];
   for( i = 0; i < count_; i++)
      [array addObject:storage_[ i].obj];
   return( array);
}


// zzzzzzloooow
- (unsigned int) tagForObject:(id) object
{
   unsigned int     i;
   for( i = 0; i < count_; i++)
      if( [storage_[ i].obj isEqual:object])
         return( storage_[ i].tag);
   return( (unsigned int) NSNotAnIntMapKey);
}


- (void) dump
{
   unsigned int   i;
   
   fprintf( stderr, "[\n");
   for( i = 0; i < count_; i++)
      fprintf( stderr, "   #%u = %s, %u\n", i, [storage_[ i].obj cString], storage_[ i].tag);
   fprintf( stderr, "]\n");
}
@end
