/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOActionAssociation.h"

#import <AppKit/AppKit.h>
#ifdef GNUSTEP
#include <gnustep/base/objc-gnu2next.h>
#else
#import <objc/objc-runtime.h>
#endif
#import "EODisplayGroup.h"
#import "EODisplayGroup+Selection.h"
#import "AspectsAndKeys.h"



@implementation EOActionAssociation

+ (NSString *) primaryAspect
{
   return( MulleEOInterfaceActionAspect);  // or ??
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceActionAspect, MulleEOInterfaceArgumentAspect, MulleEOInterfaceEnabledAspect, nil];
   return( aspects);
}


+ (NSArray *)  aspectSignatures
{
   static NSArray  * aspectSignatures;

   if( !  aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"A", @"A1M", @"A", nil];

   return(  aspectSignatures);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:MulleEOInterfaceTargetKey, nil];
   
   return( objectKeysTaken);
}


+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[NSControl class]]);
}


- (void) establishConnection
{
   EOAssociationObjectState   state;

#if DEBUG   
   if( ! [self displayGroupForAspect:MulleEOInterfaceActionAspect])
   {
      NSLog( @"warning: %@ action aspect not connected", self);
   }
   else
      if( ! [self displayGroupKeyForAspect:MulleEOInterfaceActionAspect])
      {
         NSLog( @"warning: %@ action aspect no key connected", self);
      }
#endif
   
   [_object setTarget:self];
   [_object setAction:@selector( action:)];

   state = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
      [_object setEnabled:(BOOL) state];

   [super establishConnection];
}


- (void) breakConnection
{
   id    obj;
   
   obj = [self object];
   [obj setTarget:nil];
   [obj setAction:0];
   [super breakConnection];
}


- (EOAssociationObjectState) objectStateForEnabledAspect
{
   return( [self objectStateForAspect:MulleEOInterfaceEnabledAspect
                editingDisplayGroup:[self displayGroupForAspect:MulleEOInterfaceActionAspect]]);
}


- (void) action:(id) sender
{
   EODisplayGroup   *displayGroup;
   NSString         *string;
   id               argument;
   SEL              selector;
   NSArray          *objects;
   id               object;
   unsigned int     i, n;
   
// superflous safety ??
//   if( ! [self objectStateForAspect:MulleEOInterfaceEnabledAspect])
//      return;

   string   = [self displayGroupKeyForAspect:MulleEOInterfaceActionAspect];
   selector = NSSelectorFromString( string);

   displayGroup = [self displayGroupForAspect:MulleEOInterfaceArgumentAspect];
   object       = [displayGroup selectedObject];
   
   if( ! displayGroup)
   {
      // to account for second example in
      // EOActionAssociation documentation (no key set, take sel obj)
      argument = object;
   }
   else
      argument = [object valueForKey:string];

   objects = [[self displayGroupForAspect:MulleEOInterfaceActionAspect] selectedObjects];
   n = [objects count];
   if( n == 0)
      return;

   object = [objects lastObject];
   if( ! [object respondsToSelector:selector])
   {
      NSLog( @"%@: Objects of Class %@ do not respond to action key %@", self, [object class], string);
      return;
   }

#if GNUSTEP   
   for( i = 0; i < n; i++)
      [[objects objectAtIndex: i] performSelector:selector
                                       withObject: argument];
#else
   for( i = 0; i < n; i++)
      objc_msgSend( [objects objectAtIndex:i], selector, argument);
#endif   
}


- (void) subjectChanged
{
   EOAssociationObjectState   oldstate;
   EOAssociationObjectState   state;

#if DEBUG_VERBOSE
   NSLog( @"%s (%@)", __PRETTY_FUNCTION__, self);
#endif   
   state = [self objectStateForEnabledAspect];
   if( state == EOAssociationObjectStateKeepState)
      return;
   
   oldstate = [_object isEnabled];
   if( state != oldstate)
      [_object setEnabled:(BOOL) state];
}


@end


