/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOActionInsertionAssociation.h"

#import "EODisplayGroup.h"
#import "EODisplayGroup+InsertUpdateDelete.h"
#import "EODisplayGroup+Selection.h"
#import "AspectsAndKeys.h"
#import <AppKit/AppKit.h>


@implementation EOActionInsertionAssociation


+ (NSString *) primaryAspect
{
   return( MulleEOInterfaceSourceAspect);  // or ??
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceSourceAspect, MulleEOInterfaceDestinationAspect, MulleEOInterfaceEnabledAspect, nil];
   return( aspects);
}


+ (NSArray *)  aspectSignatures
{
   static NSArray  * aspectSignatures;

   if( !  aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"", @"1M", @"A1M", nil];

   return(  aspectSignatures);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:MulleEOInterfaceTargetKey, nil];

   return( objectKeysTaken);
}


+ (BOOL) isUsableWithObject:(id) object
{
   return( [object respondsToSelector:@selector( setAction:)] &&
           [object respondsToSelector:@selector( setTarget:)] && 
           [object respondsToSelector:@selector( setEnabled:)]);
}



- (void) establishConnection
{
   EOAssociationObjectState   state;
   
   if( ! [self displayGroupForAspect:MulleEOInterfaceSourceAspect])
   {
      NSLog( @"%@ source aspect not connected", self);
      return;
   }

   if( ! [self displayGroupForAspect:MulleEOInterfaceDestinationAspect])
   {
      NSLog( @"%@ destination aspect not connected", self);
      return;
   }

   state = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
      [_object setEnabled:(BOOL) state];
   [_object setTarget:self];
   [_object setAction:@selector( action:)];

   [super establishConnection];
}


- (void) breakConnection
{
   NSControl  *obj;
   
   obj = [self object];
   [obj setTarget:nil];
   [obj setAction:0];

   [super breakConnection];
}


- (void) action:(id) sender
{
   EODisplayGroup   *target;
   EODisplayGroup   *source;
   NSString         *key;
   NSArray          *selection;
   id               targetObject;
   id               original;
   id               value;
   
   target = [self displayGroupForAspect:MulleEOInterfaceDestinationAspect];
   source = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   key    = [self displayGroupKeyForAspect:MulleEOInterfaceDestinationAspect];
   
   targetObject = [target selectedObject];
   if( ! targetObject) // humm why am I not disabled ??
   {
#if DEBUG      
      NSLog( @"*** %s %p's action called with no selected object in place for aspect %@", __PRETTY_FUNCTION__, self, MulleEOInterfaceDestinationAspect);
#endif      
      return;
   }

   original  = [targetObject valueForKey:key];
   selection = [source selectedObjects];

   NSAssert( [[NSSet setWithArray:selection] count] == [selection count], @"Selection contains duplicates");
   
   // try to figure out if we need to insert
   // into a tomany or to one relationship
   // use EOClassDescription conveniences (might have figured that out 
   // earlier...)
   
   if( [[targetObject toManyRelationshipKeys] containsObject:key])
   {
      NSMutableArray   *set;
      NSEnumerator     *rover;
      id               object;
  
      set = [NSSet setWithArray:original];
      
      NSAssert( [set count] == [original count], @"Original value contains duplicates");
      
      rover = [selection objectEnumerator];
      while( object = [rover nextObject])
      {
         if( [set containsObject:object])
            continue;
         
         [targetObject addObject:object
toBothSidesOfRelationshipWithKey:key];
      }
      return;
   }

   //
   // we only deal with single setting then
   if( [selection count] != 1)
      return;
   
   value = [selection objectAtIndex:0];
   if( value == original)
      return;
   [target setSelectedObjectValue:value
                           forKey:key];
}


- (EOAssociationObjectState) objectStateForEnabledAspect
{
   return( [self objectStateForAspect:MulleEOInterfaceEnabledAspect
                  editingDisplayGroup:[self displayGroupForAspect:MulleEOInterfaceDestinationAspect]] &&
           [[[self displayGroupForAspect:MulleEOInterfaceSourceAspect] selectedObjects] count] > 0);
}


- (void) subjectChanged
{
   BOOL                       oldstate;
   EOAssociationObjectState   state;

#if DEBUG_VERBOSE
   NSLog( @"%s (%@)", __PRETTY_FUNCTION__, self);
#endif   
   state = [self objectStateForEnabledAspect];
   if( state == EOAssociationObjectStateKeepState)
      return;

   oldstate = [_object isEnabled];
   if( state != oldstate)
      [_object setEnabled:(BOOL) state];
}


@end


