/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOColumnAssociation.h"

#import <AppKit/AppKit.h>
#import "AspectsAndKeys.h"
#import "EODisplayGroup.h"
#import "EODisplayGroup+Associations.h"
#import "EODisplayGroup+InsertUpdateDelete.h"
#import "EODisplayGroup+Objects.h"
#import "EOTableViewAssociation.h"
#import "NSString+MulleKeyPrettyPrint.h"
#import "EOColumnAssociationProxy.h"
#import "EOAssociation+RenderCellByAspect.h"


#define ENCODING_VERSION   229
#define USE_STUPID_HACK    0

#define ALSO_SET_SORT_DESCRIPTOR  0  // academic
#define ONLY_HIT_INDICATOR        1  // more academia

// 
// Column Associations are pretty fragile. As the tableview can be manipulated
// by the user to remove columns, the "object" we have can be stale or 
// deallocated already. What we do is to ask the tableView *everytime* if our
// column still exists. If not, we break the connection ourselves.
// For that purposed our instance variable object_ contains the possibly stale
// NSTableColumn, but we always refresh it when we are called using
// [self object].
//
@implementation EOColumnAssociation

+ (void) initialize
{
   [EOColumnAssociation setVersion:ENCODING_VERSION];
}


+ (NSString *) primaryAspect
{
   return( MulleEOInterfaceValueAspect);  // or ??
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceValueAspect, MulleEOInterfaceEnabledAspect, nil];
   return( aspects);
}


+ (NSArray *)  aspectSignatures
{
   static NSArray  * aspectSignatures;

   if( !  aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"A", @"A1M", nil];

   return(  aspectSignatures);
}


+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[NSTableColumn class]]);
}

#if USE_STUPID_HACK
- (id) object
{
   NSTableColumn   *column;
   
   // we are live! use the tableView_
   if( ! [_object isKindOfClass:[NSTableView class]])
      return( _object);
   column = [_object tableColumnWithIdentifier:self];
   return( column);
}


static BOOL  attemptStupidHack( EOColumnAssociation *self, NSTableColumn *column)
{
   NSTableView   *tableView;
   
   tableView = [column tableView];
   if( tableView)
   {
      self->_object = tableView;       // we keep the tableView in _object
      [column setIdentifier:self];  // and memorize us in the column like this
      return( YES);
   }
   return( NO);
}

#endif


- (id) initWithObject:(id) obj
{
   NSParameterAssert( [obj isKindOfClass:[NSTableColumn class]]);
   
   [super initWithObject:obj];

   //
   // this looks like it should be done in "establishConnection" alas
   // backwards compatibility forces us to do it here
   //
#if USE_STUPID_HACK
   attemptStupidHack( self, obj);
#endif                                 
   return( self);
}


- (void) _giveHeaderCellATitleIfNeeded
{
   NSTableColumn   *column;
   NSCell          *cell;
   NSString        *title;
   NSString        *s;
   
   column = [self object];
   cell   = [column headerCell];
   
   if( ! [cell image])
   {
      title = [cell stringValue];
      if( ! [title length])
      {
         title = [self displayGroupKeyForAspect:MulleEOInterfaceValueAspect];
         s = [title pathExtension];  // cleverly grab last part of path
         if( [s length])
            title = s;
         title = [title mulleDecapslifiedMethodString:YES]; 
      }
      if( ! title)
         title = @"???";
      [cell setStringValue:title];
   }   
}


#if ALSO_SET_SORT_DESCRIPTOR      
- (void) _reflectSortOrderSelectorInSortDescriptor
{
   NSTableView     *tableView;
   NSTableColumn   *tableColumn;
   NSString        *key;
   id              sortDescriptor;
   NSImage         *image;
   BOOL            flag;
   
   tableColumn = [self object];
   
   sortDescriptor = nil;
   if( _sortingSelector)
   {
      flag = _sortingSelector == EOCompareAscending || _sortingSelector == EOCompareCaseInsensitiveAscending;
      key  = [self displayGroupKeyForAspect:MulleEOInterfaceValueAspect];
      sortDescriptor = [[NSClassFromString( @"NSSortDescriptor") alloc] initWithKey:key
                                                                          ascending:flag
                                                                           selector:_sortingSelector];
   }
   
   [tableColumn setSortDescriptorPrototype:sortDescriptor];
   [sortDescriptor release];
}
#endif


- (unsigned int) sortIndicatorWidth
{
   static unsigned int  width;
  
   if( ! width)
      width = [[NSImage imageNamed:@"NSAscendingSortIndicator"] size].width;
   return( width);
}


- (void) _reflectSortOrderInTableColumn
{
   NSTableView     *tableView;
   NSTableColumn   *tableColumn;
   NSImage         *image;
   BOOL            flag;
   
   tableColumn = [self object];
   
   image = nil;
   if( _sortingSelector)
   {
      flag  = _sortingSelector == EOCompareAscending || _sortingSelector == EOCompareCaseInsensitiveAscending;
      image = [NSImage imageNamed:flag ? @"NSAscendingSortIndicator" :  @"NSDescendingSortIndicator"];
   }
   
   tableView = [tableColumn tableView];
   [tableView setIndicatorImage:image
                  inTableColumn:tableColumn];
}


- (void) _reflectSortOrderSelector
{
#if ALSO_SET_SORT_DESCRIPTOR      
   [self _reflectSortOrderSelectorInSortDescriptor];
#endif   

   if( ! _showIndicatorImage)
      return;
   [self _reflectSortOrderInTableColumn];
}


- (void) establishConnection
{
   EODisplayGroup           *group;
   NSTableColumn            *column;
   EOTableViewAssociation   *parent;
   
   // actually unusal, see above bizarre hack
#if USE_STUPID_HACK
   if( [_object isKindOfClass:[NSTableColumn class]])
      attemptStupidHack( self, _object);
   NSParameterAssert( [_object isKindOfClass:[NSTableView class]]); // sic!
#endif

   [EOColumnAssociationProxy hookColumnAssociation:self
                                     toTableColumn:_object];
   
   //
   // The first EOColumnAssociation to be bound automatically creates and binds
   // an EOTableView association.
   //
   group  = [self displayGroupForAspect:MulleEOInterfaceValueAspect];
   parent = [EOTableViewAssociation bindToTableView:[_object tableView]
                                       displayGroup:group];
   
   column = [self object];
   
   if( ! column)
   {
      NSLog( @"Do not establish connections on NSTableColumns, that aren't wired to the table yet");
#if DEBUG
      abort();
#endif
      return;
   }
   
   _colorAspectBound   = [self displayGroupKeyForAspect:MulleEOInterfaceTextColorAspect] != nil;
   _boldAspectBound    = [self displayGroupKeyForAspect:MulleEOInterfaceBoldAspect] != nil;
   _italicAspectBound  = [self displayGroupKeyForAspect:MulleEOInterfaceItalicAspect] != nil;
   _enabledAspectBound = [self displayGroupKeyForAspect:MulleEOInterfaceEnabledAspect] != nil;

   [self _giveHeaderCellATitleIfNeeded];
   
   // EOColumnAssociation is sometimes used by other assocs
   if( [parent respondsToSelector:@selector( sortMode)])
   {
      unsigned int   sortMode;
      
      sortMode = [parent sortMode];
      _sortingEnabled      = sortMode != EOCompatibilitySortMode; 
      _cycleWithNilSorting = sortMode & EOAllowNilInSorts;
      _showIndicatorImage  = sortMode & EOShowSortIndicatorImage;
   }

   if( [parent respondsToSelector:@selector( sortsByColumnOrder)])
      if( [parent sortsByColumnOrder])
         _sortingEnabled = NO;
   
   if( _sortingEnabled) 
   {
      if( ! _cycleWithNilSorting)
         _sortingSelector = EOCompareAscending;
      [self _reflectSortOrderSelector];
   }

   [super establishConnection];
}


- (void) breakConnection
{
   id   obj;
   
   obj = [self object];
#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@", self, __PRETTY_FUNCTION__, _object);
#endif
   [EOColumnAssociationProxy unhookColumnAssociationFromTableColumn:obj];
   [super breakConnection];
}


- (void) setSortingEnabled:(BOOL) flag
{
   _sortingEnabled  = flag;
#if ALSO_SET_SORT_DESCRIPTOR      
   [self _reflectSortOrderSelectorInSortDescriptor];
#endif   
   [self _reflectSortOrderInTableColumn];
}


- (BOOL) sortingEnabled
{
   return( _sortingEnabled);
}


- (void) _setSortingSelectorIfNeeded:(SEL) selector
{
   if( _sortingSelector == NULL)
      [self setSortingSelector:selector];
}


- (void) setSortingSelector:(SEL) selector
{
   _sortingSelector        = selector;
   _caseInsensitiveSorting = selector == EOCompareCaseInsensitiveAscending || selector == EOCompareCaseInsensitiveDescending;
   [self _reflectSortOrderSelector];
}


- (SEL) sortingSelector
{
   return( _sortingSelector);
}


- (void) tableColumnDealloced
{
#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@", self, __PRETTY_FUNCTION__, _object);
#endif
   _object = nil;
   [super breakConnection];
}


#if DEBUG

- (id) retain
{
   return( [super retain]);
}


- (void) release
{
   [super release];
}


- (id) autorelease
{
   return( [super autorelease]);
}


- (void) dealloc
{
   [super dealloc];
}

#endif


static BOOL   boolForEnablingObjectAndKey( id object, NSString *key)
{
   if( ! key)
      return( YES);
   if( [object isKindOfClass:[NSNumber class]])
      return( [object boolValue]);
   return( object != nil);
}


static inline id   emptyValueForCell( NSCell *cell)
{
   return( [cell isKindOfClass:[NSTextFieldCell class]] ? @"" : nil);
}


static inline id   emptyValueForTableColumn( NSTableColumn *column)
{
   return( emptyValueForCell( [column dataCell]));
}


static inline id   emptyValueForTableColumnAndRow( NSTableColumn *column, int row)
{
   return( emptyValueForCell( [column dataCellForRow:row]));
}


//
// We don't like it
//
- (BOOL) endEditing
{
#if DEBUG_VERBOSE
   NSLog( @"%@ %s", self, __PRETTY_FUNCTION__);
#endif
   return( NO);
}


- (id)          tableView:(NSTableView *) tableView
objectValueForTableColumn:(NSTableColumn *) column
                      row:(int) row
{
   id               value;
   id               eo;
   NSArray          *objects;
   unsigned int     count;
   NSString         *valueKey;
   EODisplayGroup   *valueGroup;
   
   valueGroup = [self displayGroupForAspect:MulleEOInterfaceValueAspect];
   objects    = [valueGroup displayedObjects];
   count      = [objects count];

   value = nil;
   if( row < count)
   {
      eo       = [objects objectAtIndex:row];
      valueKey = [self displayGroupKeyForAspect:MulleEOInterfaceValueAspect];
      value    = [eo valueForKeyPath:valueKey];
   }
   return( value ? value : emptyValueForTableColumnAndRow( column, row));
}


- (void)   tableView:(NSTableView *) tableView
      setObjectValue:(id) value
      forTableColumn:(NSTableColumn *) tableColumn
                 row:(int) row
{
   EODisplayGroup   *valueGroup;
   NSString         *key;
   
#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@ %@ %@ %u", self, __PRETTY_FUNCTION__, tableView, value, tableColumn, row);
#endif
   
   valueGroup = [self displayGroupForAspect:MulleEOInterfaceValueAspect];
   key        = [self displayGroupKeyForAspect:MulleEOInterfaceValueAspect];
   if( ! [valueGroup setValue:value
             forObjectAtIndex:row
                          key:key])
   {	
      NSArray  *array;

      array = [valueGroup displayedObjects];
      if( row >= [array count])
      {
         NSLog( @"Index out of bounds %u", row);
         return;
      }

      NSLog( @"Setting of object value %@ for key %@ failed for EO %@ at row %u",
             value, key, [array objectAtIndex:row], row);
   }
}


- (void)   tableView:(NSTableView *) tableView
     willDisplayCell:(id) cell
      forTableColumn:(NSTableColumn *) tableColumn
                 row:(int) row
{
   [self _renderCell:cell
       enabledAspect:_enabledAspectBound ?  MulleEOInterfaceEnabledAspect : nil
         colorAspect:_colorAspectBound ? MulleEOInterfaceTextColorAspect : nil
          boldAspect:_boldAspectBound ? MulleEOInterfaceBoldAspect : nil
        italicAspect:_italicAspectBound ? MulleEOInterfaceItalicAspect : nil
              forRow:row];
}


- (BOOL)     tableView:(NSTableView *) tableView
 shouldEditTableColumn:(NSTableColumn *) tableColumn
                   row:(int) row
{
   id   object;

#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@ %@ %u", self, __PRETTY_FUNCTION__, tableView, tableColumn, row);
#endif
   object = [[[self displayGroupForAspect:MulleEOInterfaceEnabledAspect] displayedObjects] objectAtIndex:row];
   return( boolForEnablingObjectAndKey( object, [self displayGroupKeyForAspect:MulleEOInterfaceEnabledAspect]));
}


- (void) _toggleSortingSelector
{
   if( _sortingSelector == EOCompareAscending)
   {
      _sortingSelector = EOCompareDescending;
      return;
   }
   if( _sortingSelector == EOCompareCaseInsensitiveAscending)
   {
      _sortingSelector = EOCompareCaseInsensitiveDescending;
      return;
   }
   

   if( _sortingSelector == EOCompareDescending)
   {
      _sortingSelector = _cycleWithNilSorting ? nil : EOCompareAscending;
      return;
   }
   if( _sortingSelector == EOCompareCaseInsensitiveDescending)
   {
      _sortingSelector = _cycleWithNilSorting ? nil : EOCompareCaseInsensitiveAscending;
      return;
   }
   
   if( _sortingSelector != nil)  // can't handle others
      return;
   
#if DEBUG
   if( ! _cycleWithNilSorting)
      NSLog( @"%@ should not have nil as an initial sort", self);
#endif
   
   if( _caseInsensitiveSorting)
   {
      _sortingSelector = EOCompareCaseInsensitiveAscending;
      return;
   }
   
   _sortingSelector = EOCompareAscending;
}


- (EOSortOrdering *) sortOrdering
{
   NSString         *key;
   EOSortOrdering   *sort;
   
   if( ! _sortingSelector)
      return( nil);
   
   key  = [self displayGroupKeyForAspect:MulleEOInterfaceValueAspect];
   sort = [[[EOSortOrdering alloc] initWithKey:key
                                      selector:_sortingSelector] autorelease];
   return( sort);
}


- (void)   tableView:(NSTableView *) tableView
 didClickTableColumn:(NSTableColumn *) tableColumn
{
#if ONLY_HIT_INDICATOR   
   NSEvent             *event;
   NSPoint             mouseXY;
   NSPoint             xy;
   float               width;
   unsigned int        hitWidth;
   NSTableHeaderView   *headerView;
   NSRect              rect;
   unsigned int        i;
#endif
   
   // if we don't sort, ignore
   if( ! _sortingEnabled)
      return;

   /* check that the general area of the sorting selector
    * is hit, otherwise selecting columns is too hard
    */
#if ONLY_HIT_INDICATOR   
   headerView = [tableView headerView];
   event      = [[tableView window] currentEvent];
   mouseXY    = [event locationInWindow];
   xy         = [headerView convertPoint:mouseXY
                                fromView:nil];
   
   width = 0;
   for( i = 0;; i++)
   {
      rect   = [headerView headerRectOfColumn:i];
      width += rect.size.width;
      if( xy.x < width)
         break;
   }
   
   hitWidth = [self sortIndicatorWidth] + 4 + 4;  /* image + 4 pix border */
   
   if( rect.size.width <= hitWidth || xy.x < (width - hitWidth))
      return;
#endif   
   {
      [self _toggleSortingSelector];
      [self _reflectSortOrderSelector];
   }
   /* otherwise do what.. column selection */
}


- (BOOL)   control:(NSControl *) control
     isValidObject:(id) obj
{
#if DEBUG_VERBOSE
   NSLog( @"%s %@ %@", __PRETTY_FUNCTION__, control, obj);
#endif
   return( YES);
}


- (BOOL)       control:(NSControl *) control
 didFailToFormatString:(NSString *) string
      errorDescription:(NSString *) errorDescription
{
   return( [self shouldEndEditingForAspect:MulleEOInterfaceValueAspect
                              invalidInput:string
                          errorDescription:errorDescription]);
}

- (BOOL)        control:(NSControl *) control
 textShouldBeginEditing:(NSText *) fieldEditor
{
#if DEBUG_VERBOSE
   NSLog( @"%s %@ %@", __PRETTY_FUNCTION__, control, fieldEditor, [fieldEditor string]);
#endif
   return( YES);
}


- (BOOL)        control:(NSControl *) control
   textShouldEndEditing:(NSText *) fieldEditor
{
   id           value;
   id           original;
   NSException  *exception;

#if DEBUG_VERBOSE
   NSLog( @"%s %@ %@", __PRETTY_FUNCTION__, control, [fieldEditor string]);
#endif   
   original  = [[[fieldEditor string] copy] autorelease];
   value     = original;
   exception = [self validateValue:&value
                         forAspect:MulleEOInterfaceValueAspect];
   if( exception != nil)
      return( [self shouldEndEditingForAspect:MulleEOInterfaceValueAspect
                                 invalidInput:original
                             errorDescription:[exception reason]]);
   
   if( original && ! [original isEqual:value])
      [fieldEditor setString:value];
   
   return( YES);
}


- (void)   controlTextDidBeginEditing:(NSNotification *) notification
{
   EODisplayGroup   *valueGroup;
#if DEBUG_VERBOSE
   NSLog( @"%@ %s", self, __PRETTY_FUNCTION__);
#endif
   valueGroup = [self displayGroupForAspect:MulleEOInterfaceValueAspect];
   [valueGroup associationDidBeginEditing:self];
}


// i don't know if this actually gets called...
- (void) controlTextDidEndEditing:(NSNotification *) notification
{
   EODisplayGroup   *valueGroup;
   
#if DEBUG_VERBOSE
   NSLog( @"%@ %s", self, __PRETTY_FUNCTION__);
#endif
#if 0   
   NSText        *fieldEditor;
   id            value;
   NSException   *exception;

   fieldEditor = [[notification userInfo] objectForKey:@"NSFieldEditor"];
   value = [[fieldEditor string] copy];
   exception = nil;
NS_DURING   
   [self setValue:value
        forAspect:MulleEOInterfaceValueAspect];
NS_HANDLER
   exception = localException;
NS_ENDHANDLER

   [value release];
#endif
   valueGroup = [self displayGroupForAspect:MulleEOInterfaceValueAspect];
   [valueGroup associationDidEndEditing:self];
//   [exception raise];
}


- (NSString *) debugDescription
{
   NSTableColumn   *column;
   
   column = [self object];
   return( [NSString stringWithFormat:@"<%@ %p (%p column \"%@\", %@)>", 
      isa, self, column, [[column headerCell] stringValue], [self mulleSelectedValueObject]]);
}


+ (EOColumnAssociation *) columnAssociationForTableColumn:(NSTableColumn *) column
{
   id   identifier;
   
   identifier = [column identifier];
   if( ! identifier)
      return( nil);
   if( _isa( identifier) == [EOColumnAssociationProxy class])
      return( [(EOColumnAssociationProxy *) identifier columnAssociation]);
   if( [identifier isKindOfClass:[EOColumnAssociation class]])
      return( (EOColumnAssociation *) identifier);
   return( nil);
}


//
// this gets called often, so make it speedy
//
+ (NSArray *) columnAssociationsForTableView:(NSTableView *) tableView
{
   NSTableColumn    *column;
   id               proxy;
   Class            proxyClass;
   Class            realClass;
   NSArray          *columns;
   NSMutableArray   *array;
   unsigned int     i, n;
   
   array = [NSMutableArray array];
   
   proxyClass = [EOColumnAssociationProxy class];
   realClass  = Nil;
   
   columns = [tableView tableColumns];
   n       = [columns count];
   for( i = 0; i < n; i++)
   {
      column = [columns objectAtIndex:i];
      proxy  = [column identifier];
      if( ! proxy)
         continue;
      
      if( [proxy isKindOfClass:proxyClass])
      {
         [array addObject:[proxy columnAssociation]];
         continue;
      }
      
      if( ! realClass)
         realClass  = [EOColumnAssociation class];
      //
      // for ultra backwards compatibility...
      if( [proxy isKindOfClass:realClass])
         [array addObject:proxy];
   }
   return( array);
}


- (BOOL) shouldRetainObjectWhileBeingEstablished
{
   return( NO);
}


static void   SafelyAddDisplayGroupKeyToSet( EOAssociation *self, 
                                             NSMutableSet *set, 
                                             NSString *aspect)
{
   NSString   *key;
   
   key = [self displayGroupKeyForAspect:aspect];
   if( key)
      [set addObject:key];
}


- (void) addBoundKeysForDisplayedObjectsToSet:(NSMutableSet *) set
{
   SafelyAddDisplayGroupKeyToSet( self, set, MulleEOInterfaceTextColorAspect);
   SafelyAddDisplayGroupKeyToSet( self, set, MulleEOInterfaceBoldAspect);
   SafelyAddDisplayGroupKeyToSet( self, set, MulleEOInterfaceItalicAspect);
   SafelyAddDisplayGroupKeyToSet( self, set, MulleEOInterfaceEnabledAspect);
   SafelyAddDisplayGroupKeyToSet( self, set, MulleEOInterfaceValueAspect);
}

@end


