/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF
 *  Copyright (C) 2006 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 *
 */
#import "Defines.h"

#import "EOColumnAssociationProxy.h"

#import <AppKit/AppKit.h>
#import "EOColumnAssociation.h"

#define DEBUG_PROXY     (0 || DEBUG_VERBOSE)

@implementation EOColumnAssociationProxy

+ (void) hookColumnAssociation:(EOColumnAssociation *) assoc
                 toTableColumn:(NSTableColumn *) column
{
   NSAutoreleasePool   *pool;
   id                  proxy;

   NSParameterAssert( [column isKindOfClass:[NSTableColumn class]]);
   NSParameterAssert( [assoc isKindOfClass:[EOColumnAssociation class]]);
   NSParameterAssert( _isa( assoc) != [EOColumnAssociationProxy class]);
   NSAssert( ! [[column identifier] isKindOfClass:[EOColumnAssociationProxy class]] ||
                      [[column identifier] columnAssociation] != assoc, @"you must not hook the same association twice");
   
   proxy = [[self alloc] initWithColumnAssociation:assoc];
#if DEBUG_PROXY
   NSLog( @"%@ %s %@ %@", proxy, __PRETTY_FUNCTION__, assoc, column);
#endif
   pool = [NSAutoreleasePool new];
   [column setIdentifier:proxy];
   [pool release];
   
   NSParameterAssert( [proxy retainCount] == 2);
   [proxy release];  // only column holding us
#if DEBUG_NO_RETAIN   
   [proxy disableRetain];
#endif   
}


+ (void) unhookColumnAssociationFromTableColumn:(NSTableColumn *) column
{
   NSAutoreleasePool   *pool;
   id                  proxy;

   //
   // if we get unhooked, then don't tel the columnAssociation anything
   // it's unhooking us.
   //
   proxy = [column identifier];
   if( _isa( proxy) != [EOColumnAssociationProxy class])
   {
      NSLog( @"TableColumn %@ identifier is no longer a %@ as expected", column, self);
#if DEBUG
      abort();
#endif      
      return;
   }
   
#if DEBUG_PROXY
   NSLog( @"%@ %s %@ %@", proxy, __PRETTY_FUNCTION__, [proxy columnAssociation], column);
#endif
   [proxy setColumnAssociation:nil];
   
   //
   // what we really don't want is that we stay autoreleased, because
   // that would mean we are dangling in the autoreleasePool and the
   // association might notice to late that the object is gone.
   //
   pool = [NSAutoreleasePool new];
   [column setIdentifier:nil];
   [pool release];
}


- (id) initWithColumnAssociation:(EOColumnAssociation *) assoc
{
   NSParameterAssert( [assoc isKindOfClass:[EOColumnAssociation class]]);

   assoc_ = assoc;
   
   return( self);
}


#if DEBUG_NO_RETAIN
- (void) disableRetain
{
   noRetain_ = YES;
}


- (id) retain
{
   if( noRetain_)
      [NSException raise:NSInternalInconsistencyException
                  format:@"The %@ Proxy can not be retained (otherwise its useless)", self];
   return( [super retain]);
}
#endif

   
- (void) dealloc
{
#if DEBUG_PROXY
   NSLog( @"%@ %s", self, __PRETTY_FUNCTION__);
#endif
   [assoc_ tableColumnDealloced];
   [super dealloc];
}


#if DEBUG
- (id) autorelease
{
   return( [super autorelease]);
}
#endif


- (void) setColumnAssociation:(EOColumnAssociation *) assoc
{
   assoc_ = assoc;
}


- (EOColumnAssociation *) columnAssociation
{
   return( assoc_);
}


//
// fake (it to the max)
//
- (Class) class
{
   if( assoc_)
      return( [assoc_ class]);
   return( [super class]);
}


- (BOOL) isKindOfClass:(Class) class
{
   if( isa == class)
      return( YES);
   if( assoc_)
      return( [assoc_ isKindOfClass:class]);
   return( [super isKindOfClass:class]);
}


- (NSMethodSignature *) methodSignatureForSelector:(SEL) sel
{
   if( assoc_)
      return( [assoc_ methodSignatureForSelector:sel]);
   return( [super methodSignatureForSelector:sel]);
}


- (void) forwardInvocation:(NSInvocation *) invocation
{
   [invocation invokeWithTarget:assoc_]; 
}


/*
 * STUPID HACK II
 *
 * InterfaceBuilder asks the NSTableColumn for its identifier and assumes it's
 * a string. This is not really true to the API. So we fake a few string 
 * methods...
 */
- (unsigned int) length
{
   return( 0);
}


- (id) copyWithZone:(NSZone *) zone
{
   return( [self retain]);
}

@end
