/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOComboBoxAssociation.h"

#import <AppKit/AppKit.h>
#import "AspectsAndKeys.h"
#import "EOAssociation+PopUpComboSupport.h"
#import "_EOInterfaceUniqueObjectsAndTags.h"
#import "EOActionAssociation.h"
#import "EOControlAssociation.h"
#import "EODisplayGroup.h"
#import "EODisplayGroup+Associations.h"
#import "EODisplayGroup+Selection.h"
#import "EODisplayGroup+Objects.h"


@implementation EOComboBoxAssociation


+ (NSString *) primaryAspect
{
   return( MulleEOInterfaceTitlesAspect);  // or ??
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceTitlesAspect, MulleEOInterfaceSelectedTitleAspect,
         MulleEOInterfaceSelectedObjectAspect, MulleEOInterfaceEnabledAspect, nil];
   return( aspects);
}


+ (NSArray *) aspectSignatures
{
   static NSArray  *aspectSignatures;
   
   if( ! aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"A", @"A", @"1", @"A1M", nil];
   
   return( aspectSignatures);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:MulleEOInterfaceTargetKey, MulleEOInterfaceDataSourceKey, MulleEOInterfaceDelegateKey, nil];
   return( objectKeysTaken);
}



+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[NSComboBox class]]);
}


+ (NSArray *) associationClassesSuperseded
{
   static NSArray  *associationClassesSuperseded;

   if( ! associationClassesSuperseded)
      associationClassesSuperseded = [[NSArray alloc] initWithObjects:[EOControlAssociation class],
         [EOActionAssociation class], nil];
   return( associationClassesSuperseded);
}


static void   uncacheFilteredTitles( EOComboBoxAssociation *self)
{
   id  old;

   old = self->filteredTitles_;
   self->filteredTitles_ = 0;
   [self->filteredTitles_ release];
   old = self->objectIndexes_;
   self->objectIndexes_ = 0;
   [self->objectIndexes_ release];
}


static void   cachefilteredTitles( EOComboBoxAssociation *self)
{
   _EOInterfaceUniqueObjectsAndTags   *titlesAndTags;
   
   titlesAndTags = [self uniquedStringsForAspect:MulleEOInterfaceTitlesAspect];
   if( ! self->disableSort_)
      [titlesAndTags sortUsingSelector:@selector( compare:)];
   
   self->filteredTitles_  = [titlesAndTags allObjects];
   [self->filteredTitles_ retain];

   self->objectIndexes_   = [titlesAndTags allIndexes];
   [self->objectIndexes_ retain];
   
   self->nFilteredTitles_ = [self->filteredTitles_ count];
}


static inline NSArray   *filteredTitles( EOComboBoxAssociation *self)
{
   if( ! self->filteredTitles_)
      cachefilteredTitles( self);
   return( self->filteredTitles_);
}


static inline NSArray   *objectIndexes( EOComboBoxAssociation *self)
{
   if( ! self->objectIndexes_)
      cachefilteredTitles( self);
   return( self->objectIndexes_);
}


- (EOAssociationObjectState) objectStateForEnabledAspect
{
   return( [self objectStateForAspect:MulleEOInterfaceEnabledAspect
                  editingDisplayGroup:[self displayGroupForAspect:MulleEOInterfaceValueAspect]]);
}


- (void) establishConnection
{
   EOAssociationObjectState   state;
   
   _wasEnabled = [_object isEnabled];

   _selectedObjectConnected = [self displayGroupForAspect:MulleEOInterfaceSelectedObjectAspect] != nil;
   _selectedTitleConnected  = [self displayGroupForAspect:MulleEOInterfaceSelectedTitleAspect] != nil;

#if DEBUG   
   if( _selectedObjectConnected + _selectedTitleConnected )
      NSLog( @"%@ : Both selectedTitle and selectedObject connected. It may not be what you want.", self);
#endif
   
   // ensure if both are connected, that they both point to the same group
   NSParameterAssert( ! (_selectedTitleConnected && _selectedObjectConnected) || 
                      [self displayGroupForAspect:MulleEOInterfaceSelectedObjectAspect] == 
                      [self displayGroupForAspect:MulleEOInterfaceSelectedTitleAspect]);
   
   state = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
      [_object setEnabled:(BOOL) state];
   [_object setTarget:self];
   [_object setAction:@selector( objectChanged:)];
   
   if( [self displayGroupForAspect:MulleEOInterfaceTitlesAspect])
   {
      [_object setUsesDataSource:YES];
      [_object setDataSource:self];
   }
   [_object setDelegate:self];

   [super establishConnection];
}


- (void) breakConnection
{
   id    old;
   id    obj;

   obj = [self object];
   [obj setDelegate:nil];
   [obj setAction:nil];
   [obj setTarget:nil];
   if( [obj usesDataSource])
   {
      [obj setDataSource:nil];
      [obj setUsesDataSource:NO];
   }
   uncacheFilteredTitles( self);
   _selectedObjectConnected = NO;
   _selectedTitleConnected  = NO;
   
   old = _lastStringValue;
   _lastStringValue = nil;
   [old release];

   [obj setEnabled:_wasEnabled];

   [super breakConnection];
}


- (BOOL) endEditing
{
   EODisplayGroup   *group;
   
   group = [self displayGroupForAspect:_selectedTitleConnected ? MulleEOInterfaceSelectedTitleAspect : MulleEOInterfaceSelectedObjectAspect];
   if( group)
   {
      [self objectChanged:self];    // or keep on editing if that failed ???
      [group associationDidEndEditing:self];
   }
   return( YES);
}


- (void) selectItemWithTitle:(NSString *) title
{
   if( ! [title isEqualToString:_lastStringValue])
   {
      id   old;
      
      [_object setStringValue:title];
//      [_object selectItemWithObjectValue:title];

      old = _lastStringValue;
      _lastStringValue = [title copyWithZone:[self zone]];
      [old release];
   }
}


static BOOL  syncSelectionWithObjectOfGroup( EOComboBoxAssociation *self, EODisplayGroup *group, EODisplayGroup *titlesGroup)
{
   NSString   *s;
   NSString   *key;
   NSString   *path;
   NSString   *keyPath;
   id         eo;
   
   eo = [group selectedObject];
   if( eo == nil)
   {
      [self selectItemWithTitle:@""];
      return( NO);
   }

   // figure out the objects behind the title
   path    = [self displayGroupKeyForAspect:MulleEOInterfaceSelectedObjectAspect];
   key     = [self displayGroupKeyForAspect:MulleEOInterfaceTitlesAspect];
   keyPath = [path stringByAppendingFormat:@".%@", key];
   
   s = [eo valueForKeyPath:keyPath]; // sic!
   if( ! s)
      s = @"";
   [self selectItemWithTitle:s];
   return( YES);
}


static BOOL  syncSelectionWithSelectedTitleOfGroup( EOComboBoxAssociation *self, EODisplayGroup *group)
{
   NSString  *s;
   NSString  *key;

   key = [self displayGroupKeyForAspect:MulleEOInterfaceSelectedTitleAspect];
   s   = [[group selectedObject] valueForKeyPath:key];
   if( s)
   {
      s = [s description];
      [self selectItemWithTitle:s];
   }
   else
      [self selectItemWithTitle:@""];
   return( YES);
}


static BOOL  syncSelectionWithTitlesOfGroup( EOComboBoxAssociation *self, EODisplayGroup *group)
{
   NSString  *s;
   NSString  *key;
   id        eo;
   
   eo = [group selectedObject];
   if( eo == nil)
   {
      [self selectItemWithTitle:@""];
      return( NO);
   }
   
   key = [self displayGroupKeyForAspect:MulleEOInterfaceTitlesAspect];
   
   s = [eo valueForKeyPath:key]; // sic!
   s = [s description];
   if( ! s)
      s = @"";
   [self selectItemWithTitle:s];
   return( YES);
}




- (void) subjectChanged
{
   EODisplayGroup  *titlesGroup;
   EODisplayGroup  *valueGroup;
   EODisplayGroup  *group;
   BOOL            oldstate;
   BOOL            state;
   BOOL            flag;

#if DEBUG
   NSLog( @"%s (%@)", __PRETTY_FUNCTION__, self);
#endif
   state      = [self objectStateForEnabledAspect];
   oldstate   = [_object isEnabled];
   valueGroup = [self displayGroupForAspect:MulleEOInterfaceValueAspect];

   flag = NO;
   if( titlesGroup = [self displayGroupForAspect:MulleEOInterfaceTitlesAspect])
   {
      if( [titlesGroup contentsChanged])
      {
         flag = YES;
         uncacheFilteredTitles( self);
         [_object reloadData];
      }
   }

   if( _selectedObjectConnected)
   {
      group = [self displayGroupForAspect:MulleEOInterfaceSelectedObjectAspect];
      if( [group selectionOrContentsChanged])
      {
         flag = YES;
         if( ! syncSelectionWithObjectOfGroup( self, group, titlesGroup))
            if( state != EOAssociationObjectStateKeepState)
               state = NSOffState;
      }
   }

   if( _selectedTitleConnected)
   {
      group = [self displayGroupForAspect:MulleEOInterfaceSelectedTitleAspect];
      if( [group selectionOrContentsChanged])
      {
         flag = YES;
         if( ! syncSelectionWithSelectedTitleOfGroup( self, group))
            if( state != EOAssociationObjectStateKeepState)
               state = NSOffState;
      }
   }

   if( ! _selectedObjectConnected && ! _selectedTitleConnected)
   {
      if( [titlesGroup selectionChanged])
      {
         flag = YES;
         if( ! syncSelectionWithTitlesOfGroup( self, titlesGroup))
            if( state != EOAssociationObjectStateKeepState)
               state = NSOffState;
      }
   }

   if( state != EOAssociationObjectStateKeepState && state != oldstate)
   {
      flag = YES;
      [_object setEnabled:state];
   }
   
   if( flag)
      [_object setNeedsDisplay:YES];   
}


- (void) objectChanged:(id) sender
{
   int              index;
   EODisplayGroup   *titlesGroup;
   EODisplayGroup   *group;
   NSArray          *titles;
   NSString         *title;
   NSString         *string;
   NSString         *key;
   id               eo;
   id               old;
   id               value;
   
   eo     = nil;
   string = [_object stringValue];
   old    = _lastStringValue;
   _lastStringValue = [string copyWithZone:[self zone]];
   [old release];
   
   title = nil;

   index = [filteredTitles( self) indexOfObject:string];
   if( index != NSNotFound)
   {
      title = string;
      index = [[objectIndexes( self) objectAtIndex:index] intValue];
   }

   // string matches a title in our Combobox
   if( index != NSNotFound)
   {
      if( _selectedObjectConnected)
      {
         titlesGroup = [self displayGroupForAspect:MulleEOInterfaceTitlesAspect];
         titles      = [titlesGroup displayedObjects];
         value       = [titles objectAtIndex:index];
         eo          = [[self displayGroupForAspect:MulleEOInterfaceSelectedObjectAspect] selectedObject];
         key         = [self displayGroupKeyForAspect:MulleEOInterfaceSelectedObjectAspect];
         [eo setValue:value
           forKeyPath:key];
      }

      if( _selectedTitleConnected)
      {
         [self setValue:title
              forAspect:MulleEOInterfaceSelectedTitleAspect];
      }
      
      if( ! _selectedObjectConnected && ! _selectedTitleConnected)
      {  // operate on titles only...
         group = [self displayGroupForAspect:MulleEOInterfaceTitlesAspect];
         eo    = [[group displayedObjects] objectAtIndex:index];
         [group selectObject:eo];
      }
      return;
   }

   // no match
   if( _selectedObjectConnected)
   {
      eo    = [[self displayGroupForAspect:MulleEOInterfaceSelectedObjectAspect] selectedObject];
      key   = [self displayGroupKeyForAspect:MulleEOInterfaceSelectedObjectAspect];
      [eo setValue:nil
        forKeyPath:key];
   }
   
   if( _selectedTitleConnected)
   {
      // special combo box thing, here we allow the user to write the title!
      if( ! title)
         title = string;
      [self setValue:title
           forAspect:MulleEOInterfaceSelectedTitleAspect];
   }

   if( ! _selectedObjectConnected && ! _selectedTitleConnected)
   {  // operate on titles only...
      group = [self displayGroupForAspect:MulleEOInterfaceTitlesAspect];
      [group clearSelection];
   }
}


- (void) comboBoxSelectionDidChange:(NSNotification *) notification
{
// we get called later again by the target/action 
//   [self objectChanged:self];
}



- (NSString *) comboBox:(NSComboBox *) aComboBox
        completedString:(NSString *) aString
{
   unsigned int    i, n;
   unsigned int    min;
   NSString        *s;
   NSString        *lower;
   NSArray         *titles;
   NSString        *memo;

   min = [aString length];
   if( ! min)
      return( aString);

   memo    = nil;
   lower   = [aString lowercaseString];
   titles  = filteredTitles( self);
   n       = [titles count];
   
   for( i = 0; i < n; i++)
   {
      s = [titles objectAtIndex:i];
      if( ! [[s lowercaseString] hasPrefix:lower])
         continue;

      if( ! memo)
         memo = s; // sic!
      else
         memo = [s commonPrefixWithString:memo
                                  options:NSCaseInsensitiveSearch];
      if( [memo length] == min)
         break;
   }

   return( memo ? memo : aString);
}


- (int) numberOfItemsInComboBox:(NSComboBox *)aComboBox
{
   return( [filteredTitles( self) count]);
}


- (id) comboBox:(NSComboBox *) aComboBox
 objectValueForItemAtIndex:(int)index
{
   return( [filteredTitles_ objectAtIndex:index]);
}


- (unsigned int) comboBox:(NSComboBox *) aComboBox 
indexOfItemWithStringValue:(NSString *) aString
{
   unsigned int    i, n;
   NSArray         *titles;

   titles = filteredTitles( self);
   n = [titles count];
   for( i = 0; i < n; i++)
      if( [[titles objectAtIndex:i] isEqualToString:aString])
         return( i);
   return( NSNotFound);
   
}


- (void) controlTextDidBeginEditing:(NSNotification *) aNotification
{
   [[self displayGroupForAspect:MulleEOInterfaceSelectedTitleAspect] associationDidBeginEditing:self];
}


- (void) controlTextDidEndEditing:(NSNotification *) aNotification
{
   [self endEditing];
}


static void   SafelyAddDisplayGroupKeyToSet( EOAssociation *self, 
                                             NSMutableSet *set, 
                                             NSString *aspect)
{
   NSString   *key;
   
   key = [self displayGroupKeyForAspect:aspect];
   if( key)
      [set addObject:key];
}


- (void) addBoundKeysForDisplayedObjectsToSet:(NSMutableSet *) set
{
   SafelyAddDisplayGroupKeyToSet( self, set, MulleEOInterfaceTitlesAspect);
}


@end


