/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOControlAssociation.h"

#import <AppKit/AppKit.h>
#import "AspectsAndKeys.h"
#import "EODisplayGroup.h"
#import "EODisplayGroup+Associations.h"


@implementation EOControlAssociation


+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[NSControl class]]);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:MulleEOInterfaceTargetKey, MulleEOInterfaceDelegateKey, nil];
   
   return( objectKeysTaken);
}


- (NSControl *) control
{
   return( _object);
}


- (EOGenericControlAssociation *) editingAssociation
{
   return( self);
}



- (void) establishConnection
{
   NSControl  *control;
   
   [super establishConnection];

   control = [self control];
#if DEBUG
   if( [control isKindOfClass:[NSForm class]])
   {
      unsigned int   i;
      unsigned int   j;
      unsigned int   rows;
      unsigned int   columns;
      NSFormCell     *cell;
      BOOL           flag;
      
      rows    = [(NSForm *) control numberOfRows];
      columns = [(NSForm *) control numberOfColumns];
      cell    = (NSFormCell *) [(NSForm *) control prototype];
      flag    = ! [cell sendsActionOnEndEditing];
      if( ! flag)
         for( j = 0; j < rows; j++)
            for( i = 0; i < columns; i++)
            {
               cell = (NSFormCell *) [(NSForm *) control cellAtRow:j
                                                            column:i];
               
               flag |= ! [cell sendsActionOnEndEditing];
            }

      if( flag)
         NSLog( @"*** %s <%@ %p> Warning: NSForm %@ cells may not send ActionOnEndEditing. Edits may be therefore be lost", __PRETTY_FUNCTION__, [self class], self, control);
   }
#endif      
   if( [control respondsToSelector:@selector( setDelegate:)])
      [(id) control setDelegate:self];
}


- (void) breakConnection
{
   NSControl  *control;

   control = [self control];
   if( [control respondsToSelector:@selector( setDelegate:)])
      [(id) control setDelegate:nil];

   [super breakConnection];
}


#warning (nat) straighten up use of _objectChanged objectChanged and associationDidEndEditing order
- (BOOL) endEditing
{
   BOOL   flag;
   
   flag = [self _objectChanged]; // i think _objectChanged is used for historic reasons
   // or keep on editing if that failed ???
   [[self displayGroupForAspect:MulleEOInterfaceValueAspect] associationDidEndEditing:self];
   return( YES);
}


// when is this really called ??
#if 0
- (BOOL)        control:(NSControl *) control 
 textShouldBeginEditing:(NSText *) fieldEditor
{
   [[self displayGroupForAspect:MulleEOInterfaceValueAspect] associationDidBeginEditing:self];
   return( YES);  // i mean , why not ??
}
#endif


- (BOOL)      control:(NSControl *) control
 textShouldEndEditing:(NSText *) fieldEditor
{
   id           value;
   id           original;
   NSException  *exception;
   
   // DL this means that the control must have a formatter!
   // (in case of a string value in a text field, this formatter
   // must return the unmodified string! is this right?)
   original  = [control objectValue];
   value     = original;
   exception = [self validateValue:&value
                         forAspect:MulleEOInterfaceValueAspect]; 
   if( exception != nil)
      return( [self shouldEndEditingForAspect:MulleEOInterfaceValueAspect
                                 invalidInput:original
                             errorDescription:[exception reason]]);

   // validation might have changed value
   if( original && ! [original isEqual:value])
      [control setObjectValue:value];
   return( YES);
}


// supposedly we should handle this
// control:didFailToFormatString:errorDescription: 

#warning (nat) might want to push this down to "generic" level...
- (void) controlTextDidBeginEditing:(NSNotification *) aNotification
{
   [[self displayGroupForAspect:MulleEOInterfaceValueAspect] associationDidBeginEditing:self];
}


static EODisplayGroup  *beginEditingIfNotYet( EOControlAssociation *self)
{
   EODisplayGroup   *dg;
   
   dg = [self displayGroupForAspect:MulleEOInterfaceValueAspect];
   if( [dg editingAssociation] != self)
      [dg associationDidBeginEditing:self];
   return( dg);
}


- (void) controlTextDidChange:(NSNotification *) notification
{
   _editedYet = YES;
   beginEditingIfNotYet( self);
}


- (void) objectChanged
{
   EODisplayGroup  *dg;
   
   dg = beginEditingIfNotYet( self);
   [dg associationDidEndEditing:self];
   [super objectChanged]; 
   _editedYet = NO;
//   [dg associationDidEndEditing:self];  // danger of recursion! so do it before objectChanged
}   


//
// we usually get called twice because of the action. So for NSTextFields
// the action is not wired up. This is a bug, because we can't choose
// the RETURN as enter method
//
- (void) controlTextDidEndEditing:(NSNotification *) aNotification
{
#warning (nat) possibly use "_objectChange" and if that fails...
   // updating GUI with current state may be to painful for the user...
   if( _editedYet)
      [self objectChanged];
}

@end


