/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOGenericControlAssociation.h"

#import <AppKit/AppKit.h>
#import "AspectsAndKeys.h"
#import "EODisplayGroup.h"


#define USE_LAST_VALUE   0   // i think it's shit


@implementation EOGenericControlAssociation

+ (NSString *) primaryAspect
{
   return( MulleEOInterfaceValueAspect);
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceValueAspect, MulleEOInterfaceEnabledAspect, nil];
   return( aspects);
}


+ (NSArray *) aspectSignatures
{
   static NSArray  *aspectSignatures;
   
   if( ! aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"A", @"A1M", nil];
   
   return( aspectSignatures);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:MulleEOInterfaceTargetKey, MulleEOInterfaceDelegateKey, nil];
   return( objectKeysTaken);
}


+ (BOOL) isUsableWithObject:(id) object
{
   return( NO);
}


#if USE_LAST_VALUE
- (void) _setLastValue:(id) value
{
   id   old;
   
   old = _lastValue;
   _lastValue = [value retain];
   [old release];
}
#endif


- (EOAssociationObjectState) objectStateForEnabledAspect
{
   return( [self objectStateForAspect:MulleEOInterfaceEnabledAspect
                  editingDisplayGroup:[self displayGroupForAspect:MulleEOInterfaceValueAspect]]);
}


- (void) establishConnection
{
   EOAssociationObjectState   state;
   id                         obj;
   
   if( ! [self displayGroupForAspect:MulleEOInterfaceValueAspect])
   {
      NSLog( @"%@ value aspect not connected", self);
      return;
   }
   
   if( ! [self displayGroupKeyForAspect:MulleEOInterfaceValueAspect])
   {
      NSLog( @"%@ value aspect no key connected", self);
      return;
   }

   obj = [self object];
   if( ! [obj isKindOfClass:[NSTextField class]])
   {
      [obj setTarget:self];
      [obj setAction:@selector( action:)];
   }
   
   state = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
      [_object setEnabled:(BOOL) state];

#if USE_LAST_VALUE
   [self _setLastValue:[obj objectValue]];
#endif   
   [super establishConnection];
}


- (void) breakConnection
{
   id   obj;

   obj = [self object];
   [obj setTarget:nil];
   [obj setAction:0];

   [super breakConnection];
}


- (NSControl *) control
{
   return( nil);
}


- (EOGenericControlAssociation *) editingAssociation
{
   return( nil);
}


- (void) dealloc
{
   [_lastValue release];  // for subclasses
   [super dealloc];
}


- (BOOL) setValue:(id) value
        forAspect:(NSString *) MulleEOInterfaceValueAspect
{
   if( _copyObjects)
      value = [[value copy] autorelease];
   
   // mirror UI in lastValue regardless
#if USE_LAST_VALUE
   [self _setLastValue:value];
#endif   
   if( [super setValue:value
             forAspect:MulleEOInterfaceValueAspect])
   {
      return( YES);
   }
   
   return( NO);
}


// Since this is also used by EOActionCell, we ask the object not the control
// for objectValues...
//
- (void) subjectChanged
{
   id          value;
   id          oldValue;
   BOOL        old;
   BOOL        state;
   BOOL        flag;
   id          obj;
   
#if DEBUG_VERBOSE
   NSLog( @"%s (%@)", __PRETTY_FUNCTION__, self);
#endif
   obj   = [self object];
   old   = [obj isEnabled];
   state = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
   {
      flag = state != old;
      if( flag)
         [obj setEnabled:state];
   }

   value = [self valueForAspect:MulleEOInterfaceValueAspect];
   
#if USE_LAST_VALUE
   //
   // if objectValue and _lastValue differ, then _lastValue is no good
   // indeed, but some of the associations are too tricky to enforce this
   //
   NSParameterAssert( [obj objectValue] == _lastValue || [[obj objectValue] isEqual:_lastValue]);
   oldValue = _lastValue;
#else
   oldValue = [obj objectValue];
#endif
   
   //
   // this _lastValue seems like a stupid idea, except if it's very costly
   // for the UI element to provide the value (especially by copying) do 
   // we know that it has to copy ???
   //
#if DEBUG_SUBJECT_CHANGED
   NSLog( @"%s %p control %@, object %@ value %p (before %p)", __PRETTY_FUNCTION__, self, [self control], obj, value, oldValue);
#endif     
   
   if( value != oldValue || (value && ! [value isEqual:oldValue]))
   {
      if( _copyObjects)
         value = [[value copy] autorelease];
#if DEBUG_VERBOSE
     NSLog( @"Set %@ to value %@", control, value);
#endif
     [obj setObjectValue:value];
#if USE_LAST_VALUE
     newValue = [obj objectValue];
     NSParameterAssert( ! value && ! newValue || value && newValue || (! value && ([newValue isKindOfClass:[NSString class]] || [newValue isKindOfClass:[NSNumber class]]) && ! [newValue intValue]));
     [self _setLastValue:newValue];  // profit from conversions ...
#endif     
     flag = YES;
   }
   
   if( flag)
      [[self control] setNeedsDisplay:YES];
}


- (BOOL) _objectChanged
{
   id      value;
   BOOL    flag;
   id      obj;
   
   flag  = NO;
   obj   = [self object];
   /*
    * There was an interesting bug here. There was a number 
    * formatter that formatted the numbers to 9,999 but it
    * wasn't able to read it back (strange as it seems) so it
    * returned nil on the second read.
    * 0;0;-1 - Add 1000 Sep. - All others off, Run in German locale.
    */
   value = [obj objectValue];
#if USE_LAST_VALUE
   /* stupid hack, but for some reason we sometimes read @"" back
    * when we put in a nil although the user didn't edit anything 
    */
   if( [value isKindOfClass:[NSString class]] && [value length] == 0 && _lastValue == nil)
      return( flag);
   
   if( value != _lastValue || (value && ! [value isEqual:_lastValue]))
   {
      flag = [self setValue:value
                  forAspect:MulleEOInterfaceValueAspect];
   }
#else
   flag = [self setValue:value
               forAspect:MulleEOInterfaceValueAspect];
#endif
   return( flag);
}


- (void) objectChanged
{
   [self _objectChanged];
}   


- (void) action:(id) sender
{
   [self objectChanged];
}


@end


