/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOMasterCopyAssociation.h"

#import "AspectsAndKeys.h"
#import "EOAssociation+EODisplayGroupAssociation.h"
#import "EODisplayGroup.h"
#import "EODisplayGroup+Objects.h"
#import "EODisplayGroup+Selection.h"
#import "EOInterfaceNotifications.h"

@implementation EOMasterCopyAssociation


+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[EODisplayGroup class]]);
}


+ (NSString *) primaryAspect
{
   return( MulleEOInterfaceParentAspect);  // or ??
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceParentAspect, nil];
   return( aspects);
}


+ (NSArray *) aspectSignatures
{
   static NSArray  *aspectSignatures;

   if( ! aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"", nil];

   return( aspectSignatures);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] init];

   return( objectKeysTaken);
}


- (EOObserverPriority) priority
{
   return( EOObserverPrioritySecond);
}


- (void) establishConnection
{
   EODisplayGroup         *group;
   EODisplayGroup         *dad;
   NSNotificationCenter   *center;

   [super establishConnection];
   [self establishConnectionWithDisplayGroup];

   group = [self object];
   dad   = [self displayGroupForAspect:MulleEOInterfaceParentAspect];
   [group setDataSource:[dad dataSource]];

   center = [NSNotificationCenter defaultCenter];
   [center postNotificationName:EOMasterCopyAssociationEstablishedConnection
                         object:self];
}


- (void) breakConnection
{
   EODisplayGroup   *group;

   group = [self object];
   [group setDataSource:nil];

   [self breakConnectionWithDisplayGroup];
   [super breakConnection];
}


- (void) subjectChanged
{
   EODisplayGroup   *master;
   EODisplayGroup   *group;
   
   group  = [self object];
   master = [self displayGroupForAspect:MulleEOInterfaceParentAspect];
   if( [master contentsChanged])
      [group updateDisplayedObjects];    // displayedObjects updaten
   else
      if( [group contentsChanged])
         [master updateDisplayedObjects];  // displayedObjects updaten

   // selection kopieren
   // the following code works, because we select lazily in the EODisplayGroup
   // we allow the selectedObjects to contain objects that are not displayed
   // objects (on passing in) the user of EODisplayGroup will never notice that...
   if( [master selectionChanged])
      [group selectObjectsIdenticalTo:[master selectedObjects]];
   else
      if( [group selectionChanged])
         [master selectObjectsIdenticalTo:[group selectedObjects]];
}


- (BOOL) shouldRetainObjectWhileBeingEstablished
{
   return( NO);
}

@end


