/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOMasterDetailAssociation.h"

#import "AspectsAndKeys.h"
#import "EOAssociation+EODisplayGroupAssociation.h"
#import "EODisplayGroup.h"
#import "EODisplayGroup+Selection.h"


@implementation EOMasterDetailAssociation


+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[EODisplayGroup class]]);
}


+ (NSString *) primaryAspect
{
   return( MulleEOInterfaceParentAspect);  // or ??
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceParentAspect, nil];
   return( aspects);
}


+ (NSArray *) aspectSignatures
{
   static NSArray  *aspectSignatures;

   if( ! aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"1M", nil];

   return( aspectSignatures);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:nil];

   return( objectKeysTaken);
}


- (EOObserverPriority) priority
{
   return( EOObserverPrioritySecond);
}


- (EODisplayGroup *) _qualifyWithParent:(EODisplayGroup *) parent
                                  force:(BOOL) force
{
   NSString               *key;
   EODisplayGroup         *group;
   EODetailDataSource     *dataSource;
   id                     master;
   NSArray                *selection;
   //
   // the "automatic creation of missing dataSources
   // is compatible with original EOF
   //
   key        = [self displayGroupKeyForAspect:MulleEOInterfaceParentAspect];
   group      = [self object];
   dataSource = (EODetailDataSource *) [group dataSource];
   
   // this fixes "en passant" some faulty nibs I created
   if( force || [dataSource masterDataSource] != [parent dataSource])
   {
      dataSource = (EODetailDataSource *) [[parent dataSource] dataSourceQualifiedByKey:key];
      [group setDataSource:dataSource];
   }
   
   NSParameterAssert( [dataSource isKindOfClass:[EODetailDataSource class]]);
   
   selection = [parent selectedObjects];

   master = nil;
   if( [selection count] == 1)
      master = [selection lastObject];
   
   [dataSource qualifyWithRelationshipKey:key
                                 ofObject:master];
   
   NSAssert3( dataSource, @"%@: %@ has no detailDataSource and parent %@'s dataSource won't create one for us", self, group, parent);
   return( group);
}



- (void) establishConnection
{
   EODisplayGroup         *parent;
   EODisplayGroup         *group;
   EODetailDataSource     *dataSource;
   
   group      = [self object];
   dataSource = (EODetailDataSource *) [group dataSource];
   
   // allow to not have a dataSource 
   // but should we establish the Connection ??
   
   NSAssert1( ! dataSource || [dataSource isKindOfClass:[EODetailDataSource class]], @"DataSource %@ unusable for MasterDetail, need EODetailDataSource", dataSource);

   [super establishConnection];
   [self establishConnectionWithDisplayGroup];

   parent = [self displayGroupForAspect:MulleEOInterfaceParentAspect];
   [self _qualifyWithParent:parent
                      force:YES];
}


- (void) breakConnection
{
   EODetailDataSource   *dataSource;
   EODisplayGroup       *group;
   NSString             *key;
   
   key        = [self displayGroupKeyForAspect:MulleEOInterfaceParentAspect];
   group      = (EODisplayGroup *) [self object];
   dataSource = (EODetailDataSource *)  [group dataSource];
   [dataSource qualifyWithRelationshipKey:key
                                 ofObject:nil];
   [self breakConnectionWithDisplayGroup];
   [super breakConnection];
}


//
// We should try to avoid actual refetches as much as possible
// because it does make the interface shakier
//
- (void) subjectChanged
{
   EODisplayGroup   *group;
   EODisplayGroup   *parent;

#if DEBUG_VERBOSE
   NSLog( @"%s (%@)", __PRETTY_FUNCTION__, self);
#endif   
   parent = [self displayGroupForAspect:MulleEOInterfaceParentAspect];
   if( ! [parent selectionOrContentsChanged])
      return;
   
#if DEBUG_SUBJECT_CHANGED
   NSLog( @"%s %p parent %@", __PRETTY_FUNCTION__, self, parent);
#endif      

   group = [self _qualifyWithParent:parent
                              force:NO];
   [group fetch];
#if DEBUG_SUBJECT_CHANGED
   NSLog( @"%s %p after %p refetches %u displayed, %u selected", __PRETTY_FUNCTION__, self, group, [[group displayedObjects] count], [[group selectedObjects] count]);
#endif      
}


- (BOOL) shouldRetainObjectWhileBeingEstablished
{
   return( NO);
}

@end


