/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOMatrixAssociation.h"

#import <AppKit/AppKit.h>
#import "AspectsAndKeys.h"
#import "EODisplayGroup.h"
#import "EODisplayGroup+Objects.h"
#import "EODisplayGroup+Selection.h"


@implementation EOMatrixAssociation


+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[NSMatrix class]]);
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceEnabledAspect, MulleEOInterfaceImageAspect, MulleEOInterfaceTitleAspect, nil];
   return( aspects);
}


+ (NSArray *) aspectSignatures
{
   static NSArray  *aspectSignatures;
   
   if( ! aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"A1M", @"A", @"A", nil];
   
   return( aspectSignatures);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:MulleEOInterfaceTargetKey, nil];
   return( objectKeysTaken);
}


- (void) establishConnection
{
   EOAssociationObjectState   state;
   NSMatrix                   *matrix;
   
   matrix = [self object];
   
   /* only usurp target/action if we actually manage the matrix */
   if( [self displayGroupForAspect:MulleEOInterfaceImageAspect] ||
       [self displayGroupForAspect:MulleEOInterfaceTitleAspect])
   {
      [matrix setTarget:self];
      [matrix setAction:@selector( action:)];
   }
   
   state = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
      [matrix setEnabled:(BOOL) state];
   
   [super establishConnection];

   // either one must be connected (not true)
/*   
   NSParameterAssert( [self class] == [EOMatrixAssociation class] || 
                      [self displayGroupKeyForAspect:MulleEOInterfaceImageAspect] ||
                      [self displayGroupKeyForAspect:MulleEOInterfaceTitleAspect]);
*/
                      
   // if both are connected, both must point to same displayGroup
   NSParameterAssert( [self class] == [EOMatrixAssociation class] || /* subclasses can do what they want */
                      ! [self displayGroupKeyForAspect:MulleEOInterfaceImageAspect] ||
                      ! [self displayGroupKeyForAspect:MulleEOInterfaceTitleAspect] ||
                      [self displayGroupForAspect:MulleEOInterfaceImageAspect] == 
                      [self displayGroupForAspect:MulleEOInterfaceTitleAspect]);
}


- (void) breakConnection
{
   NSMatrix   *matrix;
   
   if( [self displayGroupForAspect:MulleEOInterfaceImageAspect] ||
       [self displayGroupForAspect:MulleEOInterfaceTitleAspect])
   {
      matrix = [self object];
   
      [matrix setTarget:nil];
      [matrix setAction:0];
   }
   [super breakConnection];
}


static void   setCell( NSCell *cell, NSImage *image, NSString *title, BOOL isEnabled)
{
   BOOL   flag;
   
   NSCParameterAssert( [cell isKindOfClass:[NSCell class]]);
   
   flag = image || title;
   if( [cell respondsToSelector:@selector( setTransparent:)])
      [(NSButtonCell *) cell setTransparent:! flag];
   
   [cell setImage:image];
   if( ! [cell isKindOfClass:[NSImageCell class]])
      [cell setTitle:title];
   [cell setEnabled:isEnabled && flag];
}


static void   setCellFromEO( NSCell *cell, id eo, NSString *imageKey, NSString *titleKey, BOOL isEnabled)
{
   NSImage    *image;
   NSString   *title;
   
   NSCParameterAssert( [cell isKindOfClass:[NSCell class]]);

   image = imageKey ? [eo valueForKey:imageKey] : nil;
   title = titleKey ? [eo valueForKey:titleKey] : nil;
   
   setCell( cell, image, title, isEnabled);
}


typedef struct
{
   NSMatrix       *matrix;
   unsigned int   nCells;
   unsigned int   columns;
   unsigned int   rows;
   SEL            selCellAtRow;
   IMP            impCellAtRow;
   
   NSArray        *displayedObjects;
   unsigned int   nDisplayedObjects;
   SEL            selObjectAtIndex;
   IMP            impObjectAtIndex;
   
   BOOL           resizeFlag;
} sync_context;


static void   changeMatrix( sync_context *ctxt, NSString *imageKey, NSString *titleKey, BOOL isEnabled)
{
   unsigned int   i, n;
   unsigned int   x, y;
   NSCell         *cell;
   id             eo;
   
   x = 0;
   y = 0;
   n = ctxt->nDisplayedObjects < ctxt->nCells ? ctxt->nDisplayedObjects : ctxt->nCells;
   
   for( i = 0; i < n; i++)
   {
      if( x >= ctxt->columns)
      {
         x = 0;
         y++;
      }
      
      cell = (*ctxt->impCellAtRow)( ctxt->matrix, ctxt->selCellAtRow, y, x);
      eo   = (*ctxt->impObjectAtIndex)( ctxt->displayedObjects, ctxt->selObjectAtIndex, i);
      
      setCellFromEO( cell, eo, imageKey, titleKey, isEnabled);
      ++x;
   }
}


static void   disableColumnTail( sync_context *ctxt, int xEnd, int y)
{
   NSCell   *cell;
   int      x;
   
   x = ctxt->columns - 1;
   for( ;x >= xEnd; x--)
   {      
      // disable cells and try to make them invisible
      cell = (*ctxt->impCellAtRow)( ctxt->matrix, ctxt->selCellAtRow, y, x);
      setCell( cell, nil, nil, NO);
   }
}


static BOOL   shrinkMatrixTail( sync_context *ctxt)
{
   int            x, y;
   int            xEnd, yEnd;
   BOOL           flag;
   
   flag = NO;
   
   y = ctxt->rows    - 1;
   x = ctxt->columns - 1;

   yEnd = ctxt->nDisplayedObjects / ctxt->columns;
   xEnd = ctxt->nDisplayedObjects % ctxt->columns;

   //
   // if we have unfilled rows then keep last column 
   // alive
   //
   if( xEnd)
      yEnd++;
   
   // remove superflous rows
   for(; y >= yEnd; y--)
   {
      if( y)  // we keep the last row around (so we don't lose columns info)
      {
         [ctxt->matrix removeRow:y];
         ctxt->nCells -= ctxt->columns;
         ctxt->rows--;
         flag = YES;
      }
   }

   if( xEnd || ! y)
      disableColumnTail( ctxt, xEnd, y);
   
   return( flag);
}


static BOOL   growMatrixTail( sync_context *ctxt, NSString *imageKey, NSString *titleKey, BOOL isEnabled)
{
   unsigned int   i, n;
   unsigned int   x, y;
   id             eo;
   NSCell         *cell;
   BOOL           flag;
   
   flag = NO;
   
   y = ctxt->rows    - 1;
   x = ctxt->columns - 1;
   n = ctxt->nDisplayedObjects;

   for( i = ctxt->nCells; i < n; i++)
   {
      if( x++ >= ctxt->columns)
      {
         x = 0;
         y++;
         
         [ctxt->matrix addRow];
         ctxt->rows++;
         ctxt->nCells += ctxt->columns;
         
         if( ctxt->nCells > n)
            n = ctxt->nCells;
         
         flag = YES;
      }
      
      eo   = nil;
      cell = (*ctxt->impCellAtRow)( ctxt->matrix, ctxt->selCellAtRow, y, x);
      if( i < ctxt->nDisplayedObjects)
         eo = (*ctxt->impObjectAtIndex)( ctxt->displayedObjects, ctxt->selObjectAtIndex, i);
      
      setCellFromEO( cell, eo, imageKey, titleKey, isEnabled);
   }
   
   return( flag);
}


static BOOL   syncContents( NSMatrix *matrix, NSArray *array, NSString *imageKey, NSString *titleKey, BOOL isEnabled)
{
   sync_context     ctxt;
   BOOL             flag;
   
   flag = NO;
   
   // reload matrix
   ctxt.matrix       = matrix;
   ctxt.rows         = [matrix numberOfRows];
   ctxt.columns      = [matrix numberOfColumns];
   ctxt.nCells       = ctxt.rows * ctxt.columns;
   ctxt.selCellAtRow = @selector( cellAtRow:column:);
   ctxt.impCellAtRow = [matrix methodForSelector:ctxt.selCellAtRow];
   
   ctxt.displayedObjects  = array;
   ctxt.nDisplayedObjects = [array count];   
   ctxt.selObjectAtIndex  = @selector( objectAtIndex:);
   ctxt.impObjectAtIndex  = [array methodForSelector:ctxt.selObjectAtIndex];
   
   changeMatrix( &ctxt, imageKey, titleKey, isEnabled);
   //
   // more old cells than new cells ?
   // remove rows as much as can do
   //
   if( ctxt.nDisplayedObjects < ctxt.nCells)
      flag = shrinkMatrixTail( &ctxt);
   else
      if( ctxt.nCells < ctxt.nDisplayedObjects)
         flag = growMatrixTail( &ctxt, imageKey, titleKey, isEnabled);
   
   return( flag);
}


static BOOL   syncSelection( NSMatrix *matrix, NSArray *array)
{
   unsigned int     i, n;
   unsigned int     x, y;
   unsigned int     columns;
   IMP              impSelectCellAtRow;
   IMP              impObjectAtIndex;
   id               value;
   int              index;
   
   [matrix deselectAllCells];
   
   impSelectCellAtRow = [matrix methodForSelector:@selector( selectCellAtRow:column:)];
   impObjectAtIndex   = [array methodForSelector:@selector( objectAtIndex:)];
   
   columns = [matrix numberOfColumns];
   n       = [array count];
   for( i = 0; i < n; i++)
   {
      value = (*impObjectAtIndex)( array, @selector( objectAtIndex:), i);
      index = [value intValue];
      
      y = index / columns;
      x = index % columns;
      (*impSelectCellAtRow)( matrix, @selector( selectCellAtRow:column:), y, x);
   }
   return( YES);
}


//
// we manage this by noting that the matrix is just ordered like our
// displayedObjects. In the PopUp box we have some lossage because of
// duplicate titles. But here the relationship is 1:1
//
- (void) subjectChanged
{
   EODisplayGroup   *group;
   NSMatrix         *matrix;
   NSString         *titleKey;
   NSString         *imageKey;
   BOOL             isEnabled;
   BOOL             redrawFlag;
   BOOL             resizeFlag;
   EOAssociationObjectState   state;
   
   matrix = [self object];
   
   group    = nil;
   imageKey = [self displayGroupKeyForAspect:MulleEOInterfaceImageAspect];
   titleKey = [self displayGroupKeyForAspect:MulleEOInterfaceTitleAspect];
   if( imageKey)
      group = [self displayGroupForAspect:MulleEOInterfaceImageAspect];
   else
      if( titleKey)
         group = [self displayGroupForAspect:MulleEOInterfaceTitleAspect];
   
   state = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
      isEnabled = state;
   else
      isEnabled = [matrix isEnabled];
         
   redrawFlag = NO;
   resizeFlag = NO;
   
   if( [group contentsChanged])
   {
      redrawFlag = YES;
      resizeFlag = syncContents( matrix, [group displayedObjects], imageKey, titleKey, isEnabled);
   }
      
   if( [group selectionChanged])
      redrawFlag = syncSelection( matrix, [group selectionIndexes]);
   
   if( resizeFlag)
   {
      [matrix sizeToCells];
      [[matrix superview] setNeedsDisplay:YES];
   }
   
   if( [matrix isEnabled] != isEnabled)
      [matrix setEnabled:isEnabled];
   if( redrawFlag)
      [matrix setNeedsDisplay:YES];
}


- (void) objectChanged
{
#if 0   
   EODisplayGroup   *group;
   NSMatrix         *matrix;
   NSCell           *cell;
   NSString         *title;
   NSString         *key;
   int              tag;
   NSArray          *cells;
   
   matrix = [self object];
   cells  = [matrix selectedCells];
   
   key = [self displayGroupKeyForAspect:MulleEOInterfaceImageAspect];
   if( key)
   {
      group = [self displayGroupForAspect:MulleEOInterfaceImageAspect];
      image = [group selectedObjectValueForKey:key];
      [cell = 
   }
   
   key   = [self displayGroupKeyForAspect:MulleEOInterfaceTitleAspect];
   if( key)
   {
      group = [self displayGroupForAspect:MulleEOInterfaceSelectedTagAspect];
      tag   = [cell tag];
      [group setSelectedObjectValue:[NSNumber numberWithInt:tag]
                             forKey:key];
   }
#endif
}   


- (void) action:(id) sender
{
   [self objectChanged];
}

@end


