/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOOutlineViewAssociation+DragNDrop.h"

#import "EODisplayGroup.h"


NSString   *EOOutlineViewAssociationDragNDrop         = @"EOOutlineViewAssociationDragNDrop";
NSString   *EOOutlineViewAssociationInternalDragNDrop = @"EOOutlineViewAssociationInternalDragNDrop";


@implementation EOOutlineViewAssociationDragNDropInfo : EOTableViewAssociationDragNDropInfo

- (int)    childIndex { return( childIndex_); }
- (id)     item       { return( item_); }
- (BOOL)   isLocal    { return( isLocal_); }

- (void) setChildIndex:(int) index
{
   childIndex_ = index;
}


- (void) setItem:(id) item
{
   item_ = [[item retain] autorelease];
}


- (unsigned int) draggedItemsCount
{
   NSData   *data;

   if( ! [types_ containsObject:EOOutlineViewAssociationInternalDragNDrop])
      return( nil);
   data = [pasteboard_ dataForType:EOOutlineViewAssociationInternalDragNDrop];
   return( [items_ count]);
}


- (NSArray *) draggedItems
{
   NSData   *data;

   if( ! [types_ containsObject:EOOutlineViewAssociationInternalDragNDrop])
      return( nil);
   data = [pasteboard_ dataForType:EOOutlineViewAssociationInternalDragNDrop];
   return( items_);
}

@end


@implementation EOOutlineViewAssociation ( DragNDrop)

- (id) standardDragNDropInfo
{
   EOOutlineViewAssociationDragNDropInfo  *info;
   
   info = [[[EOOutlineViewAssociationDragNDropInfo alloc] init] autorelease];
   info->association_  = self;
   info->displayGroup_ = sourceGroup_; 
   return( info);
}


static EOOutlineViewAssociationDragNDropInfo  *notify( EOOutlineViewAssociation *self,
                                                       EOOutlineViewAssociationDragNDropInfo *info)
{
   NSDictionary   *userInfo;

   [self setDragNDropInfo:nil];

   userInfo = [NSDictionary dictionaryWithObject:info
                                          forKey:EODragNDropInfo];
   [[NSNotificationCenter defaultCenter] postNotificationName:EOOutlineViewAssociationDragNDrop
                                                       object:self->_object
                                                     userInfo:userInfo];
   info = (id) [self dragNDropInfo];
   [self setDragNDropInfo:nil];
   return( info);
}


- (void) registerForDragNDrops
{
   EOOutlineViewAssociationDragNDropInfo  *info;
   
   info = [self standardDragNDropInfo];
   info->stage_ = EODragNDropRegistration;
   info->types_ = [NSArray arrayWithObjects:EOOutlineViewAssociationInternalDragNDrop, nil];

   info = notify( self, info);
   if( info)
      [_object registerForDraggedTypes:info->types_];
}


- (unsigned int) outlineView:(NSOutlineView *) view
draggingSourceOperationMaskForLocal:(BOOL) isLocal
{
   EOOutlineViewAssociationDragNDropInfo  *info;

   info = [self standardDragNDropInfo];
   info->stage_         = EODragNDropOperation;
   info->dragOperation_ = isLocal ? NSDragOperationGeneric : NSDragOperationNone;
   info->isLocal_       = isLocal;

   info = notify( self, info);
   if( info)
      return( info->dragOperation_);
   return( NSDragOperationNone);
}


//
// Put items on paste board, interested parties may intercept
// this by the notification and call setDragNDropInfo: on us
// if the dictionary is nil, this means that the receiver already
// set things up for us, so we return with YES.
// id the types array is now empty, we return NO
//
- (BOOL) outlineView:(NSOutlineView  *) outlineView
          writeItems:(NSArray *) items
        toPasteboard:(NSPasteboard *) pasteboard
{
   EOOutlineViewAssociationDragNDropInfo  *info;

   [self setDragNDropItems:items]; 
   
   info = [self standardDragNDropInfo];
   info->stage_      = EODragNDropPreparation;
   info->types_      = [NSArray arrayWithObjects:EOOutlineViewAssociationInternalDragNDrop, nil];
   info->pasteboard_ = pasteboard;
   info->items_      = dragNDropItems_;
   
   info = notify( self, info);
   
   if( ! info)
      return( YES);
   
   if( ! [info->types_ count])
      return( NO);

   [info->pasteboard_ declareTypes:info->types_
                             owner:self];
   if( [info->types_ containsObject:EOOutlineViewAssociationInternalDragNDrop])
      [info->pasteboard_ setData:[NSData data] 
                         forType:EOOutlineViewAssociationInternalDragNDrop];
   return( YES);
}



- (unsigned int) outlineView:(NSOutlineView *) outlineView
                validateDrop:(id <NSDraggingInfo>) dragInfo
                proposedItem:(id) item
          proposedChildIndex:(int) childIndex
{
   EOOutlineViewAssociationDragNDropInfo  *info;

   info = [self standardDragNDropInfo];
   info->stage_         = EODragNDropValidation;
   info->dragOperation_ = NSDragOperationGeneric;
   info->draggingInfo_  = dragInfo;
   info->pasteboard_    = [info->draggingInfo_ draggingPasteboard];
   info->types_         = [info->pasteboard_ types];
   info->item_          = item;
   info->childIndex_    = childIndex;
   info->items_         = dragNDropItems_;

   info = notify( self, info);

   if( ! info)
      return( NSDragOperationNone);

#if DEBUG_VERBOSE
   NSLog( @"%d %@", [info childIndex], [info item]);
#endif   
   [_object setDropItem:info->item_
         dropChildIndex:info->childIndex_];

   return( info->dragOperation_);
}


- (BOOL) outlineView:(NSOutlineView *) outlineView
          acceptDrop:(id <NSDraggingInfo>) dragInfo
                item:(id) item
          childIndex:(int) childIndex
{
   EOOutlineViewAssociationDragNDropInfo  *info;

   info = [self standardDragNDropInfo];
   info->stage_         = EODragNDropAcceptance;
   info->dragOperation_ = NSDragOperationGeneric;
   info->draggingInfo_  = dragInfo;
   info->pasteboard_    = [info->draggingInfo_ draggingPasteboard];
   info->types_         = [info->pasteboard_ types];
   info->item_          = item;
   info->childIndex_    = childIndex;
   info->items_         = dragNDropItems_;

   info = notify( self, info);
   
   if( ! info)
      return( NO);
   return( YES);
}

@end
