/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 *
 */
#import "Defines.h"

#import "EOOutlineViewAssociation.h"

#import <AppKit/AppKit.h>
#import "AspectsAndKeys.h"
#import "EODisplayGroup.h"
#import "EODisplayGroup+Objects.h"
#import "EODisplayGroup+Selection.h"
#import "EOColumnAssociation.h"
#import "EOControlAssociation.h"
#import "NSObject+TableViewAssociation.h"
#import <objc/objc-class.h>


@implementation EOOutlineViewAssociation

static inline eo_tableview_association_extra   *tableviewAssociationExtra( EOAssociation *self)
{
   return( (eo_tableview_association_extra *) &((char *) self)[ ((struct objc_class *) self->isa)->instance_size]);
}


+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[NSOutlineView class]]);
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceSourceAspect, MulleEOInterfaceChildrenAspect, MulleEOInterfaceParentAspect, MulleEOInterfaceEnabledAspect,
         MulleEOInterfaceTextColorAspect, MulleEOInterfaceBoldAspect, MulleEOInterfaceItalicAspect, nil];
   return( aspects);
}


+ (NSArray *)  aspectSignatures
{
   static NSArray  * aspectSignatures;

   if( !  aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"", @"M", @"1", @"A1M", @"A", @"A", @"A", nil];

   return(  aspectSignatures);
}


+ (NSArray *) associationClassesSuperseded
{
   static NSArray  *associationClassesSuperseded;

   if( ! associationClassesSuperseded)
      associationClassesSuperseded = [[NSArray alloc] initWithObjects:[EOControlAssociation class], [EOTableViewAssociation class], nil];
   return( associationClassesSuperseded);
}


/*#
### The smart stuff
#*/
- (void) releaseCachedValues;
{
    [rootChildren_ release];
    rootChildren_ = nil;
    nRootChildren_ = 0;
}


- (void) dealloc
{
    [self releaseCachedValues];
    [super dealloc];
}


- (void) setDragNDropItems:(NSArray *) items
{
   id   old;

   old = dragNDropItems_;
   dragNDropItems_ = [[NSMutableArray alloc] initWithArray:items];
   [old release];
}


//
//  CONNECTING
//
// Es ist nicht so, dass wir immer eine breakConnection kriegen
// zwischen zwei establishConnection!!
//
- (void) establishConnection;
{
   childrenKey_ = [self displayGroupKeyForAspect:MulleEOInterfaceChildrenAspect];
   if( ! childrenKey_)
   {
      NSLog( @"%@ %@ aspect not connected", self, MulleEOInterfaceChildrenAspect);
      return;
   }
   dadKey_ = [self displayGroupKeyForAspect:MulleEOInterfaceParentAspect];
   if( ! dadKey_)
   {
      NSLog( @"%@ %@ aspect not connected", self, MulleEOInterfaceParentAspect);
      return;
   }

   [super establishConnection];
   
   sourceGroup_ = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
}


- (void) breakConnection;
{
   id   obj;
   
   obj = [self object];
   
   [self setDragNDropItems:nil];
   [_object setDataSource:nil];
   [_object setDelegate:nil];

   sourceGroup_ = nil;
   childrenKey_ = nil;
   dadKey_      = nil;
   [self releaseCachedValues];

   [super breakConnection];
}


//
// letzten nicht expandieren
//
- (void) expandItemPath:(id) item
{
   NSMutableArray   *path;
   NSEnumerator     *rover;
   unsigned int     recursionProtection;
      
#if DEBUG_VERBOSE
   NSLog( @"%s %@ %@", __PRETTY_FUNCTION__, self, item);
#endif   
   path = [NSMutableArray array];
   recursionProtection = 64;
   while( item = [item valueForKey:dadKey_])
   {
      [path addObject:item];
      if( ! --recursionProtection)
      {
         NSLog( @"%s: item %@ traversal on %@ appeared to recurse forever", item, dadKey_);
         return;
      }
   }

   rover = [path reverseObjectEnumerator];
   while( item = [rover nextObject])
   {
      if( [_object isItemExpanded:item] == NO)
          [_object expandItem:item];
   }
}


//
// OutlineView Protokoll
//
- (id) outlineView:(NSOutlineView *) view
             child:(int) index
            ofItem:(id) item
{
#if DEBUG_VERBOSE
   NSLog( @"%s %@ %d %@ -> %@", __PRETTY_FUNCTION__, self,
         index, item,
          item ? [[item valueForKey:childrenKey_] objectAtIndex:index]
               : [rootChildren_ objectAtIndex:index]);
#endif
   return( item ? [[item valueForKey:childrenKey_] objectAtIndex:index]
                : [rootChildren_ objectAtIndex:index]);
}


- (BOOL) outlineView:(NSOutlineView *) view
    isItemExpandable:(id) item
{
#if DEBUG_VERBOSE
   NSLog( @"%s (%@) %@ %@ %d", __PRETTY_FUNCTION__, self,
          item, childrenKey_, [[item valueForKey:childrenKey_] count]);
#endif

    return( [[item valueForKey:childrenKey_] count] > 0 ? YES : NO);
}


- (int)           outlineView:(NSOutlineView *) view
       numberOfChildrenOfItem:(id) item
{
#if DEBUG_VERBOSE
   NSLog( @"%s %@ %@ -> %d", __PRETTY_FUNCTION__, self,
          item,
          item ? [[item valueForKey:childrenKey_] count]
                           : nRootChildren_);
#endif
    return( item ? [[item valueForKey:childrenKey_] count]
                 : nRootChildren_);
}


static inline EOColumnAssociation   *columnAssociationForIdentifier( NSString *identifier)
{
   if( [identifier isKindOfClass:[EOColumnAssociation class]])
      return( (EOColumnAssociation *) identifier);
   return( nil);
}




- (id) 	       outlineView:(NSOutlineView *) outlineView
 objectValueForTableColumn:(NSTableColumn *) tableColumn
                    byItem:(id) item
{
   EOColumnAssociation  *assoc;
   id   value;
   id   identifier;


   identifier = [tableColumn identifier];
   if( assoc = columnAssociationForIdentifier( identifier))
   {
      return( [assoc outlineView:outlineView
       objectValueForTableColumn:tableColumn
                          byItem:item]);
   }

   value = nil;
   if( identifier)
      value = [item valueForKeyPath:identifier];
   return( value ? value : @"???");
}


//
// not time criticial
//
- (void) outlineView:(NSOutlineView *) outlineView
      setObjectValue:(id) value
      forTableColumn:(NSTableColumn *) tableColumn
              byItem:(id) item
{
   EOColumnAssociation  *assoc;
   NSString             *identifier;

   identifier = [tableColumn identifier];
   if( assoc = columnAssociationForIdentifier( identifier))
   {
      [assoc outlineView:outlineView
          setObjectValue:value
          forTableColumn:tableColumn
                  byItem:item];
      return;
   }

   if( [identifier length])
   {
      //
      // don't use the EODisplayGroup, because
      // we may set values on objects, it doesn't
      // know about (deep in the tree)
      //
NS_DURING
      [item takeValue:value
           forKeyPath:identifier];
NS_HANDLER
      NSLog( @"%@: Setting of object value %@ for key %@ failed for EO %@ ",
             self, value, identifier, item);
NS_ENDHANDLER
   }
}


//
// Collapsing wenn ein Kind oder kindes kind davon selektiert
// ist, ist "nicht cool", also diese deselektieren
//
- (BOOL)      outlineView:(NSOutlineView *) view
       shouldCollapseItem:(id) item
{
   NSMutableArray  *array;
   NSArray         *selection;
   NSEnumerator    *rover;
   id              p;

   selection = [sourceGroup_ selectedObjects];
   array     = [NSMutableArray arrayWithArray:selection];
   
   rover = [[sourceGroup_ selectedObjects] objectEnumerator];
   while( p = [rover nextObject])
      while( p = [p valueForKey:dadKey_])
         if( p == item)
         {
            [array removeObjectIdenticalTo:p];
            break;
         }
            
  if( ! [selection isEqualToArray:array])
     [sourceGroup_ selectObjectsIdenticalTo:array];
  return( YES);
}


//
// Keep our displayGroup in sync....
//
- (void) outlineViewSelectionDidChange:(NSNotification *) notification
{
   NSMutableArray   *selection;
   NSEnumerator     *rover;
   id               item;
   NSNumber         *number;
   
   rover     = [_object selectedRowEnumerator];
   selection = [NSMutableArray array];

   while( number = [rover nextObject])
   {
      item  = [_object itemAtRow:[number intValue]];
      [selection addObject:item];
   }

   [sourceGroup_ selectObjectsIdenticalTo:selection];
}


//
// Eleganter waere es noch die momentan aktuelle
// Row festzustellen und nur zu scrollen, wenn es
// unbedingt sein muss
//
- (void) displayGroupDidChangeSelection
{
   NSMutableArray  *selectedObjects;
   NSMutableArray  *deselectCandidates;
   NSEnumerator	   *rover;
   id              item;
   int             firstRow;
   int             row;
   BOOL            flag;
   NSNumber        *number;
   unsigned int    index;
   
   firstRow = -1;

   selectedObjects = [NSMutableArray arrayWithArray:[sourceGroup_ selectedObjects]];
   if( [selectedObjects count] == 0)
   {
      [_object deselectAll:self];
      return;
   }
   
   deselectCandidates = [NSMutableArray array];

   flag = NO;
   rover = [_object selectedRowEnumerator];
   while( number = [rover nextObject])
   {
      row   = [number intValue];
      item  = [_object itemAtRow:row];
      index = [selectedObjects indexOfObjectIdenticalTo:item];
      if( index != NSNotFound)
      {
         [selectedObjects removeObjectAtIndex:index];
         flag = YES;
         if( firstRow < 0)
            firstRow = row;
      }
      else
         [deselectCandidates addObject:item];
   }

   rover = [selectedObjects objectEnumerator];
   while( item = [rover nextObject])
   {
      row = [_object rowForItem:item];
      [_object selectRow:row
    byExtendingSelection:flag];
      flag = YES;
      
      if( firstRow < 0)
         firstRow = row;
   }

   rover = [deselectCandidates objectEnumerator];
   while( item = [rover nextObject])
      [_object deselectRow:[_object rowForItem:item]];

   if( firstRow >= 0)
      [_object scrollRowToVisible:firstRow];
}


- (void) sync
{
   NSMutableArray   *newRootChildren;
   NSEnumerator	    *enumerator;
   NSArray          *possibleRootChildren;
   id               possibleRootChild;

   [self setDragNDropItems:nil];	// get rid of them
   possibleRootChildren = [sourceGroup_ displayedObjects];

   if( [possibleRootChildren count] == 0)
   {
      [rootChildren_ release];
      rootChildren_  = 0;
      nRootChildren_ = 0;
      return;
   }

   newRootChildren = [NSMutableArray array];
   enumerator = [possibleRootChildren objectEnumerator];
   while( possibleRootChild = [enumerator nextObject])
   {
      if( [possibleRootChild valueForKey:dadKey_] == nil)
         [newRootChildren addObject:possibleRootChild];
   }
   
#warning (nat) I don''t get it, what happens if we have none ??
   
   if( [newRootChildren count])
   {
      [rootChildren_ release];
      [newRootChildren retain];
      rootChildren_ = newRootChildren;
   }

   nRootChildren_ = [rootChildren_ count];
}


- (void) subjectChanged
{
#if DEBUG_VERBOSE
   NSLog( @"%s (%@)", __PRETTY_FUNCTION__, self);
#endif   

   if( [sourceGroup_ contentsChanged])
   {
      [self sync];
      [_object reloadData];
   }

   if( [sourceGroup_ selectionChanged])
   {
      [self displayGroupDidChangeSelection];
   }
}



@end
