/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOPickTextAssociation.h"

#import <AppKit/AppKit.h>
#import "AspectsAndKeys.h"
#import "EODisplayGroup.h"
#import "EODisplayGroup+Objects.h"


@implementation EOPickTextAssociation


+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[NSControl class]] ||
           [object isKindOfClass:[NSActionCell class]]);
}


+ (NSString *) primaryAspect
{
   return( MulleEOInterfaceMatchKey1Aspect);  // or ??
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceMatchKey1Aspect, MulleEOInterfaceMatchKey2Aspect, MulleEOInterfaceMatchKey3Aspect, nil];
   return( aspects);
}


+ (NSArray *) aspectSignatures
{
   static NSArray  *aspectSignatures;
   
   if( ! aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"A", @"A", @"A", nil];
   
   return( aspectSignatures);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:MulleEOInterfaceTargetKey, MulleEOInterfaceDelegateKey, nil];
   return( objectKeysTaken);
}


- (void) establishConnection
{
   EOAssociationObjectState   state;
   EODisplayGroup             *group;
   EODisplayGroup             *other;
   EODisplayGroup             *last;

   group = [self displayGroupForAspect:MulleEOInterfaceMatchKey1Aspect];
   if( ! group)
   {
      NSLog( @"%@ matchKey1 aspect not connected", self);
      return;
   }

   other = [self displayGroupForAspect:MulleEOInterfaceMatchKey2Aspect];
   if( other && other != group)
   {
      NSLog( @"%@ matchKey2 aspect connected to other EODisplayGroup than matchKey1 aspect", self);
      return;
   }

   last = [self displayGroupForAspect:MulleEOInterfaceMatchKey3Aspect];
   if( last)
   {
      if( ! other)
      {
         NSLog( @"%@ matchKey3 aspect connected but not matchKey2 aspect", self);
         return;
      }
      if( last != group)
      {
         NSLog( @"%@ matchKey3 aspect connected to other EODisplayGroup than matchKey1 aspect", self);
         return;
      }
   }
   
   [_object setTarget:self];
   [_object setAction:@selector( action:)];
   
   state = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
      [_object setEnabled:(BOOL) state];

   if( [_object respondsToSelector:@selector( setDelegate:)])
      [_object setDelegate:self];

   [super establishConnection];
}


- (void) breakConnection
{
   id   obj;
   
   obj = [self object];
   if( [obj respondsToSelector:@selector( setDelegate:)])
      [obj setDelegate:nil];
   [obj setTarget:nil];
   [obj setAction:0];

   [super breakConnection];
}


- (void) objectChanged
{
   id               value;
   NSString         *key[ 3];
   EODisplayGroup   *group;
   EOQualifier      *q;
   NSMutableArray   *qualifiers;
   unsigned int     i;

   value      = [_object objectValue];
   qualifiers = [NSMutableArray array];
   group      = [self displayGroupForAspect:MulleEOInterfaceMatchKey1Aspect];

   value = [value description];
   if( [(NSString *) value length])
   {
      SEL   op;
      
      key[ 0] = [self displayGroupKeyForAspect:MulleEOInterfaceMatchKey1Aspect];
      key[ 1] = [self displayGroupKeyForAspect:MulleEOInterfaceMatchKey2Aspect];
      key[ 2] = [self displayGroupKeyForAspect:MulleEOInterfaceMatchKey3Aspect];

      op = [EOQualifier operatorSelectorForString:[group defaultStringMatchOperator]];
      if( ! op)
         op = EOQualifierOperatorCaseInsensitiveLike;

      for( i = 0; i < 3; i++)
      {
         if( key[ i])
         {
            q  = [[EOKeyValueQualifier alloc] initWithKey:key[ i]
                                         operatorSelector:op
                                                    value:[NSString stringWithFormat:@"*%@*", value]];
            [qualifiers addObject:q];
            [q release];
         }
      }
   }

   switch( [qualifiers count])
   {
   case 0 :
      q = nil;
      break;

   case 1 :
      q = [qualifiers lastObject];
      break;

   default :
      q = [[[EOOrQualifier alloc] initWithQualifierArray:qualifiers] autorelease];
   }
#if DEBUG_VERBOSE
   NSLog( @"Qualifying %@ with %@", group, q);
#endif   
   [group setQualifier:q];
   [group updateDisplayedObjects];
}   


- (void) subjectChanged
{
   BOOL                       oldstate;
   EOAssociationObjectState   state;
   
#if DEBUG_VERBOSE
   NSLog( @"%s (%@)", __PRETTY_FUNCTION__, self);
#endif   
   state = [self objectStateForEnabledAspect];
   if( state == EOAssociationObjectStateKeepState)
      return;
   
   oldstate = [_object isEnabled];
   if( state != oldstate)
      [_object setEnabled:(BOOL) state];
}


- (void) controlTextDidChange:(id) sender
{
//   if( _applyImmediately)
      [self objectChanged];
}


- (void) action:(id) sender
{
#if DEBUG_VERBOSE
   NSLog( @"%s", __PRETTY_FUNCTION__);
#endif   
   [self objectChanged];
}


@end


