/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOPopUpAssociation.h"

#import <AppKit/AppKit.h>
#import "AspectsAndKeys.h"
#import "EOAssociation+PopUpComboSupport.h"
#import "_EOInterfaceUniqueObjectsAndTags.h"
#import "EOActionAssociation.h"
#import "EOControlAssociation.h"
#import "EODisplayGroup.h"
#import "EODisplayGroup+Selection.h"


#define FORCE_VALID_VALUE_INSTEAD_OF_DISABLE   0
#define DISABLE_IF_INVALID                     0
#define POPUP_DEBUG                            (defined( DEBUG) & 1)


@implementation EOPopUpAssociation


+ (void) initialize
{
   [self setVersion:1];
}


- (id) initWithCoder:(NSCoder *) decoder
{
   [super initWithCoder:decoder];

#if DEBUG_CODER
   NSLog( @"Continueing %@ %s 0x%lx (version=%d)",  isa, __PRETTY_FUNCTION__, (long) self, [decoder versionForClassName:@"EOPopUpAssociation"]);
#endif   

   [decoder decodeValueOfObjCType:@encode( int)
                               at:&_tagValueForOther];

#if DEBUG_CODER
   NSLog( @"Leave %@ %s 0x%lx", isa, __PRETTY_FUNCTION__, (long) self);
#endif   
  
   return( self);
}



- (void) encodeWithCoder:(NSCoder *) encoder
{
   [super encodeWithCoder:encoder];

#if DEBUG_CODER
   NSLog( @"Continueing %@ %s 0x%lx",  isa, __PRETTY_FUNCTION__, (long) self);
#endif   

   [encoder encodeValueOfObjCType:@encode( int)
                               at:&_tagValueForOther];
#if DEBUG_CODER
  NSLog( @"Leave %@ %s 0x%lx", isa, __PRETTY_FUNCTION__, (long) self);
#endif   
}


- (id) initWithObject:(id) object
{
   [super initWithObject:object];

   _tagValueForOther  = -1;
   _lastSelectedIndex = -2;
   _forceValidValue   = FORCE_VALID_VALUE_INSTEAD_OF_DISABLE;
   _disableIfInvalid  = DISABLE_IF_INVALID;
   
   titleIndexMapping_ = NSCreateMapTable( NSObjectMapKeyCallBacks,NSIntMapValueCallBacks, 128);
   
   return( self);
}


+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[NSPopUpButton class]]);
}


+ (NSString *) primaryAspect
{
   return( MulleEOInterfaceTitlesAspect);  // or ??
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceTitlesAspect, MulleEOInterfaceSelectedTitleAspect, MulleEOInterfaceSelectedTagAspect,
         MulleEOInterfaceSelectedObjectAspect, MulleEOInterfaceEnabledAspect, nil];
   return( aspects);
}


+ (NSArray *) aspectSignatures
{
   static NSArray  *aspectSignatures;
   
#warning (nat) nur mal so zum testen....
   if( ! aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"A", @"A", @"A",
         @"1", @"A1M", nil];
   return( aspectSignatures);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:MulleEOInterfaceTargetKey, nil];
   return( objectKeysTaken);
}


+ (NSArray *) associationClassesSuperseded
{
   static NSArray  *associationClassesSuperseded;

   if( ! associationClassesSuperseded)
      associationClassesSuperseded = [[NSArray alloc] initWithObjects:[EOControlAssociation class],
         [EOActionAssociation class], nil];
   return( associationClassesSuperseded);
}


- (void) dealloc
{
   NSFreeMapTable( titleIndexMapping_);
   [_tempTitle release];
   [super dealloc];
}


- (EODisplayGroup *) targetGroup
{
   EODisplayGroup  *group;
   
   group = nil;
   if( _selectedObjectConnected)
      group = [self displayGroupForAspect:MulleEOInterfaceSelectedObjectAspect];
   if( _selectedTitleConnected)
      group = [self displayGroupForAspect:MulleEOInterfaceSelectedTitleAspect];
   if( _selectedTagConnected)
      group = [self displayGroupForAspect:MulleEOInterfaceSelectedTagAspect];
   return( group);
}

static void   syncEmptyTitlesBit( EOPopUpAssociation *self)
{
   self->_emptyTitles = ! [self->_object lastItem];
}


- (EOAssociationObjectState) objectStateForEnabledAspect
{
   return( [self objectStateForAspect:MulleEOInterfaceEnabledAspect
                        editingDisplayGroup:[self targetGroup]]);
}


- (void) establishConnection
{
   EOAssociationObjectState   state;
   id                         obj;
   
   if( _isConnected)
   {
      NSLog( @"%s: connection already established for %@", __PRETTY_FUNCTION__, self);
      return;
   }

   obj = [self object];

   _selectedTagConnected    = [self displayGroupForAspect:MulleEOInterfaceSelectedTagAspect] != nil;
   _selectedObjectConnected = [self displayGroupForAspect:MulleEOInterfaceSelectedObjectAspect] != nil;
   _selectedTitleConnected  = [self displayGroupForAspect:MulleEOInterfaceSelectedTitleAspect] != nil;
   _lastSelectedIndex       = -2;

   if( _selectedObjectConnected + _selectedTitleConnected + _selectedTagConnected > 1)
      NSLog( @"%@ : Too many aspects connected. Do not combine selectedObject, selectedTitle, selectedTag", self);

#if DEBUG_VERBOSE
   NSLog(@"connect popup: %d, %d, %d", _selectedTagConnected, _selectedObjectConnected, _selectedTitleConnected);
   NSLog(@"enabled: %d", [obj isEnabled]);
#endif   
   _wasEnabled = [obj isEnabled];

   state  = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
      [obj setEnabled:(BOOL) state];
   
   if( [self displayGroupForAspect:MulleEOInterfaceTitlesAspect])
      [obj removeAllItems];
   
   syncEmptyTitlesBit( self);
   
#if DEBUG_VERBOSE
   NSLog(@"enabled: %d", [obj isEnabled]);
#endif   
   [obj setTarget:self];
   [obj setAction:@selector( controlActed:)];  // nicely compatible...

   [super establishConnection];
}


- (void) breakConnection
{
   id   obj;
   
   obj = [self object];
   [obj setTarget:nil];
   [obj setAction:nil];

   _selectedTagConnected    = NO;
   _selectedObjectConnected = NO;
   _selectedTitleConnected  = NO;

   [obj setEnabled:_wasEnabled];

   [super breakConnection];
}


- (void) setTagValueForOther:(int) newValue
{
   _tagValueForOther = newValue;;
}


- (int) tagValueForOther
{
   return( _tagValueForOther);
}


- (void) reloadTitles
{
   _EOInterfaceUniqueObjectsAndTags   *titlesAndTags;
   NSArray         *filteredTitles;
   id <NSMenuItem> item;
   NSString        *title;
   unsigned int    count;
   unsigned int    i;
   unsigned int    tag;
   
#if DEBUG_VERBOSE
   NSLog( @"%s %@", __PRETTY_FUNCTION__, self);
#endif
   //
   // remove duplicates, keep index in tag
   //
   titlesAndTags = [self uniquedStringsForAspect:MulleEOInterfaceTitlesAspect];

   if( ! _disableSorting)
      [titlesAndTags sortUsingSelector:@selector( compare:)];
   
   //
   // Special nicety, allow null as selection (will be tag: NSNotFound)
   //
   if( _allowNullSelection)
      [titlesAndTags insertObject:@" -- "
                          withTag:NSNotFound
                          atIndex:0];
   filteredTitles = [titlesAndTags allObjects];
   
   ++ignore_;
   [_object removeAllItems]; 	// lame because it calls synchronizeTitleAndSelectedItem
   [_object addItemsWithTitles:filteredTitles];
   --ignore_;
   
   // why was this commented out ?
   _lastSelectedIndex = [_object indexOfSelectedItem]; // XXX DL right undefined value?

   [_tempTitle release];
   _tempTitle     = nil; 
   _addedTempItem = NO;

   //
   // use internal map tags to record index in original title list
   //
   NSResetMapTable( titleIndexMapping_);

   // we ask the popup and not titlesAndTags box, because the popup can't be trusted
   // to not have weeded out some items
   count = [_object numberOfItems];
   for( i = 0; i < count; i++)
   {
      item  = [_object itemAtIndex:i];
      title = [item title];
      tag   = [titlesAndTags tagForObject:title];
      
      NSMapInsertKnownAbsent( titleIndexMapping_, title, (void *) tag);
   }

   syncEmptyTitlesBit( self);
}


- (void) removeTempForNewTitle:(NSString *) title
{
   if( _addedTempItem)
   {
      if( ! [_tempTitle isEqualToString:title])
      {
         [_object removeItemWithTitle:_tempTitle];
         _addedTempItem = NO;
         syncEmptyTitlesBit( self);

         [_tempTitle release];
         _tempTitle = nil;
      }
   }
}


static void  selectPopUpItemAtIndex(EOPopUpAssociation *self, int index)
{
   if( self->_emptyTitles && index == 0)
      index = -1;
   
   if( self->_lastSelectedIndex != index)
   {
      [self->_object selectItemAtIndex:index];
      self->_lastSelectedIndex = index;
   }
}


- (void) selectItemWithTitle:(NSString *) title
                addIfMissing:(BOOL) flag
{
   int   index;

   [self removeTempForNewTitle:title];
   _lastSelectedIndex = [_object indexOfSelectedItem];
   
   index = -1;
   if( title)
   {
      index = [_object indexOfItemWithTitle:title];
      if( index == -1 && flag)
      {
         _tempTitle = [title copyWithZone:[self zone]];
         [_object addItemWithTitle:title];
         syncEmptyTitlesBit( self);
         _addedTempItem = YES;
         index = [_object indexOfItemWithTitle:title];
      }
   }
   else
      if( _allowNullSelection)
         index = [_object indexOfItemWithTag:NSNotFound];
   
   selectPopUpItemAtIndex( self, index);
}


- (void) selectItemWithTag:(int) tag
{
   int   index;

   [self removeTempForNewTitle:nil];

   index = [_object indexOfItemWithTag:tag];
   if( index == -1)
      index = _tagValueForOther;
   
   selectPopUpItemAtIndex( self, index);
}


- (void) selectItemAtIndex:(int) index
{
   [self removeTempForNewTitle:nil];
   selectPopUpItemAtIndex( self, index);
}


static BOOL  syncNoSelection( EOPopUpAssociation *self, EODisplayGroup *group)
{
   if( ! self->_allowNullSelection)
   {
#if DEBUG      
      NSLog( @"%s %@: null value (DG: %@) unusable and not allowed.", __PRETTY_FUNCTION__, self, group);
#endif      
      if( self->_forceValidValue)
      {
         [self selectItemAtIndex:0];
         if( self->_lastSelectedIndex == 0)
         {
            [self _updateWithIndexOfSelectedItem:self->_lastSelectedIndex];  // and mirror it back
            return( YES);
         }
      }
      [self selectItemAtIndex:-1];
      return( NO);
   }
   
   [self selectItemAtIndex:0];
   return( YES);
}


//
// popup indexes with titles into a displaygroup 
// (title strings are taken from key)
// the selected object in the "titles" aspect dg 
// is used as the value for the selected object
// in the "selectedObject" aspect dg
static BOOL  syncSelectionWithObjectOfGroup( EOPopUpAssociation *self, EODisplayGroup *group)
{
   NSString   *title;
   NSString   *key;
   id         eo;
   BOOL       enabled;
   
   key     = [self displayGroupKeyForAspect:MulleEOInterfaceSelectedObjectAspect];
   enabled = [[group selectedObjects] count] == 1;
   eo      = [group selectedObject];
   if( key)
      eo = [eo valueForKeyPath:key];
   
   if( eo)
   {
      title = [self uniquedStringForObject:eo
                                    aspect:MulleEOInterfaceTitlesAspect];
      [self selectItemWithTitle:title
                   addIfMissing:YES];
      return( enabled);
   }
   
   return( syncNoSelection( self, group));
}


static BOOL  syncSelectionWithSelectedTitleOfGroup( EOPopUpAssociation *self, EODisplayGroup *group)
{
   NSString   *s;
   NSString   *key;
   
   key = [self displayGroupKeyForAspect:MulleEOInterfaceSelectedTitleAspect];
   s   = [[group selectedObject] valueForKeyPath:key];
   if( s)
   {
      s = [s description];
      [self selectItemWithTitle:s
                   addIfMissing:YES];
      return( YES);
   }

   return( syncNoSelection( self, group));
}


static BOOL  syncSelectionWithSelectedTagOfGroup( EOPopUpAssociation *self, EODisplayGroup *group)
{
   NSNumber   *number;
   NSString   *key;
   
   key    = [self displayGroupKeyForAspect:MulleEOInterfaceSelectedTagAspect];
   number = [[group selectedObject] valueForKeyPath:key];
   if( [number isKindOfClass:[NSNumber class]])
   {
      [self selectItemWithTag:[number intValue]];
      return( YES);
   }

   return( syncNoSelection( self, group));
}


- (void) subjectChanged
{
   EOAssociationObjectState  state;
   EODisplayGroup  *titlesGroup;
   EODisplayGroup  *targetGroup;
   EODisplayGroup  *group;
   BOOL            oldstate; 
   BOOL            flag;
   BOOL            invalid;

#if DEBUG_VERBOSE
   NSLog( @"%s (%@) %@", __PRETTY_FUNCTION__, self, [self displayGroupForAspect:MulleEOInterfaceTitlesAspect]);
#endif

   targetGroup = [self targetGroup]; 
   state       = [self objectStateForEnabledAspect];
   oldstate    = [_object isEnabled];

   //
   // suspend objectChanged execution during
   // this
   //
   ++ignore_;
   
   invalid = NO;
   flag    = NO;
   titlesGroup = [self displayGroupForAspect:MulleEOInterfaceTitlesAspect];
   if( titlesGroup)
   {
      if( [titlesGroup contentsChanged])
      {
         flag = YES;
         [self reloadTitles];
         _lastSelectedIndex = -2;  // set to undefined
      }
   }

   if( _selectedObjectConnected)
   {
      group = [self displayGroupForAspect:MulleEOInterfaceSelectedObjectAspect];
      if( _lastSelectedIndex == -2 || [group selectionOrContentsChanged])
      {
         flag    = YES;
         invalid = ! syncSelectionWithObjectOfGroup( self, group);
      }
   }
   else
      if( _selectedTitleConnected)
      {
         group = [self displayGroupForAspect:MulleEOInterfaceSelectedTitleAspect];
         if( _lastSelectedIndex == -2 || [group selectionOrContentsChanged])
         {
            flag    = YES;
            invalid = ! syncSelectionWithSelectedTitleOfGroup( self, group);
         }
      }
      else
         if( _selectedTagConnected)
         {
            group = [self displayGroupForAspect:MulleEOInterfaceSelectedTagAspect];
            if( _lastSelectedIndex == -2 || [group selectionOrContentsChanged])
            {
               flag    = YES;
               invalid = ! syncSelectionWithSelectedTagOfGroup( self, group);
            }
         }
         else
         {
            if( _lastSelectedIndex == -2 || [titlesGroup selectionChanged])
            {
               flag    = YES;
               invalid = ! syncSelectionWithObjectOfGroup( self, titlesGroup);
            }
         }

   if( state != EOAssociationObjectStateKeepState)
   {
      if( invalid && _disableIfInvalid)
         state = NSOffState;
   
      if( state != oldstate)
      {
         flag = YES;
         [_object setEnabled:(BOOL) state];
      }
   }
   
   if( flag)
      [_object setNeedsDisplay:YES];   
   
   --ignore_;
}


- (BOOL) __updateWithIndexOfSelectedItem:(int) index
{
   EODisplayGroup   *group;
   NSArray          *array;
   id <NSMenuItem>  item;
   NSString         *title;
   int              tag;
   id               eo;
   int              flag;
   
   flag  = NO;
   group = [self displayGroupForAspect:MulleEOInterfaceTitlesAspect];
   eo    = nil;
   item  = [_object itemAtIndex:index];
   title = [item title];
   tag   = [item tag];
   
   // if we have created titles, then the tags will be possibly supplied
   // by a subclass, so we use our own tags
   if( group)
   {
      tag = NSNotFound;
      if( title)
         tag = (int) NSMapGet( titleIndexMapping_, title);
   }
   
   if( tag == NSNotFound)
   {
      if( ! _allowNullSelection)
         return( NO);
   }
   
   if( _selectedObjectConnected)
   {
      if( item)
      {
         array = [self originatingObjectsOfUniquedStringsForAspect:MulleEOInterfaceTitlesAspect];
         eo    = nil;
         if( [array count] > tag)
            eo = [array objectAtIndex:tag];
         flag  = [self setValue:eo
                      forAspect:MulleEOInterfaceSelectedObjectAspect];
      }
      else
      {
         group = [self displayGroupForAspect:MulleEOInterfaceSelectedObjectAspect];
         [group clearSelection];
      }
      return( flag);
   }
   
   if( _selectedTitleConnected)
   {
      if( item)
         flag = [self setValue:[item title]
                     forAspect:MulleEOInterfaceSelectedTitleAspect];
      else
         flag = [self setValue:nil
                     forAspect:MulleEOInterfaceSelectedTitleAspect];
      return( flag);
   }
   
   if( _selectedTagConnected)
   {
      if( item)
         flag = [self setValue:[NSNumber numberWithInt:tag]
                     forAspect:MulleEOInterfaceSelectedTagAspect];
      else
         flag = [self setValue:[NSNumber numberWithInt:-1]
                     forAspect:MulleEOInterfaceSelectedTagAspect];
      return( flag);
   }
   
   if( item)
   {
      // operate on titles only...
      if( tag != NSNotFound)
      {
         array = [group displayedObjects];
         NSParameterAssert( (unsigned int) tag < [array count]);
         eo    = [array objectAtIndex:tag];
         [group selectObject:eo];
      }
      else
         [group clearSelection];
      return( flag);
   }

   // operate on titles only...
   group = [self displayGroupForAspect:MulleEOInterfaceTitlesAspect];
   [group clearSelection];

   return( flag);
}


- (BOOL) _updateWithIndexOfSelectedItem:(int) index
{
   BOOL   flag;

   flag = [self __updateWithIndexOfSelectedItem:index];
   _lastSelectedIndex = flag ? index : -2;
}


//
// named thusly, for Apple EOInterface compatibility
//
- (void) controlActed:(id) sender
{
   int    index;
   BOOL   flag;
   
   // Change was forced by ourselves ? Then ignore_
   if( ignore_)
      return;
   
   // if selection is same, no need to fire 
   index = [_object indexOfSelectedItem];
   if( index == _lastSelectedIndex)
      return;

   flag = [self _updateWithIndexOfSelectedItem:index];
   
   //
   // means setValue didn't work out
   // reload... still no clue how to handle the exception here
   //
   if( flag == NO)
   {
      _lastSelectedIndex == -2;  // current state is undefined (wrong)
      [self subjectChanged];
   }
}


//
// for backwards compatibility with previous MulleEOInterface
// but controlActed is wired to the PopUpButton
- (void) objectChanged:(id)sender
{
   [self controlActed:sender];
}


// added API
- (void) setAllowsNullSelection:(BOOL) flag
{
   _allowNullSelection = flag;
}


- (BOOL) allowsNullSelection
{
   return( _allowNullSelection);
}


static void   SafelyAddDisplayGroupKeyToSet( EOAssociation *self, 
                                             NSMutableSet *set, 
                                             NSString *aspect)
{
   NSString   *key;
   
   key = [self displayGroupKeyForAspect:aspect];
   if( key)
      [set addObject:key];
}


- (void) addBoundKeysForDisplayedObjectsToSet:(NSMutableSet *) set
{
   SafelyAddDisplayGroupKeyToSet( self, set, MulleEOInterfaceTitlesAspect);
}

@end


