/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:36  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EORadioMatrixAssociation.h"

#import <AppKit/AppKit.h>
#import "AspectsAndKeys.h"
#import "EODisplayGroup.h"


@implementation EORadioMatrixAssociation

+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[NSMatrix class]]);
}


+ (NSString *) primaryAspect
{
   return( @"selectedTitle");  // or ??
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceSelectedTitleAspect,
         MulleEOInterfaceSelectedTagAspect, MulleEOInterfaceEnabledAspect, nil];
   return( aspects);
}


+ (NSArray *) aspectSignatures
{
   static NSArray  *aspectSignatures;
   
   if( ! aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"A", @"A", @"A1M", nil];
   
   return( aspectSignatures);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:@"target", nil];
   return( objectKeysTaken);
}


- (void) setTagValueForOther:(int) newValue
{
   _tagValueForOther = newValue;
}


- (int) tagValueForOther
{
   return( _tagValueForOther);
}


- (id) initWithCoder:(NSCoder *) decoder
{
   [super initWithCoder:decoder];
#if DEBUG_CODER
   NSLog( @"Continueing %@ %s 0x%lx (version=%d)",  isa, __PRETTY_FUNCTION__, (long) self, [decoder versionForClassName:@"EORadioMatrixAssociation"]);
#endif   
   [decoder decodeValueOfObjCType:"i"
                               at:&_tagValueForOther];
      
   return( self);
}


- (void) encodeWithCoder:(NSCoder *) encoder
{
   [super encodeWithCoder:encoder];
   [encoder encodeValueOfObjCType:"i"
                                 at:&_tagValueForOther];
}


- (void) establishConnection
{
   EOAssociationObjectState   state;
   NSMatrix                   *matrix;
   
   matrix = [self object];
   [matrix setTarget:self];
   [matrix setAction:@selector( action:)];
   
   state = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
      [matrix setEnabled:(BOOL) state];

   [super establishConnection];

   NSParameterAssert( [self displayGroupKeyForAspect:MulleEOInterfaceSelectedTitleAspect] ||
                      [self displayGroupKeyForAspect:MulleEOInterfaceSelectedTagAspect]);
}


- (void) breakConnection
{
   NSMatrix   *matrix;

   matrix = [self object];
   
   [matrix setTarget:nil];
   [matrix setAction:0];

   [super breakConnection];
}


- (void) subjectChanged
{
   NSMatrix         *matrix;
   NSCell           *cell;
   NSString         *key;
   NSString         *title;
   NSString         *s;
   int              tag;
   unsigned int     rows;
   unsigned int     columns;
   unsigned int     x, y;
   IMP              impCellAtRow;
   EOAssociationObjectState   state;
   
   matrix = [self object];
   cell   = [matrix selectedCell];
   
   state = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
      [matrix setEnabled:(BOOL) state];
   
   key = [self displayGroupKeyForAspect:MulleEOInterfaceSelectedTagAspect];
   if( key)
   {
      tag = [[self valueForAspect:MulleEOInterfaceSelectedTagAspect] intValue];
   
      if( ! [matrix selectCellWithTag:tag])
         [matrix selectCellWithTag:_tagValueForOther];
      return;
   }
   
   
   title = [self valueForAspect:MulleEOInterfaceSelectedTitleAspect];
      
   // find cell with title
   impCellAtRow = [matrix methodForSelector:@selector( cellAtRow:column:)];
   
   rows    = [matrix numberOfRows];
   columns = [matrix numberOfColumns];
   for( y = 0; y < rows; y++)
      for( x = 0; x < columns; x++)
      {
         cell = (*impCellAtRow)( matrix, @selector( cellAtRow:column:), y, x);
         s    = [cell title];
         if( s == title || [title isEqualToString:s])
         {
            [matrix selectCell:cell];
            return;
         }
      }
         
   [matrix selectCellWithTag:_tagValueForOther];
}


- (void) objectChanged
{
   NSMatrix         *matrix;
   NSCell           *cell;
   NSString         *title;
   NSString         *key;
   int              tag;
   
   matrix = [self object];
   cell   = [matrix selectedCell];
   
   // allow both tag and title to be set, sounds "more compatible"
   key = [self displayGroupKeyForAspect:MulleEOInterfaceSelectedTitleAspect];
   if( key)
   {
      title = [cell title];
      [self setValue:title
           forAspect:MulleEOInterfaceSelectedTitleAspect];
   }
   
   key = [self displayGroupKeyForAspect:MulleEOInterfaceSelectedTagAspect];
   if( key)
   {
      tag = [cell tag];
      [self setValue:[NSNumber numberWithInt:tag]
           forAspect:MulleEOInterfaceSelectedTagAspect];
   }
}   


- (void) action:(id) sender
{
   [self objectChanged];
}

@end


