/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 *  $Id$
 *
 * 
 *  $Log$
 *  Revision 1.1.1.1  2002/04/21 23:58:37  nat
 *  Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#ifndef _EOTABLEVIEWASSOCIATION_H
#define _EOTABLEVIEWASSOCIATION_H

#import "EOAssociation.h"


@class EOColumnAssociation;
@class EOTableViewAssociationDragNDropInfo;
@class NSControl;
@class NSTableView;
@class NSTableColumn;
@class NSText;


typedef struct
{
   eo_association_extra                  extra;
   EOTableViewAssociationDragNDropInfo   *dragNDropInfo_;
} eo_tableview_association_extra;


typedef enum
{
   EOCompatibilitySortMode     = 0x0,  // no sorting (only column order sorting if enabled)

   EOUseSergioSorting          = 0x1,  // all these only enabled if column order sorting is DISABLED!
   EOAllowMultipleSorts        = 0x2,
   EOUseAltKeyForMultipleSorts = 0x4,
   EOShowSortIndicatorImage    = 0x8,
   EOAllowNilInSorts           = 0x10,
   
   EOAlwaysSortMode            = EOAllowMultipleSorts | EOShowSortIndicatorImage,
   EOInterfaceSortMode         = EOAllowMultipleSorts | EOShowSortIndicatorImage | EOAllowNilInSorts,
   EOSergioSortMode            = EOUseSergioSorting | EOAllowMultipleSorts | EOUseAltKeyForMultipleSorts | EOShowSortIndicatorImage | EOAllowNilInSorts
} EOTableViewAssociationSortBits;


@interface EOTableViewAssociation : EOAssociation
{
   unsigned   _updating:1;	
   unsigned   _enabledAspectBound:1;
   unsigned   _colorAspectBound:1;
   unsigned   _boldAspectBound:1;

   unsigned   _italicAspectBound:1;
   unsigned   _sortsByColumnOrder:1;      // traditional
   unsigned   _didSetSortOrdering:1;
   unsigned   _autoCreated:1;

   unsigned   _sortMode:8;

   unsigned   _dragNDropDisabled:1;
   unsigned   _allowsEmptySelection:1;
   unsigned   _inQueryMode:1;
   unsigned   _pingPongDetected:1;
   
   unsigned   _unused:12;
}


// only used for some notifications!!
- (void) setDragNDropInfo:(EOTableViewAssociationDragNDropInfo *) info;
- (EOTableViewAssociationDragNDropInfo *) dragNDropInfo;

// used to be void
+ (id) bindToTableView:(NSTableView *) tableView
          displayGroup:(EODisplayGroup *) displayGroup;

- (void) setSortsByColumnOrder:(BOOL)yn;
- (BOOL) sortsByColumnOrder;
- (void) setDragNDropEnabled:(BOOL)yn;
- (BOOL) dragNDropEnabled;

- (EOColumnAssociation *) editingAssociation;

- (int) numberOfRowsInTableView:(NSTableView *) tableView;
- (void) tableViewSelectionDidChange:(NSNotification *) notification;

- (id)          tableView:(NSTableView *) tableView
objectValueForTableColumn:(NSTableColumn *) tableColumn
                      row:(int)row;
- (void)        tableView:(NSTableView *) tableView
           setObjectValue:(id) object
           forTableColumn:(NSTableColumn *) tableColumn
                      row:(int) row;
- (void)        tableView:(NSTableView *) tableView
          willDisplayCell:(id) cell
           forTableColumn:(NSTableColumn *) tableColumn
                      row:(int) row;
- (BOOL)        tableView:(NSTableView *) tableView
    shouldEditTableColumn:(NSTableColumn *) tableColumn
                      row:(int) row;

- (BOOL)          control:(NSControl *) control
            isValidObject:(id)obj;
- (BOOL)          control:(NSControl *) control
    didFailToFormatString:(NSString *) string
         errorDescription:(NSString *) errorDescription;
- (BOOL)          control:(NSControl *) control
   textShouldBeginEditing:(NSText *) fieldEditor;

// mulle additions:

+ (void) setDefaultSortMode:(unsigned int) sortBits;
- (void) setSortMode:(unsigned int) sortBits;
- (unsigned int) sortMode;

@end

#endif
