/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOTableViewAssociation.h"

#import "AspectsAndKeys.h"
#import "EOColumnAssociation.h"
#import "EOColumnAssociationProxy.h"
#import "EOControlAssociation.h"
#import "EODisplayGroup.h"
#import "EODisplayGroup+Associations.h"
#import "EODisplayGroup+Objects.h"
#import "EODisplayGroup+Selection.h"
#import "EODisplayGroup+InsertUpdateDelete.h"
#import "EOTableViewAssociation+DragNDrop.h"
#import "NSObject+TableViewAssociation.h"
#import "EOAssociation+RenderCellByAspect.h"
#import <objc/objc-class.h>

@implementation EOTableViewAssociation


+ (size_t) associationExtraSize
{
   return( sizeof( eo_tableview_association_extra));
}


static inline eo_tableview_association_extra   *tableviewAssociationExtra( EOAssociation *self)
{
   return( (eo_tableview_association_extra *) &((char *) self)[ ((struct objc_class *) self->isa)->instance_size]);
}


static inline EOTableViewAssociationDragNDropInfo   *dragNDropInfo( EOAssociation *self)
{
   return( tableviewAssociationExtra( self)->dragNDropInfo_);
}


static inline void   setDragNDropInfo( EOAssociation *self, EOTableViewAssociationDragNDropInfo *info)
{
   tableviewAssociationExtra( self)->dragNDropInfo_ = info;
}


- (id) initWithObject:(id) obj
{
   [super initWithObject:obj];

//   _sortsByColumnOrder = YES;

   [[NSNotificationCenter defaultCenter] addObserver:self
                                            selector:@selector( columnDidMoveNotification:)
                                                name:NSTableViewColumnDidMoveNotification
                                              object:obj];
   return( self);
}


- (id) initWithCoder:(NSCoder *) decoder
{
   unsigned short   bits;

   [super initWithCoder:decoder];

#if DEBUG_CODER
   NSLog( @"Continueing %@ %s 0x%lx (version=%d)",  isa, __PRETTY_FUNCTION__, (long) self, [decoder versionForClassName:@"EOTableViewAssociation"]);
#endif
   [decoder decodeValueOfObjCType:@encode( unsigned short)
                               at:&bits];
   [self setSortsByColumnOrder:bits];
//   _sortsByColumnOrder = bits;  // i just gzess

   // keep this for query mode (query mode no empty sel!)
#if DEBUG_CODER
   NSLog( @"Leave %@ %s %p", isa, __PRETTY_FUNCTION__, self);
#endif

   return( self);
}


- (void) dealloc
{
   [[NSNotificationCenter defaultCenter] removeObserver:self];
   [super dealloc];
}


- (void) encodeWithCoder:(NSCoder *) encoder
{
   unsigned short   bits;

   [super encodeWithCoder:encoder];

#if DEBUG_CODER   
   NSLog( @"Continueing %@ %s 0x%lx",  isa, __PRETTY_FUNCTION__, (long) self);
#endif
   bits = _sortsByColumnOrder;
   [encoder encodeValueOfObjCType:@encode( unsigned short)
                               at:&bits];

#if DEBUG_CODER   
   NSLog( @"Leave %@ %s 0x%lx", isa, __PRETTY_FUNCTION__, (long) self);
#endif
}


+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[NSTableView class]]);
}


+ (NSString *) primaryAspect
{
   return( MulleEOInterfaceSourceAspect);
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceSourceAspect, MulleEOInterfaceEnabledAspect,
         MulleEOInterfaceTextColorAspect, MulleEOInterfaceBoldAspect, MulleEOInterfaceItalicAspect, nil];
   return( aspects);
}


+ (NSArray *)  aspectSignatures
{
   static NSArray  * aspectSignatures;

   if( !  aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"", @"A1M", @"A", @"A", @"A", nil];

   return(  aspectSignatures);
}


#warning (nat) hmm, DL changed this around but its now inverse to my original intention. peculiar
- (EOObserverPriority) priority
{
   //return( EOObserverPriorityThird);
   return( EOObserverPriorityFifth); //DL
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:MulleEOInterfaceDataSourceKey,
         MulleEOInterfaceDelegateKey, MulleEOInterfaceTargetKey, nil];
   return( objectKeysTaken);
}


+ (NSArray *) associationClassesSuperseded
{
   static NSArray  *associationClassesSuperseded;

   if( ! associationClassesSuperseded)
      associationClassesSuperseded = [[NSArray alloc] initWithObjects:[EOControlAssociation class], nil];
   return( associationClassesSuperseded);
}


//
// this routine is called by EOColumnAssociation mostly...
//
+ (id) bindToTableView:(NSTableView *) tableView
            displayGroup:(EODisplayGroup *) displayGroup
{
   id   association;
   id   delegate;
   
   // already set up ?
   // don't enforce EOTableAssociation!!
   delegate = [tableView delegate];
   if( [delegate isKindOfClass:[EOAssociation class]])  // already set up
   {
      // ensure that columns have same group
      NSParameterAssert( ! [delegate isKindOfClass:[EOTableViewAssociation class]] ||
                         [delegate displayGroupForAspect:MulleEOInterfaceSourceAspect] == displayGroup);
      return( delegate);
   }
   
#if DEBUG
   if( delegate)
      NSLog( @"TableView %@ has a delegate %@ set up already. We'll torch it now", tableView, delegate);
#endif
   
   association = [[[self alloc] initWithObject:tableView] autorelease];
   [association bindAspect:MulleEOInterfaceSourceAspect
              displayGroup:displayGroup
                       key:nil];
   [association establishConnection];
   return( association);
}


static unsigned char   EODefaultSortMode = EOCompatibilitySortMode;

+ (void) setDefaultSortMode:(unsigned int) bits
{
   NSParameterAssert( bits < 0x20);
   
   EODefaultSortMode = bits;
}


- (void) setSortMode:(unsigned int) bits
{
   NSParameterAssert( bits < 0x20);
   
   _sortMode = bits;
   if( bits)
      _sortsByColumnOrder = NO;     //otherwise keep previous value
}


- (unsigned int) sortMode
{
   return( _sortMode);
}


- (NSArray *) columnOrderedSortOrderings
{
   NSMutableArray        *array;
   EOColumnAssociation   *association;
   EOSortOrdering        *sort;
   NSArray               *columnAssociations;
   unsigned int          i, n;
   NSTableView           *tableView;
   
   tableView          = [self object];
   columnAssociations = [EOColumnAssociation columnAssociationsForTableView:tableView];
   array              = [NSMutableArray array];

   //
   // force update once on column associations, now is the
   // time because they should be available
   //
   if( ! _didSetSortOrdering)
      [self setSortsByColumnOrder:[self sortsByColumnOrder]];
   
   n = [columnAssociations count];
   for( i = 0; i < n; i++)
   {
      association = [columnAssociations objectAtIndex:i];
      sort        = [association sortOrdering];
      if( ! sort)    // as an alternative we might want to set it here
         continue;
      [array addObject:sort];
   }
   return( array);
}


static BOOL  _isEqualSortOrderings( EOSortOrdering *a, EOSortOrdering *b)
{
   return( [[a key] isEqualToString:[b key]] &&
           [a selector] == [b selector]);
}


static unsigned int   _sortOrderingsIndexOfSortOrdering( NSArray *array, EOSortOrdering *ordering)
{
   EOSortOrdering  *candidate;
   unsigned int    i, n;
   
   n = [array count];
   for( i = 0; i < n; i++)
   {
      candidate = [array objectAtIndex:i];
      if( _isEqualSortOrderings( candidate, ordering))
         return( i);
   }
   return( NSNotFound);
}


- (void) resortForChangedSortOrderingFrom:(EOSortOrdering *) before
                                       to:(EOSortOrdering *) now
{
   EODisplayGroup   *sourceGroup;
   NSMutableArray   *sortOrderings;
   unsigned int     index;

   sourceGroup   = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   
   sortOrderings = [NSMutableArray arrayWithArray:[sourceGroup sortOrderings]];
   
   index = _sortOrderingsIndexOfSortOrdering( sortOrderings, before);
   if( index != NSNotFound)
   {
      [sortOrderings removeObjectAtIndex:index];
      if( now)
         [sortOrderings insertObject:now
                             atIndex:index];
   }
   else
      if( now)
         [sortOrderings addObject:now];

   [sourceGroup setSortOrderings:sortOrderings];
   [sourceGroup updateDisplayedObjects];
   _didSetSortOrdering = YES;
}


- (void) resortByColumnOrder
{
   EODisplayGroup  *sourceGroup;
   NSArray         *sortOrderings;
   
   sourceGroup   = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   sortOrderings = [self columnOrderedSortOrderings];
   [sourceGroup setSortOrderings:sortOrderings];
   [sourceGroup updateDisplayedObjects];
   _didSetSortOrdering = YES;
}


- (void) resortByColumnOrderIfWanted
{
   if( ! _didSetSortOrdering && [self sortsByColumnOrder])
      [self resortByColumnOrder];
}


- (void) resortIfWanted
{
   if( ! _didSetSortOrdering)
      [self resortByColumnOrder];
}


- (void) setSortsByColumnOrder:(BOOL) flag
{
   NSArray   *associations;
   
   _sortsByColumnOrder = flag;
   if( _sortsByColumnOrder)
   {
      associations = [EOColumnAssociation columnAssociationsForTableView:[self object]];
      [associations makeObjectsPerformSelector:@selector( _setSortingSelectorIfNeeded:)
                                    withObject:(id) EOCompareAscending];
//   [associations makeObjectsPerformSelector:@selector( setSortingEnabled:)
//                                 withObject:(id) (int) flag];
   }
}


- (BOOL) sortsByColumnOrder
{
   return( _sortsByColumnOrder);
}


- (void) setDragNDropEnabled:(BOOL) flag
{
   _dragNDropDisabled = ! flag;
}


- (BOOL) dragNDropEnabled
{
   return( ! _dragNDropDisabled);
}


- (void) setDragNDropInfo:(EOTableViewAssociationDragNDropInfo *) info
{
   setDragNDropInfo( self, info);  // not retained
}


- (EOTableViewAssociationDragNDropInfo *) dragNDropInfo
{
   return( dragNDropInfo( self));
}


- (NSString *) description
{
   EODataSource   *dataSource;
   EODisplayGroup  *sourceGroup;
   
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   dataSource  = [sourceGroup dataSource];
   if( ! dataSource)
      return( [super description]);
   
   return( [NSString stringWithFormat:@"<%@ 0x%lx: %@>", isa, (long) self,
      [[dataSource classDescriptionForObjects] entityName]]);
}


- (void) registerForDragNDrops
{
   // nuthing
}


- (void) unregisterForDragNDrops
{
   [[self object] unregisterDraggedTypes];
}


- (void) _breakWrongConnectionToObject:(id) other
{
   if( [other isKindOfClass:[EOTableViewAssociation class]])
   {
      // don't warn, it's normal
      [other breakConnection];
      return;
   }
   
   [super _breakWrongConnectionToObject:other];
}


- (void) establishConnection
{
   NSTableView   *tableView;
   
#if DEBUG_VERBOSE
   NSLog( @"%s %@", __PRETTY_FUNCTION__, self);
#endif   
/*
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   if( ! sourceGroup)
   {
      NSLog( @"%@ %@ aspect not connected", self, MulleEOInterfaceSourceAspect);
      return;
   }
*/
   tableView = [self object];
   //
   // by design, the first EOColumnAssociation creates a EOTableView.
   // Now when the real one gets deserialized later, it's a problem.
   // what we do is, remove the created TableView replace it with self 
   // rewire the EOColumnAssociations (if needed).
   // breakConnection on the other
   //
   [self _breakWrongConnectionToObject:[tableView delegate]];
   [self _breakWrongConnectionToObject:[tableView dataSource]];

   [tableView setDelegate:self];
   [tableView setDataSource:self];

   if( ! _dragNDropDisabled)
   {
      //
      // perform this delayed, so that recipients have a chance to set
      // themselves up in their awakeFromNib
      //
      [self performSelector:@selector( registerForDragNDrops)
                 withObject:nil
                 afterDelay:0.01];
   }

   _enabledAspectBound = [self displayGroupKeyForAspect:MulleEOInterfaceEnabledAspect] != nil;
   if( _enabledAspectBound && [self displayGroupForAspect:MulleEOInterfaceEnabledAspect] != [self displayGroupForAspect:MulleEOInterfaceSourceAspect])
      NSLog( @"%@ Enabled aspect bound to different displayGroup, which will be ignored but may lead to problems", self);

   _colorAspectBound  = [self displayGroupKeyForAspect:MulleEOInterfaceTextColorAspect] != nil;
   if( _colorAspectBound && [self displayGroupForAspect:MulleEOInterfaceTextColorAspect] != [self displayGroupForAspect:MulleEOInterfaceSourceAspect])
      NSLog( @"%@: Color aspect bound to different displayGroup, which will be ignored but may lead to problems", self);

   _boldAspectBound   = [self displayGroupKeyForAspect:MulleEOInterfaceBoldAspect] != nil;
   if( _boldAspectBound && [self displayGroupForAspect:MulleEOInterfaceBoldAspect] != [self displayGroupForAspect:MulleEOInterfaceSourceAspect])
      NSLog( @"%@: Bold aspect bound to different displayGroup, which will be ignored but may lead to problems", self);

   _italicAspectBound = [self displayGroupKeyForAspect:MulleEOInterfaceItalicAspect] != nil;
   if( _italicAspectBound && [self displayGroupForAspect:MulleEOInterfaceItalicAspect] != [self displayGroupForAspect:MulleEOInterfaceSourceAspect])
      NSLog( @"%@ Italic aspect bound to different displayGroup, which will be ignored but may lead to problems", self);

   _sortMode = EODefaultSortMode; 
   
//   [self resortByColumnOrderIfWanted];

   _inQueryMode = NO;
   [super establishConnection];
   
}


- (void) breakConnection
{
   NSTableView   *tableView;
   
   tableView = [self object];
   if( ! _dragNDropDisabled)
      [self unregisterForDragNDrops];

   if( _inQueryMode)
      [tableView setAllowsEmptySelection:_allowsEmptySelection];
   [tableView setDelegate:nil];
   [tableView setDataSource:nil];

   [super breakConnection];
}


- (EOColumnAssociation *) editingAssociation
{
   id              association;
   EODisplayGroup  *sourceGroup;
   
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   association = [sourceGroup editingAssociation];

   if( [association isKindOfClass:[EOColumnAssociation class]])
      return( association);
   return( nil);
}


static NSTableColumn  *editedTableColumn( EOTableViewAssociation *self)
{
   NSTableView   *tableView;
   int           index;
   
   tableView = [self object];
   index     = [tableView editedColumn];
   if( index == -1)
   {
      NSLog( @"Your %@ sends funny messages", tableView);
      return( nil);
   }
   return( [[tableView tableColumns] objectAtIndex:index]);
}



- (int) numberOfRowsInTableView:(NSTableView *) tableView
{
   EODisplayGroup  *sourceGroup;
   int              n;
   
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   n           = [[sourceGroup displayedObjects] count];
   return( n);
}


- (void) _reflectTableViewSelection
{
   NSArray         *selectedRows;
   EODisplayGroup  *sourceGroup;
   NSTableView     *tableView;
   
   tableView    = [self object];
   selectedRows = [[tableView selectedRowEnumerator] allObjects];
   //   memorizeSelectedRows( self, selectedRows);
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   [sourceGroup setSelectionIndexes:selectedRows];  // this will create a subject changed message
}


- (void) tableViewSelectionDidChange:(NSNotification *) notification
{
   if( _updating)  // ignore these recursive calls
      return;
   [self _reflectTableViewSelection];
}


- (id)          tableView:(NSTableView *) tableView
objectValueForTableColumn:(NSTableColumn *) tableColumn
                      row:(int) row
{
   EODisplayGroup        *sourceGroup;
   EOColumnAssociation   *assoc;
   NSArray               *objects;
   id                    value;
   id                    eo;
   id                    identifier;
   int                   count;

   assoc = [EOColumnAssociation columnAssociationForTableColumn:tableColumn];
   if( assoc)
   {
      return( [assoc tableView:tableView
     objectValueForTableColumn:tableColumn
                           row:row]);
   }
   
   identifier = [tableColumn identifier];
   if( ! identifier)
   {
#if DEBUG      
      return( @"Identifier missing. No EOColumnAssociation found");
#endif      
      return( @"");
   }
   
   value       = nil;
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   objects     = [sourceGroup displayedObjects];
   count       = [objects count];
   
   if( row <= count)
   {
      eo    = [objects objectAtIndex:row];
      value = [eo valueForKeyPath:identifier];
   }
   return( value ?: @"???");
}


//
// not time critical
//
- (void)   tableView:(NSTableView *) tableView
      setObjectValue:(id) value
      forTableColumn:(NSTableColumn *) tableColumn
                 row:(int) row
{
   EODisplayGroup        *sourceGroup;
   EOColumnAssociation   *assoc;
   NSString              *identifier;
   
#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@ %@ %@ %u", self, __PRETTY_FUNCTION__, tableView, value, tableColumn, row);
#endif
   assoc = [EOColumnAssociation columnAssociationForTableColumn:tableColumn];
   if( assoc)
   {
      [assoc tableView:tableView
        setObjectValue:value
        forTableColumn:tableColumn
                   row:row];
      return;
   }

   identifier  = [tableColumn identifier];
   if( ! [identifier isKindOfClass:[NSString class]])
   {
      NSLog( @"*** warning: %@ tableColumn %@ has no valid identifier %@", self, tableColumn, identifier);
      return;
   }
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   if( ! [sourceGroup setValue:value
               forObjectAtIndex:row
                            key:identifier])
   {
      NSArray        *objects;
      unsigned int   count;
      
      objects = [sourceGroup displayedObjects];
      count   = [objects count];
      if( row >= count)
      {
         NSLog( @"Index out of bounds %u", row);
         return;
      }
      
      NSLog( @"Setting of object value %@ for key %@ failed for EO %@ at row %u",
             value, identifier, [objects objectAtIndex:row], row);
   }
}


- (void) tableView:(NSTableView *) tableView
   willDisplayCell:(id) cell
    forTableColumn:(NSTableColumn *) tableColumn
               row:(int) row
{
   EOColumnAssociation   *assoc;
   
   [self _renderCell:cell
       enabledAspect:_enabledAspectBound ?  MulleEOInterfaceEnabledAspect : nil
         colorAspect:_colorAspectBound ? MulleEOInterfaceTextColorAspect : nil
          boldAspect:_boldAspectBound ? MulleEOInterfaceBoldAspect : nil
        italicAspect:_italicAspectBound ? MulleEOInterfaceItalicAspect : nil
              forRow:row];
   
   assoc = [EOColumnAssociation columnAssociationForTableColumn:tableColumn];
   [assoc tableView:tableView
    willDisplayCell:cell
     forTableColumn:tableColumn
                row:row];
}


- (BOOL)     tableView:(NSTableView *) tableView
 shouldEditTableColumn:(NSTableColumn *) tableColumn
                   row:(int) row;
{
   EOColumnAssociation  *assoc;

#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@ %@ %u", self, __PRETTY_FUNCTION__, tableView, tableColumn, row);
#endif
   assoc = [EOColumnAssociation columnAssociationForTableColumn:tableColumn];
   if( assoc)
      return( [assoc tableView:tableView
         shouldEditTableColumn:tableColumn
                           row:row]);
   return( [self objectStateForAspect:MulleEOInterfaceEnabledAspect] != EOAssociationObjectStateDisabled);
}


- (void)   tableView:(NSTableView *) tableView
 didClickTableColumn:(NSTableColumn *) tableColumn
{
   EOColumnAssociation  *assoc;
   EOColumnAssociation  *candidate;
   EOSortOrdering       *before;
   EOSortOrdering       *after;
   NSEvent              *event;
   BOOL                 addFlag;
   NSArray              *associations;
   unsigned int         i, n;
   
   // this means, sorting is only done when dragging not clicking, so ignore click
   if( [self sortsByColumnOrder])
      return;
   
   // means we don't do any other sorting, if compatible
   if( [self sortMode] == EOCompatibilitySortMode)
      return;
   
   assoc = [EOColumnAssociation columnAssociationForTableColumn:tableColumn];
   if( ! assoc)
      return;
   
   addFlag = _sortMode & EOAllowMultipleSorts;
   if( _sortMode & EOUseAltKeyForMultipleSorts)
   {
      event   = [[tableView window] currentEvent]; 
      addFlag = [event modifierFlags] & NSAlternateKeyMask ? YES : NO;
   }

   // if we don't add to the sorting, then nil out 
   // all other columns sorting selectors
   if( ! addFlag)
   {
      associations = [EOColumnAssociation columnAssociationsForTableView:tableView];
      n = [associations count];
      for( i = 0; i < n; i++)
      {
         candidate = [associations objectAtIndex:i];
         if( candidate != assoc)
            [candidate setSortingSelector:(SEL) 0];
      }
   }
   
   //
   // if we use sergio sorting then we want to add the 
   // sorting columns differently, otherwise one sergio
   // column is just the same as column order sort
   //
   if( addFlag && (_sortMode & EOUseSergioSorting))
   {
      before = [assoc sortOrdering];
      [assoc tableView:tableView didClickTableColumn:tableColumn];
      after  = [assoc sortOrdering];
      [self resortForChangedSortOrderingFrom:before
                                          to:after];
      return;
   }

   //
   // back to column order sort
   //
   [assoc tableView:tableView didClickTableColumn:tableColumn];
   [self resortByColumnOrder];
}


- (void)  tableView:(NSTableView*) tableView
 didDragTableColumn:(NSTableColumn *) tableColumn
{
   if( ! [self sortsByColumnOrder])
      return;
   [self resortByColumnOrder];
}

//
// it's complicated to forward these to the approriate
// EOColumnAssociation correctly, because the selection has sometimes
// moved already, when the control is telling us to textDidEndEditing
// fortunately! the EODisplayGroup knows the editing association!
// So we use that when, we got it...
//
- (BOOL) control:(NSControl *) control
   isValidObject:(id) obj
{
   EOColumnAssociation  *association;
   EODisplayGroup       *sourceGroup;
   NSArray              *columnAssociations;
   NSTableView          *tableView;
   
#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@ %@", self, __PRETTY_FUNCTION__, control, obj);
#endif
   tableView   = [self object];
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   association = (EOColumnAssociation*) [sourceGroup editingAssociation];
   if( ! association)
      return( YES);  // oder was ??

   columnAssociations = [EOColumnAssociation columnAssociationsForTableView:tableView];
   if( [columnAssociations indexOfObjectIdenticalTo:association] == NSNotFound)
   {
      NSLog( @"Strange editing association %@ of %@ does not belong to %@", association, sourceGroup, self);
      return( YES);
   }
   return( [association control:control
                  isValidObject:obj]);
}


- (BOOL)          control:(NSControl *) control
    didFailToFormatString:(NSString *) string
         errorDescription:(NSString *) errorDescription;
{
   EOColumnAssociation  *association;
   EODisplayGroup       *sourceGroup;
   NSArray              *columnAssociations;
   NSTableView          *tableView;

#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@ %@ %@", self, __PRETTY_FUNCTION__, control, string, errorDescription);
#endif   
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   association = (EOColumnAssociation*) [sourceGroup editingAssociation];
   if( ! association)
      return( YES);  // oder was ??

   tableView          = [self object];
   columnAssociations = [EOColumnAssociation columnAssociationsForTableView:tableView];
   if( [columnAssociations indexOfObjectIdenticalTo:association] == NSNotFound)
   {
      NSLog( @"Strange editing association %@ of %@ does not belong to %@", association, sourceGroup, self);
      return( YES);
   }
   return( [association control:control
          didFailToFormatString:string
               errorDescription:errorDescription]);
}


- (BOOL)          control:(NSControl *) control
   textShouldBeginEditing:(NSText *) fieldEditor
{
   EOColumnAssociation  *association;
   NSTableColumn        *column;
   
#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@ %@", self, __PRETTY_FUNCTION__, control, [fieldEditor string]);
#endif
   column      = editedTableColumn( self);
   association = [EOColumnAssociation columnAssociationForTableColumn:column];
   if( association)
      return( [association control:control
            textShouldBeginEditing:fieldEditor]);
   return( YES);
}


- (BOOL) 	control:(NSControl *) control
   textShouldEndEditing:(NSText *) fieldEditor
{
   EOColumnAssociation  *association;
   EODisplayGroup       *sourceGroup;
   NSArray              *columnAssociations;
   NSTableView          *tableView;
   
#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@ %@", self, __PRETTY_FUNCTION__, control, [fieldEditor string]);
#endif   
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   association = (EOColumnAssociation *) [sourceGroup editingAssociation];
   if( ! association)
      return( YES);

   tableView          = [self object];
   columnAssociations = [EOColumnAssociation columnAssociationsForTableView:tableView];
   if( [columnAssociations indexOfObjectIdenticalTo:association] == NSNotFound)
   {
      NSLog( @"Strange editing association %@ of %@ does not belong to %@", association, sourceGroup, self);
      return( YES);
   }
   return( [association control:control
           textShouldEndEditing:fieldEditor]);
}


- (void) controlTextDidBeginEditing:(NSNotification *) notification
{
   EOColumnAssociation  *association;
   NSTableColumn        *column;

#if DEBUG_VERBOSE
   NSLog( @"%@ %s", self, __PRETTY_FUNCTION__);
#endif
   column      = editedTableColumn( self);
   association = [EOColumnAssociation columnAssociationForTableColumn:column];
   if( association)
      [association controlTextDidBeginEditing:notification];
}


- (void) controlTextDidEndEditing:(NSNotification *) notification
{
   EOColumnAssociation  *association;
   EODisplayGroup       *sourceGroup;
   NSArray              *columnAssociations;
   NSTableView          *tableView;

#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@", self, __PRETTY_FUNCTION__, notification);
#endif
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   if( ! (association = (EOColumnAssociation *) [sourceGroup editingAssociation]))
      return;

   tableView          = [self object];
   columnAssociations = [EOColumnAssociation columnAssociationsForTableView:tableView];
   if( [columnAssociations indexOfObjectIdenticalTo:association] == NSNotFound)
   {
      NSLog( @"Strange editing association %@ of %@ does not belong to %@", association, sourceGroup, self);
      return;
   }
   [association controlTextDidEndEditing:notification];
}


- (void) columnDidMoveNotification:(NSNotification *) notification
{
#if DEBUG_VERBOSE | 1
   NSLog( @"%@ %s %@", self, __PRETTY_FUNCTION__, notification);
#endif
}


- (BOOL) reselect
{
   unsigned int     row;
   unsigned int     first;
   unsigned int     max;
   unsigned int     i, n;
   NSNumber         *number;
   IMP              impSelectRow;
   IMP              impObjectAtIndex;
   IMP              impIntValue;
   BOOL             flag;
   NSArray          *selectedRows;
   NSArray          *indexes;
   EODisplayGroup   *sourceGroup;
   NSTableView      *tableView;
   
   tableView   = [self object];
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   indexes     = [NSArray arrayWithArray:[sourceGroup selectionIndexes]];

   n     = [indexes count];
   max   = [tableView numberOfRows];
   first = 0;  // keep compiler happy  :-/
#warning (nat) the behaviour for non-empty selection is not coded yet
   flag  = NO;
   if( ! n)
   {
      [tableView deselectAll:self];
      if( [tableView allowsEmptySelection])
         return( NO);
      
      //
      // hmm the NSTableView will have selected something, because
      // it very well could have happened during updating (which we ignored)
      // we now have to sync the DG manually. Now if for some reason, the
      // DG is reset to NO selection again, this can go on forever...
      //
      if( ! _pingPongDetected)
      {
         if( [[sourceGroup displayedObjects] count])
         {
            _pingPongDetected = YES;
            NSLog( @"Endless PingPong warning: NSTableView %p does not allow empty selection, but EODisplayGroup %p had no selection.", tableView, sourceGroup);
         }
      }
      
      return( YES);
   }
      //         if( ! [tableView allowsMultipleSelection])
      //            n = 1;
      
   impSelectRow     = [tableView methodForSelector:@selector( selectRow:byExtendingSelection:)];
   impObjectAtIndex = [indexes methodForSelector:@selector( objectAtIndex:)];
   impIntValue      = [[indexes lastObject] methodForSelector:@selector( unsignedIntValue)];
   for( i = 0; i < n; i++)
   {
      number = (*impObjectAtIndex)( indexes, @selector( objectAtIndex:), i);
      row    =  (* (unsigned int (*)()) impIntValue)( number, @selector( unsignedIntValue));
      if( row >= max)
      {
         NSLog( @"%@: Selection indices and TableView out of sync", self);
         break;
      }
      (*impSelectRow)( tableView, @selector( selectRow:byExtendingSelection:), row, flag);
      if( ! flag)
      {
         first = row;
         flag  = [tableView allowsMultipleSelection];
         if( flag)
            continue;
         
         // the display group has a multiple selection, but
         // the tableview only allows one. So the tableview wins
         // and forces the selection back. So this ping-ponging is
         // intended for once
         if( n > 1)
         {
            selectedRows = [NSArray arrayWithObject:[NSNumber numberWithInt:row]];
            [sourceGroup setSelectionIndexes:selectedRows];  
         }
         break;
      }
   }
   [tableView scrollRowToVisible:first];
   return( NO);
}


//
// Alle aspects laufen und checken...
// 
#warning (nat) subjectChanged needs to be subdivided into different functions
- (void) subjectChanged
{
   BOOL             reload;
   BOOL             reselect;
   BOOL             mirrorReselect; // warning is wrong
   EODisplayGroup   *old;
   EODisplayGroup   *group;
   EODisplayGroup   *sourceGroup;
   NSTableView      *tableView;
   NSException      *exception;
   
   reload   = NO;
   reselect = NO;
   old      = nil;
   
   tableView = [self object];
   
   // (nat) might want to add a firstTimeInit flag to EODisplayGroup
#warning (nat) lame, because this will trigger another object Changed!!
   
   [self resortByColumnOrderIfWanted];

   group = nil;
   if( _enabledAspectBound)
   {
      group = [self displayGroupForAspect:MulleEOInterfaceEnabledAspect];
      if( [group contentsChanged] | [group selectionChanged])
      {
         reload = YES;
         goto test_source;
      }
   }

   if( _colorAspectBound)
   {
      old   = group;
      group = [self displayGroupForAspect:MulleEOInterfaceTextColorAspect];
      if( group != old && ([group contentsChanged] || [group selectionChanged]))
      {
         reload = YES;
         goto test_source;
      }
   }

   if( _boldAspectBound)
   {
      old   = group;
      group = [self displayGroupForAspect:MulleEOInterfaceBoldAspect];
      if( group != old && ([group contentsChanged] || [group selectionChanged]))
      {
         reload = YES;
         goto test_source;
      }
   }

   if( _italicAspectBound)
   {
      old   = group;
      group = [self displayGroupForAspect:MulleEOInterfaceItalicAspect];
      if( group != old && ([group contentsChanged] || [group selectionChanged]))
         reload = YES;
   }

test_source:
   sourceGroup = [self displayGroupForAspect:MulleEOInterfaceSourceAspect];
   if( [sourceGroup inQueryMode])
   {
      if( ! _inQueryMode)
      {
         _allowsEmptySelection = [tableView allowsEmptySelection];
         [tableView setAllowsEmptySelection:NO];
         _inQueryMode = YES;
      }
   }
   else
   {
      if( _inQueryMode)
      {
         [tableView setAllowsEmptySelection:_allowsEmptySelection];
         _inQueryMode = NO;
      }
   }
   
   if( ! reload && [sourceGroup contentsChanged])
      reload = YES;

   exception = nil;
NS_DURING   // protect _updatingFlag
   _updating = YES;
   if( reload)
   {
//      uncacheDisplayedObjects( self);
      [tableView reloadData];
      reselect = YES;
   }

   if( ! reselect && [sourceGroup selectionChanged])
   {
//      indexes = [sourceGroup selectionIndexes];
//      if( ! [indexesMemory_ isEqualToArray:indexes])
      reselect = YES;
   }
#if DEBUG_VERBOSE
   NSLog( @"%s (%@)%s%s", __PRETTY_FUNCTION__, self, reload ? " RELOAD" : "", reselect ? " RESELECT" : "");
#endif   

   mirrorReselect = NO;
   if( reselect)
      mirrorReselect = [self reselect];
NS_HANDLER
   exception = localException;
NS_ENDHANDLER
   _updating = NO;
   [exception raise];

   if( mirrorReselect)
      [self _reflectTableViewSelection];
}


- (void) addBoundKeysForDisplayedObjectsToSet:(NSMutableSet *) set
{
   NSArray   *columnAssociations;
   
   // does nothing
   columnAssociations = [EOColumnAssociation columnAssociationsForTableView:[self object]];
   [columnAssociations makeObjectsPerformSelector:_cmd
                                       withObject:set];
}

@end


