/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EOTextAssociation.h"

#import <AppKit/AppKit.h>
#import "AspectsAndKeys.h"
#import "EODisplayGroup.h"
#import "EODisplayGroup+Associations.h"


@implementation EOTextAssociation

+ (BOOL) isUsableWithObject:(id) object
{
   return( [object isKindOfClass:[NSText class]]);
}


+ (NSArray *) aspects
{
   static NSArray  *aspects;

   if( ! aspects)
      aspects = [[NSArray alloc] initWithObjects:MulleEOInterfaceValueAspect, MulleEOInterfaceEditableAspect,
                     MulleEOInterfaceEnabledAspect, nil];
   return( aspects);
}


+ (NSArray *)  aspectSignatures
{
   static NSArray  * aspectSignatures;

   if( !  aspectSignatures)
      aspectSignatures = [[NSArray alloc] initWithObjects:@"A", @"A1M", nil];

   return(  aspectSignatures);
}


+ (NSArray *) objectKeysTaken
{
   static NSArray  *objectKeysTaken;

   if( ! objectKeysTaken)
      objectKeysTaken = [[NSArray alloc] initWithObjects:@"delegate", nil];
   return( objectKeysTaken);
}


+ (NSString *) primaryAspect
{
   return( MulleEOInterfaceValueAspect);  // or ??
}


typedef enum
{
   LastModification = 0,
   LastRedraw,     
   _NumberOfDateTypes
} DateType;


typedef struct
{
   NSTimeInterval    lastModification;
   NSTimeInterval    lastTimeSpent;
   EOObserverProxy   *proxy;
} text_association_extra;


+ (size_t) associationExtraSize
{
   return( [super associationExtraSize] + sizeof( text_association_extra));
}


+ (text_association_extra *) extraStorageForTextAssociation:(EOTextAssociation *) assoc
{
   size_t   size;
   void     *extra;
   
   size  = [super associationExtraSize]; // sic!
   extra = [assoc associationExtraStorage];
   return( (text_association_extra *) &((char *) extra)[ size]);
}


#if MULLE_EO_CONTROL
// called by proxy...
- (void) subjectChangedInRunLoopOrdering:(int) ordering
{
   text_association_extra   *extra;
   NSTimeInterval           interval;
   
   if( ordering != EONotifyTextAsssociationsAfterRedrawRunLoopOrdering)
      return;
   
   extra = [isa extraStorageForTextAssociation:self];
   if( ! extra->lastModification)
      return;
   
   interval = [NSDate timeIntervalSinceReferenceDate];
   NSParameterAssert( interval >  extra->lastModification);
   
   extra->lastTimeSpent = interval - extra->lastModification;
}

- (EOObserverProxy *) _observerProxy
{
   text_association_extra   *extra;
   
   extra = [isa extraStorageForTextAssociation:self];
   if( ! extra->proxy)
      extra->proxy = [[EOObserverProxy alloc] initWithTarget:self
                                                      action:@selector( subjectChangedInRunLoopOrdering:)
                                                    priority:EOObserverPriorityFourth];
   return( extra->proxy);
}


- (void) _touchLastModificationTimestamp
{
   EODelayedObserverQueue   *queue;
   text_association_extra   *extra;
   
   extra                   = [isa extraStorageForTextAssociation:self];
   extra->lastModification = [NSDate timeIntervalSinceReferenceDate];

   queue = [EODelayedObserverQueue defaultObserverQueueForRunLoopOrdering:EONotifyTextAsssociationsAfterRedrawRunLoopOrdering];
   [queue enqueueObserver:[self _observerProxy]];
}
#endif      


float   EOTextAssociationRefreshTimeLowerLimit = 0.1;    // 0.1 seconds max
float   EOTextAssociationRefreshTimeFactor     = 20.0;    // wait up till then do it a

//
// LastModification is when we were just before modifiying the object
// LastRedraw is when we were through drawing
// 
// 
- (BOOL) _shouldSetObjectAgain
{
   text_association_extra   *extra;
   NSTimeInterval           interval;
   
   extra = [isa extraStorageForTextAssociation:self];

   if( extra->lastTimeSpent <= EOTextAssociationRefreshTimeLowerLimit)
      return( YES);
   
   interval = extra->lastTimeSpent * EOTextAssociationRefreshTimeFactor;
   if( ([NSDate timeIntervalSinceReferenceDate] - interval) > extra->lastModification)
      return( YES);
   
   return( NO);
}


- (void) dealloc
{
//   [enabledColor_ release];
   text_association_extra   *extra;
   
   extra = [isa extraStorageForTextAssociation:self];
   [extra->proxy release];
   [_lastValue release];
   [super dealloc];
}


/* ####
 */
- (id) readValueFromObject
{
   if( ! [_object respondsToSelector:@selector( textStorage)])
   {
      if( [_object isRichText])
      {
         NSRange   all;
         
         all = NSMakeRange(0, [[(id) _object textStorage] length]);
         if( [_object importsGraphics])
            return( [_object RTFDFromRange:all]);
         return( [_object RTFFromRange:all]);
      }
   }
   return( [_object string]);
}


- (void) writeValueToObject:(id) value
{
   if( value && 
       ! [value isKindOfClass:[NSString class]] && 
       [_object respondsToSelector:@selector( textStorage)])
   {
      if( [_object isRichText])
      {
         NSRange   all;
         
         NSParameterAssert( [value isKindOfClass:[NSData class]]);
         all = NSMakeRange(0, [[(id) _object textStorage] length]);
         if( [_object importsGraphics])
         {
            [_object replaceCharactersInRange:all withRTF:value];
            return;
         }
         [_object replaceCharactersInRange:all withRTFD:value];
         return;
      }
   }

   if( ! value)
      value = @"";
   NSParameterAssert( [value isKindOfClass:[NSString class]]);
   [_object setString:value];
}


- (EOAssociationObjectState) objectStateForEditableAspect
{
   return( [self objectStateForAspect:MulleEOInterfaceEditableAspect
                  editingDisplayGroup:[self displayGroupForAspect:MulleEOInterfaceValueAspect]]);
}


//
// turn it on to get http://developer.apple.com/qa/qa2006/qa1461.html 
// behaviour, or use the EnabledAspect...
//
BOOL  MulleEOTextAssociationTreatEditableLikeEnabled = YES;


- (BOOL) setObjectEnabled:(BOOL) flag
{
   NSColor   *color;
   
   if( [_object isSelectable] == flag)
      return( NO);
   
   [_object setSelectable:flag];
   [_object setEditable:flag];
   
   //
   // if it is rich text, then disabling destroys the font information
   // (not good IMO) - but is it really a problem ? Because that info
   // should be in the EO object by then
   //
//   color = flag ? enabledColor_ : [NSColor disabledControlTextColor];
   color = flag ? [NSColor controlTextColor] : [NSColor disabledControlTextColor];
   [_object setTextColor:color];

   return( YES);
}


/* ###
 */
- (void) establishConnection
{
   BOOL   flag;
   
   _wasSelectable = [_object isSelectable];
   _wasEditable   = [_object isEditable];
   
   // pull default color in
//   enabledColor_  = [[_object textColor] retain];
   _editedYet     = NO;
   
   flag = [self objectStateForEnabledAspect];
   if( flag != EOAssociationObjectStateKeepState)
      [self setObjectEnabled:flag];

   // this may override editable again (but makes no sense)
   flag = [self objectStateForEditableAspect];
   if( flag != EOAssociationObjectStateKeepState)
      [_object setEditable:flag];
   
   [_object setDelegate:self];

   // clean up object 
   if( ! keepInitialString_)
      [self writeValueToObject:nil];

   [super establishConnection];
}


- (void) breakConnection
{
   id   obj;

   obj = [self object];
   
   [obj setDelegate:nil];
   [obj setEditable:_wasEditable];
   [obj setSelectable:_wasSelectable];
//   [obj setTextColor:enabledColor_];

//   [enabledColor_ release];
//   enabledColor_ = nil;
      
   [super breakConnection];
}


- (void) _setLastValue:(id) value
{
   id  old;

   old = _lastValue;
   _lastValue = [value retain];
   [old release];
}


- (BOOL) setValue:(id) value
        forAspect:(NSString *) MulleEOInterfaceValueAspect
{
#if MULLE_EO_CONTROL
   [self _touchLastModificationTimestamp];  // touch before setting
#endif   
   if( [super setValue:value
             forAspect:MulleEOInterfaceValueAspect])
   {
      [self _setLastValue:value];
      return( YES);
   }
   return( NO);
}


static id   valueFromObject( EOTextAssociation *self)
{
   id   value;
   
   value = [self readValueFromObject];
   if( ! self->dontCopyObjects_)
      value = [[value copy] autorelease];
   return( value);
}


static void   syncWithObject( EOTextAssociation *self)
{
   id   value;
   
   value = valueFromObject( self);
   [self setValue:value
        forAspect:MulleEOInterfaceValueAspect];
}


- (BOOL) textShouldEndEditing:(NSText *) fieldEditor
{
   NSException   *exception;
   id            value;
   id            original;

   original  = valueFromObject( self);
   value     = original;
   exception = [self validateValue:&value
                         forAspect:MulleEOInterfaceValueAspect];
   if( exception != nil)
      return( [self shouldEndEditingForAspect:MulleEOInterfaceValueAspect
                                 invalidInput:original
                             errorDescription:[exception reason]]);

   if( original && ! [original isEqual:value])
   {
      [self setValue:value
           forAspect:MulleEOInterfaceValueAspect];
   }
   return( YES);
}


// ensure that, after a manual "endEditing" we don't lose
// state...

- (void) textDidChange:(NSNotification *) notification
{
   EODisplayGroup   *dg;
   BOOL             flag;
   
   dg = [self displayGroupForAspect:MulleEOInterfaceValueAspect];
   if( [dg editingAssociation] != self)
      [dg associationDidBeginEditing:self];
   
   flag = ! _editedYet || updateAfterEveryChange_;
   if( ! flag)
      flag = [self _shouldSetObjectAgain];
   
   if( flag)
   {
      syncWithObject( self);
      _editedYet = YES;
   }
}


// we get this message when actual typing starts
- (void) textDidBeginEditing:(NSNotification *) notification
{
   // do nothing currently
}


// but we get this message when we lose focus, so it's not guaranteed that
// we got a textDidBeginEditing
- (void) textDidEndEditing:(NSNotification *) notification
{
   EODisplayGroup  *dg;
   
   dg = [self displayGroupForAspect:MulleEOInterfaceValueAspect];
   if( [dg editingAssociation] == self)
   {
      syncWithObject( self);
      _editedYet = NO;
      [dg associationDidEndEditing:self];
   }
}


- (void) subjectChanged
{
   EOAssociationObjectState   state;
   id      value;
   BOOL    flag;

#if DEBUG_VERBOSE
   NSLog( @"%s (%@)", __PRETTY_FUNCTION__, self);
#endif
   
   flag  = NO;
   state = [self objectStateForEnabledAspect];
   if( state != EOAssociationObjectStateKeepState)
      flag = [self setObjectEnabled:state];
   
   // this may override editable again (but makes no sense)
   state = [self objectStateForEditableAspect];
   if( state != EOAssociationObjectStateKeepState)
   {
      if( [_object isEditable] != state)
      {
         [_object setEditable:state];
         flag = YES;
      }
   }

   value = [self valueForAspect:MulleEOInterfaceValueAspect];
   if( value != _lastValue || (value && ! [value isEqual:_lastValue]))
   {
      _editedYet = NO;
      if( ! dontCopyObjects_)
         value = [[value copy] autorelease];
     [self writeValueToObject:value];
     [self _setLastValue:value];
     flag = YES;
   }
   
   if( flag)
   {
      [_object setNeedsDisplay:YES];
   }
}


//
// we get endEditing if we are still editing usually when changes
// are to be saved or discarded.
// These are the scenarios:
//
// User pressed a different control, we are out of focus, but the
// message we'll receive comes in to late (if at all)
// -> solution, update model but setup textDidChange to auto begin
// 
// User pressed a key combination (Apple-S). So we are still editing
// and we won't get the message anyhow 
// -> solution, update model but setup textDidChange to auto begin
//
// 
- (BOOL) endEditing
{
   EODisplayGroup  *dg;

   if( ! [self textShouldEndEditing:nil])
      return( NO);
   
   syncWithObject( self);
   _editedYet = NO;
   dg = [self displayGroupForAspect:MulleEOInterfaceValueAspect];
   [dg associationDidEndEditing:self];
   return( YES);
}

@end


