/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF
 *  Copyright (C) 2006 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 *
 */
#import "Defines.h"

#import "EODisplayGroup+Associations.h"

#import <AppKit/AppKit.h>
#import "EODisplayGroup+Associations.h"
#import "EODisplayGroup+Private.h"
#import "EOAssociation.h"


@implementation EODisplayGroup ( Associations)

// mulle addition
- (id) observingAssociationForObject:(id) obj
{
   NSMutableArray   *array;
   NSArray          *observers;
   EOAssociation    *association;
   unsigned int     i, n;
   
   array     = [NSMutableArray array];
   observers = [self observingAssociations];
   
   n = [observers count];
   for( i = 0; i < n; i++)
   {
      association = [observers objectAtIndex:i];
      if( [association object] == obj)
         return( association);
   }
   return( nil);
}


- (NSArray *) observingAssociations
{
   NSMutableArray   *array;
   NSArray          *observers;
   unsigned int     i, n;
   
   observers = [EOObserverCenter observersForObject:self];
   if( n = [observers count])
   {
      Class    asoClass;
      id       association;
      
      asoClass = [EOAssociation class];
      array    = [NSMutableArray arrayWithCapacity:n];
      for( i = 0; i < n; i++)
      {
         association = [observers objectAtIndex:i];
         if( [association isKindOfClass:asoClass])
            [array addObject:association];
      }
      observers = array;
   }
   return( observers);
}


static inline void   discardEditing( EODisplayGroup *self)
{
   self->_editingAssociation = nil;
}


static BOOL  endEditing( EODisplayGroup *self)
{
   if( ! self->_editingAssociation)
      return( YES);
   
   if( [self->_editingAssociation endEditing])
   {
      NSCAssert1( self->_editingAssociation == nil, @"Association failed to call \"associationDidEndEditing:\" on %@", self);
      return( YES);
   }
   return( NO);
}


- (void) endEditingOrRaise
{
#warning (nat) NSInternalInconsistencyException ???
   if( ! endEditing( self))
      [NSException raise:NSInternalInconsistencyException  
                  format:@"User is still editing"];
}


- (BOOL) endEditing
{
   return( endEditing( self));
}


- (BOOL)  association:(EOAssociation *) association 
failedToValidateValue:(NSString *) value
               forKey:(NSString *) key
               object:(id) eo
     errorDescription:(NSString *) errorDescription
{
   NSRunCriticalAlertPanel( @"Validation Failure",
                            errorDescription,
                            @"OK", nil, nil);
   return( NO);
}


- (void) associationDidBeginEditing:(EOAssociation *) association
{
   NSParameterAssert( [association isKindOfClass:[EOAssociation class]]);
   
   //
   // tell old editing association that it is "done"
   // cell associations can't properly end their editing because
   // they don't know when the focus leaves
   //
   if( _editingAssociation)
   {
      // now that association may complain! 
      // but we can't do anything about it
      if( ! [_editingAssociation endEditing])
      {
         // assume association already used "failedToValidateValue:" or some such
         NSLog( @"Edit Error, previous edit will be lost");
         _editingAssociation = nil;         
      }
   }
   NSParameterAssert( ! _editingAssociation);
#if DEBUG
   NSLog( @"//////// %@ starts editing", association);
#endif
   _editingAssociation = association;
}


- (void) associationDidEndEditing:(EOAssociation *) association
{
   NSParameterAssert( [association isKindOfClass:[EOAssociation class]]);
   
#if DEBUG
   if( ! _editingAssociation)
   {
      NSLog( @"/******/ %@ ends editing though it never began", association);
      // could tell another association that its editing
      // days are over...
   }
   else
      NSLog( @"//////// %@ ends editing", association);
#endif
      
      NSAssert2( ! _editingAssociation || association == _editingAssociation, @"associationDidEndEditing: mismatch with associationDidBeginEditing (%@ vs. %@)", association, _editingAssociation);
      _editingAssociation = nil;
}


//
// collect keys, that are used by associatons that operate on all
// displayed objects e.g. EOTableViewAssociation
// The idea is to fetch for a partial fault here, NOT! for raw rows
// so we just need those keys that are shown over multiple rows and
// should not fire the fault
//
- (void) addBoundKeysForDisplayedObjectsToSet:(NSMutableSet *) set
{
   NSArray   *associations;
   
   associations = [self observingAssociations];
   [associations makeObjectsPerformSelector:_cmd
                                 withObject:set];
   
   /* walk through sortOrderings */
   [[self sortOrderings] makeObjectsPerformSelector:_cmd
                                         withObject:set];
}


- (EOAssociation *) editingAssociation
{
   return( _editingAssociation);
}

@end
