/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF
 *  Copyright (C) 2006 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 *
 */
#import "Defines.h"

#import "EODisplayGroup+InsertUpdateDelete.h"

#import "EODisplayGroup+Objects.h"
#import "EODisplayGroup+Private.h"
#import "EODisplayGroup+Selection.h"
#import "EODisplayGroup+Associations.h"
#import "NSArray+MulleFastRemove.h"
#import "NSArray+MulleCommonMembers.h"
#import "NSMutableArray+MulleFastRemove.h"
#import "NSObject+EOInterfaceKVCAdditions.h"


@implementation EODisplayGroup ( InsertUpdateDelete)

#if DO_UNDO  

static void   _registerUndoForDelete( EODisplayGroup *self, id obj, unsigned int anIndex);
static void   _registerUndoForInsert( EODisplayGroup *self, id newObject);

#endif


- (void) setInsertedObjectDefaultValues:(NSDictionary *) defaultValues
{
   id   old;
   
   NSParameterAssert( ! defaultValues|| [defaultValues isKindOfClass:[NSDictionary class]]);
   old = _insertedObjectDefaultValues;
   _insertedObjectDefaultValues = [defaultValues copyWithZone:_zone];
   [old release];
}


- (NSDictionary *) insertedObjectDefaultValues
{
   return( _insertedObjectDefaultValues);
}


//
// changed from void / internal implementation detail
//
static void   __insertObjectAtIndex( EODisplayGroup *self, id newObject, unsigned int anIndex)
{  
   //
   // now nothing can go wrong...
   //
   [self->_displayedObjects insertObject:newObject
                                 atIndex:anIndex];
   self->_nDisplayedObjects = [self->_displayedObjects count];

   self->_updatedObjectIndex = anIndex;
   displayedObjectsHaveChanged( self);
   
   [self->_allObjects addObject:newObject];	// sic!!
}


static BOOL   _insertObjectAtIndex( EODisplayGroup *self, id newObject, unsigned int anIndex)
{  
   if( self->_impDelegateDisplayGroupShouldInsertObjectAtIndex)
      if( ! self->_impDelegateDisplayGroupShouldInsertObjectAtIndex( self->_delegate, @selector( displayGroup:shouldInsertObject:atIndex:), self, newObject, anIndex))
         return( NO);

   if( anIndex > self->_nDisplayedObjects)
   {
      [NSException raise:NSInvalidArgumentException
                  format:@"Index %u out of bounds", anIndex];
   }

#if DO_UNDO   // undo insert
   _registerUndoForInsert( self, newObject);
#endif   
   [self->_dataSource insertObject:newObject];
   __insertObjectAtIndex( self, newObject, anIndex);
   
   if( self->_impDelegateDisplayGroupDidInsertObject)
      self->_impDelegateDisplayGroupDidInsertObject( self->_delegate, 
                                                     @selector( displayGroup:didInsertObject:), 
                                                     self, 
                                                     newObject);
   [self setSelectedObject:newObject];
   return( YES);
}


/* validation is only specified for setValue:forKey:
   the object may very well not be ready yet for insert
   into the database. In fact, the object will be modified
   by the EODetailDisplayGroup viz the masterObject 
   most likely. Therefore validation is not useful
*/
static BOOL   _insertObjectAtIndexWithValidation( EODisplayGroup *self, id newObject, unsigned int anIndex)
{
//   NSException  *exception;

#warning (nat) might want to use contexts message handler to present the error   
/*   
   if( exception = [newObject validateForInsert])  // a little late ?
      [exception raise];   // what' else ??
*/   
   return( _insertObjectAtIndex( self, newObject, anIndex));
}


static id   _insertNewObjectAtIndex( EODisplayGroup *self, unsigned int anIndex)
{
   id             obj;
   NSException   *exception;
   
   obj       = nil;
   exception = nil;
   
   if( anIndex > self->_nDisplayedObjects)
      [self->_displayedObjects objectAtIndex:anIndex]; // force exception

   obj = [self->_dataSource createObject];
   if( ! obj)
   {
      if( self->_impDelegateDisplayGroupCreateObjectFailedForDataSource)
      {
         (*self->_impDelegateDisplayGroupCreateObjectFailedForDataSource)( self->_delegate, @selector( displayGroup:createObjectFailedForDataSource:), self, self->_dataSource); 
         // delegate ignored it so... done here
         return( nil);
      }
         // dox says to do alert panel, we raise...
#warning (nat) or go through message handler of editing context
      [NSException raise:NSInternalInconsistencyException
                  format:@"Datasource %@ failed to create a new %@ object",
         self->_dataSource, [[self->_dataSource classDescriptionForObjects] entityName]];
   }
   [obj takeValuesFromDictionary:self->_insertedObjectDefaultValues];

   //
   // go to some pains to get rid of object
   // this is not compatible to Apple EO but
   // just makes more sense to me at the moment
   // [Assume that the delegate informed the
   // user, rather that it was a user choice]
   //
   if( ! _insertObjectAtIndex( self, obj, anIndex))
   {
      [self->_dataSource deleteObject:obj];
      obj = nil;
   }

   return( obj); 
}


- (BOOL) insertObject:(id) newObject
              atIndex:(unsigned int) anIndex
{
   BOOL   flag;

   NSParameterAssert( newObject != nil);
   
#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@ %u", self, __PRETTY_FUNCTION__, newObject, anIndex);
#endif
   enterCall( self);
NS_DURING 
   flag = _insertObjectAtIndexWithValidation( self, newObject, anIndex);

NS_HANDLER
#if DEBUG
   NSLog( @"******EXCEPTION: %@", localException);
#endif  
   if( _impDelegateDisplayGroupEditFailed)
   {
      NSDictionary   *userInfo;
      
      userInfo = [NSDictionary dictionaryWithObjectsAndKeys:EODisplayGroupOperationInsertValue, EODisplayGroupOperationKey, 
         [NSNumber numberWithUnsignedInt:anIndex], EODisplayGroupIndexKey,
         newObject, EODisplayGroupObjectKey,
         localException, EODisplayGroupExceptionKey,
         nil];
      
      (*_impDelegateDisplayGroupEditFailed)( _delegate, 
                                             @selector( displayGroup:editFailed:), 
                                             self, 
                                             userInfo);
   }
   flag = NO;
NS_ENDHANDLER

   leaveCall( self);
   return( flag);
}


- (id) insertObjectAtIndex:(unsigned int) anIndex
{
   id            obj;
   NSException   *exception;
   
   enterCall( self);
   
   obj       = nil;
   exception = nil;
NS_DURING   
   obj = _insertNewObjectAtIndex( self, anIndex);
NS_HANDLER
#if DEBUG
   NSLog( @"******EXCEPTION: %@", localException);
#endif  
   if( _impDelegateDisplayGroupEditFailed)
   {
      NSDictionary   *userInfo;
      
      userInfo = [NSDictionary dictionaryWithObjectsAndKeys:EODisplayGroupOperationInsertValue, EODisplayGroupOperationKey, 
         [NSNumber numberWithUnsignedInt:anIndex], EODisplayGroupIndexKey,
         localException, EODisplayGroupExceptionKey,
         nil];
      
      (*_impDelegateDisplayGroupEditFailed)( _delegate, 
                                             @selector( displayGroup:editFailed:), 
                                             self, 
                                             userInfo);
   }
   else
      exception = localException;
NS_ENDHANDLER   
   leaveCall( self);
   [exception raise];
   
   return( obj);
}


//
// Two helper functions to (a) memorize current selection
// and (b) reselect
//
static unsigned int   _suggestedSelectionIndex( EODisplayGroup *self)
{
   id   obj;
   
   if( ! self->_nSelectedObjects)
      return( 0);
   obj = [self->_selectedObjects objectAtIndex:0];
   return( [self->_displayedObjects indexOfObjectIdenticalTo:obj]);
}


static unsigned int   _adjustedSelectionIndex( EODisplayGroup *self, unsigned int index)
{
   unsigned int  count;
   
   count = self->_nDisplayedObjects;
   if( ! count)
      return( NSNotFound);
   if( index >= count)
      index = count - 1;
   return( index);
}


static NSArray   *delegateFilteredSelection( EODisplayGroup *self, NSArray *selection)
{
   unsigned int     i, n;
   NSMutableArray   *array;
   id               obj;
                
   if( ! self->_impDelegateDisplayGroupShouldDeleteObject)
      return( selection);

   n     = [selection count];
   array = [NSMutableArray array];
   for( i = 0; i < n; i++)
   {
      obj = [selection objectAtIndex:i];
      if( self->_impDelegateDisplayGroupShouldDeleteObject( self->_delegate, @selector( displayGroup:shouldDeleteObject:), self, obj))
         [array addObject:obj];
   }   
   return( array);
}


- (BOOL) _deleteObjects:(NSArray *) objects
{
   volatile NSArray          *array;
   volatile unsigned int     i, n;
   volatile BOOL             flag;
   volatile unsigned int     index;
   NSArray                   *deleted;
   NSArray                   *selection;
   NSMutableArray            *newSelection;
   NSException               *exception;
   id                        obj;
   
   index = _suggestedSelectionIndex( self);
   
   array = delegateFilteredSelection( self, objects);

   n = [array count];
   if( ! n)
      return( NO);
   
   flag     = YES;
   i         = 0;
   exception = nil;
NS_DURING   
   for( ;i < n; i++)
   {
      obj = [array objectAtIndex:i];
   
      // _delete.. can raise
      [[obj retain] autorelease];
      [self->_dataSource deleteObject:obj];
#if DO_UNDO   // undo delete
      _registerUndoForDelete( self, obj, i);
#endif   
      if( self->_impDelegateDisplayGroupDidDeleteObject)
         self->_impDelegateDisplayGroupDidDeleteObject( self->_delegate,
                                                        @selector( displayGroup:didDeleteObject:),
                                                        self,
                                                        obj);
   }
NS_HANDLER
#if DEBUG
   NSLog( @"******EXCEPTION: %@", localException);
#endif   
   flag = NO;
   exception = localException;
NS_ENDHANDLER


   if( ! flag && self->_impDelegateDisplayGroupEditFailed)
   {
      NSDictionary   *userInfo;
      
      userInfo = [NSDictionary dictionaryWithObjectsAndKeys:EODisplayGroupOperationDeleteValue, EODisplayGroupOperationKey, 
         obj, EODisplayGroupObjectKey,
         exception, EODisplayGroupExceptionKey,
         nil];

NS_DURING      
      (*self->_impDelegateDisplayGroupEditFailed)( _delegate, 
                                                   @selector( displayGroup:editFailed:), 
                                                   self, 
                                                   userInfo);
NS_HANDLER
#if DEBUG
      NSLog( @"******EXCEPTION: %@", localException);  // well we say NO anyhow
#endif   
NS_ENDHANDLER
   }

   if( ! i)
      return( flag);

   deleted = [array subarrayWithRange:NSMakeRange( 0, i)];

   [self->_displayedObjects mulleRemoveObjectsInArray:deleted];
   self->_nDisplayedObjects = [self->_displayedObjects count];
   displayedObjectsHaveChanged( self);

   [self->_allObjects mulleRemoveObjectsInArray:deleted];

   selection = [self selectedObjects];
   if( [selection mulleHasCommonMembersWithArray:deleted])
   {
      newSelection = [NSMutableArray arrayWithArray:[selection mulleArrayByRemovingObjectsInArray:deleted]];
      if( ! [newSelection count])
      {
         index = _adjustedSelectionIndex( self, index);
         if( index != NSNotFound)
            [newSelection addObject:[self->_displayedObjects objectAtIndex:index]];
      }
      
      [self _setSelectedObjects:newSelection
                       verified:YES];
   }
   
   return( YES);
}


// 
// there is a major conceptual problem here. the display group will not
// notice if the object is undone and reappears (possibly under a different
// guise) because this is done by the editing context and will not affect
// the editors or f.e. the database datasource (which must not refetch)
//
- (BOOL) deleteSelection
{
   NSArray          *selection;
   BOOL             flag;
   
   if( ! _nSelectedObjects)
      return( NO); // or yes ??
   
   enterCall( self);
   
   selection = [NSMutableArray arrayWithArray:_selectedObjects]; // keep a copy
   flag      = [self _deleteObjects:selection];

   leaveCall( self);
   return( flag);
}


- (BOOL) deleteObjectAtIndex:(unsigned int) anIndex
{
   BOOL   flag;
   id     obj;
   
   if( anIndex >= self->_nDisplayedObjects)
      return( NO);
   
   enterCall( self);
   
   obj  = [[self displayedObjects] objectAtIndex:anIndex];
   flag = [self _deleteObjects:[NSArray arrayWithObject:obj]];
   
   leaveCall( self);
   return( flag);
}


static id   _valueForObjectAndKey( EODisplayGroup *self, id obj, NSString *key)
{
   id   value;
   
   if( self->_flags.queryMode)
      value = [obj valueForKey:key];
   else
      value = [obj valueForKeyPath:key];
   return( value);
}


- (id) valueForObject:(id) obj
                  key:(NSString *) key
{
   id   value;
   
   value = _valueForObjectAndKey( self, obj, key);
   return( value); // ahemm. is this supposed to be more complicated ??
}


- (id) selectedObjectValueForKey:(NSString *) key
{
   NSParameterAssert( [key isKindOfClass:[NSString class]]);
   id    obj;
   id    value;
   
   obj   = [self selectedObject];
   value = _valueForObjectAndKey( self, obj, key);
   return( value);
}


- (id) valueForObjectAtIndex:(unsigned int)  index
                         key:(NSString *) key
{
   id   obj;
   id   value;
   NSParameterAssert( [key isKindOfClass:[NSString class]]);
   
   if( index >= _nDisplayedObjects)
      return( NO);
   obj   = [_displayedObjects objectAtIndex:index];
   value = _valueForObjectAndKey( self, obj, key);
   
   return( value);
}


static NSException   *_setValueForObjectKey( EODisplayGroup *self, id value, id obj, NSString *key)
{
   NSException    *exception;
   unsigned int   index;
   
#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@ %u %@", self, __PRETTY_FUNCTION__, value, obj, key);
#endif   
   
#if STRICT_CHECKING   
   NSCAssert2( [obj editingContext], @"Object %@ is not registered in editingContext %@", obj, [self->_dataSource editingContext]);
   NSCParameterAssert( [obj editingContext] == [self->_dataSource editingContext]);
   NSCParameterAssert( ! [value editingContext] || [value editingContext] == [obj editingContext] || ([obj isKindOfClass:[NSMutableDictionary class]] && self->_flags.queryMode));
#endif
   
   // or throw exception ?
   if( self->_flags.queryMode)
   {
      [obj takeValue:value
              forKey:key];  // NOT key path, dictionary takes full keys
      displayedObjectsHaveChanged( self);
      return( nil);
   }
   
   index = [self->_displayedObjects indexOfObjectIdenticalTo:obj];
   if( index == NSNotFound)
   {
      exception = [NSException exceptionWithName:NSInternalInconsistencyException
                                          reason:@"Object is not a member of displayedObjects"                                          
                                        userInfo:nil];
      return( exception);
   }
   
   // we always validate. The "validateImmediately" just means
   // if we tell the user
   exception = [obj validateValue:&value
                           forKey:key];
   if( exception)
      return( exception);
   
   self->_updatedObjectIndex = index;
   
   /*
    * We need to be nice here. If the value is an EO and the key is (therefore)
    * a relationship-
    * If the relationship is toOne, we really want to add it to bothsides of
    * the relationship (or take it out). Let's play nice for EOPopupButtonAssociation...
    */
   [obj eoTakeValue:value
         forKeyPath:key];
   
   displayedObjectsHaveChanged( self);
   
   if( self->_impDelegateDisplayGroupDidSetValueForObjectKey)
      self->_impDelegateDisplayGroupDidSetValueForObjectKey( self->_delegate, @selector( displayGroup:didSetValue:forObject:key:), self, value, obj, key);
   return( nil);
}


#warning (nat) change this interface to return an exception
- (BOOL) setValue:(id) value
        forObject:(id) obj
              key:(NSString *) key
{
   NSException   *exception;
   
   NSParameterAssert( [key isKindOfClass:[NSString class]]);
   
   enterCall( self);
   exception = nil;
NS_DURING   
   exception = _setValueForObjectKey( self, value, obj, key);
NS_HANDLER
   NSLog( @"****** %@ %s %@", self, __PRETTY_FUNCTION__, localException);
   exception = localException;
NS_ENDHANDLER   
   if( exception)
   {
      [obj willChange];  // force a redraw... clumsy
      if( _flags.validateImmediately)
      {
         if( _impDelegateDisplayGroupEditFailed)
         {
            NSDictionary   *userInfo;
         
            userInfo = [NSDictionary dictionaryWithObjectsAndKeys:EODisplayGroupOperationUpdateValue, EODisplayGroupOperationKey, 
               obj, EODisplayGroupObjectKey,
               value, EODisplayGroupValueKey,
               key, EODisplayGroupKeyKey,
               exception, EODisplayGroupExceptionKey,
               nil];
         
         // if it were to return "YES", this would just confuse the
         // associations.
            (*_impDelegateDisplayGroupEditFailed)( _delegate, 
                                                   @selector( displayGroup:editFailed:), 
                                                   self, 
                                                   userInfo);
         }
      }
   }
   leaveCall( self);
   return( exception == nil);
}


- (BOOL) setSelectedObjectValue:value
                         forKey:(NSString *) key
{
   id     obj;
   BOOL   flag;
   
   NSParameterAssert( [key isKindOfClass:[NSString class]]);
   
   enterCall( self);
   flag = NO;
   obj = [self selectedObject];
   if( obj)
   {
      flag = [self setValue:value      	// stay doc compatible
                  forObject:obj	// [though we now know that
                        key:key];	// object is valid....]
   }
   leaveCall( self);
   return( flag);
}



- (BOOL) setValue:(id) value
 forObjectAtIndex:(unsigned int)  index
              key:(NSString *) key
{
   id   obj;
   BOOL   flag;
   
   NSParameterAssert( [key isKindOfClass:[NSString class]]);
   
   enterCall( self);
   flag = NO;
   
   if( index < _nDisplayedObjects)
   {
      obj = [_displayedObjects objectAtIndex:index];
      flag = [self setValue:value       // stay doc compatible
                  forObject:obj	// [though we now know that
                        key:key];       // object is valid....]
   }
   leaveCall( self);
   return( flag);
}


- (int) updatedObjectIndex
{
   return( _updatedObjectIndex);
}


#if DO_UNDO  
//
// UNDO stuff
//
- (void) _undoDeleteByInsertingObjectWithGlobalID:(EOGlobalID *) gid
                                     closeToIndex:(unsigned int) index
                                      wasSelected:(int) info
{
   EOEditingContext  *context;
   id                obj;
   
#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@ %u %d", self, __PRETTY_FUNCTION__, obj, index, info);
#endif   
   enterCall( self);
   
   context = [_dataSource editingContext];
   if( [context respondsToSelector:@selector( _mappedObjectForGlobalID:)])
      obj = [context _mappedObjectForGlobalID:gid];
   else
      obj = [[_dataSource editingContext] objectForGlobalID:gid];
   
   if( ! obj)
   {
#if DEBUG      
      NSLog( @"Failed to find an object for GlobalID %@", gid);
#endif      
      return;
   }
   
   if( index > _nDisplayedObjects)
      index = _nDisplayedObjects;
   
   //
   // but don't let the datasource insert it also
   // this is stupid, but the editingcontext should have already added
   // the object back into the relationship. so there's nothing to do 
   // on the datasource level
   //
   __insertObjectAtIndex( self, obj, index);
   if( info == 1)
      [self setSelectedObject:obj];
   if( info == 2)
   {
      [_selectedObjects addObject:obj];
      _nSelectedObjects = [_selectedObjects count];
      selectionHasChanged( self);
      invalidateSelectedIndexes( self);
   }
   
   leaveCall( self);
}


//
// this could become a bottleneck, if a few thousand objects are deleted
//
static void   _registerUndoForDelete( EODisplayGroup *self, id obj, unsigned int anIndex)
{
   EOEditingContext  *context;
   EOGlobalID        *gid;
   NSUndoManager     *undo;
   int               info;
   
   if( [self->_dataSource isKindOfClass:[EODetailDataSource class]])
      return;

   context = [self->_dataSource editingContext];
   undo    = [context undoManager];
   info    = 0;
   if( self->_nSelectedObjects)
   {
      if( [self->_selectedObjects indexOfObjectIdenticalTo:obj] != NSNotFound)
         info = self->_nSelectedObjects == 1 ? 1 : 2;
   }
   
   gid = [context globalIDForObject:obj];
   NSCParameterAssert( [gid isKindOfClass:[EOGlobalID class]]);
   
   [[undo prepareWithInvocationTarget:self]
     _undoDeleteByInsertingObjectWithGlobalID:gid
                                 closeToIndex:anIndex
                                  wasSelected:info];
}


- (void) _undoInsertByDeletingObjectWithGlobalID:(EOGlobalID *) gid
{
   id   obj;
   
#if DEBUG_VERBOSE
   NSLog( @"%@ %s %@", self, __PRETTY_FUNCTION__, obj);
#endif   
   obj = [[_dataSource editingContext] objectForGlobalID:gid];
   if( ! obj)
   {
      NSLog( @"Failed to find an object for GlobalID %@", gid);
      return;
   }
   
   enterCall( self);
   [self _deleteObjects:[NSArray arrayWithObject:obj]];
   leaveCall( self);
}


static void   _registerUndoForInsert( EODisplayGroup *self, id newObject)
{
   EOEditingContext  *context;
   NSUndoManager     *undo;
   EOGlobalID        *gid;
   
   if( [self->_dataSource isKindOfClass:[EODetailDataSource class]])
      return;
   
   context = [self->_dataSource editingContext];
   undo    = [context undoManager];
   gid     = [context globalIDForObject:newObject];
   if( ! gid)
   {
      NSLog( @"%@: object %@ not known in EOEditingContext so, no undo in DG", self, newObject);
      return;
   }
   
   [[undo prepareWithInvocationTarget:self]
         _undoInsertByDeletingObjectWithGlobalID:gid];
}
#endif

@end
