/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF
 *  Copyright (C) 2006 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 *
 */
#import "Defines.h"

#import "EODisplayGroup+Objects.h"

#import "EODisplayGroup+Associations.h"
#import "EODisplayGroup+Private.h"
#import "EODisplayGroup+Selection.h"


@implementation EODisplayGroup ( Objects)

- (NSArray *) allObjects
{
   return( [[_allObjects retain] autorelease]);
}


- (NSArray *) displayedObjects
{
   return( [[_displayedObjects retain] autorelease]);
}


static BOOL   _setDisplayedObjectsArrayAlreadyRetained( EODisplayGroup *self, NSMutableArray *objects)
{
   id     old;
   void   (*f)( id, SEL, id);
   
   if( self->_flags._shutdown)
      return( NO);
   
   if( [self->_displayedObjects isEqualToArray:objects])
   {
      [objects autorelease];
      return( NO);
   }
   
   old = self->_displayedObjects;
   self->_displayedObjects  = objects;
   self->_nDisplayedObjects = [objects count];
   [old release];

   //
   // if the displayedObjects have changed we need to invalidate the indexes
   // and the selection will have changed. The selected objects themselves
   // may not have changed, but that's not important
   //
   invalidateSelectedIndexes( self);
   selectionHasChanged( self);
   
   displayedObjectsHaveChanged( self);
   
   if( self->_nSelectedObjects)
   {
      f = (void *) self->_impDelegateDisplayGroupDidChangeSelectedObjects;
      if( f)
         (*f)( self->_delegate, 
               @selector( displayGroupDidChangeSelectedObjects:), 
               self);
   }
   return( YES);
}


static BOOL   setDisplayedObjectsAlreadyRetained( EODisplayGroup *self,
                                                  NSMutableArray *objects,
                                                  BOOL updateSelection)
{
   NSMutableArray   *selection;
   BOOL             flag;
   
   flag = _setDisplayedObjectsArrayAlreadyRetained( self, objects);
   if( ! flag)
      return( YES);
   
   if( ! updateSelection)
      return( YES);
   
   selection = [self _keepableSelectionForObjects:objects];
   if( self->_nSelectedObjects == [selection count])
      return( YES);
   
   [self _setSelectedObjects:selection
                    verified:YES];
   return( NO);
}


- (BOOL) _updateDisplayedObjects:(BOOL) updateSelection
{
   NSMutableArray  *objects;
   NSArray         *array;
   SEL             sel;
   BOOL            flag;
   
   if( ! self->_flags.queryMode)
   {
      if( self->_impDelegateDisplayGroupDisplayArrayForObjects)
      {
         sel   = @selector( displayGroup:displayArrayForObjects:);
         array = self->_impDelegateDisplayGroupDisplayArrayForObjects( self->_delegate, 
                                                                       sel, 
                                                                       self, 
                                                                       self->_allObjects);
         objects = [[NSMutableArray allocWithZone:self->_zone] initWithArray:array];
      }
      else
      {
         if( self->_qualifier)
         {
            array   = [self->_allObjects filteredArrayUsingQualifier:self->_qualifier];
            objects = [[NSMutableArray allocWithZone:self->_zone] initWithArray:array];
         }
         else
            objects = [[NSMutableArray allocWithZone:self->_zone] initWithArray:self->_allObjects];
         
         if( [self->_sortOrdering count])
            [objects sortUsingKeyOrderArray:self->_sortOrdering];
      }
   } 
   else
      objects = [[NSMutableArray allocWithZone:self->_zone] initWithArray:self->_allObjects];


   flag = setDisplayedObjectsAlreadyRetained( self, objects, updateSelection);
   return( flag);
}


- (void) updateDisplayedObjects
{
   enterCall( self);
   [self _updateDisplayedObjects:YES];
   leaveCall( self);
}


static void   _setObjectArray( EODisplayGroup *self, NSArray *objects)
{
   id   old;
   
   old = self->_allObjects;
   self->_allObjects = [[NSMutableArray allocWithZone:NULL] initWithArray:objects];
   self->_allObjectsChangeCount++;
   [old release];
}


//
// shared by -fetch and -setObjectArray
//
- (void) _setObjectArray:(NSArray *) objects
{
   NSArray        *displayed;
   unsigned int   n;
   id             obj;
   BOOL           selectFirst;
   
   if( self->_flags._shutdown)
      return;
   
   //
   // if nothing's changed, we do nothing, 
   // otherwise..
   // It's hard if not impossible to avoid the
   // fault of all the objects right now
   // If there is a delegate, then we need to run through updateDisplayed
   // because the delegate may choose to do strange stuff (we do..)
   //
   if( self->_impDelegateDisplayGroupDisplayArrayForObjects ||
       (self->_allObjects != objects && ([EOFault isFault:objects] || ! [self->_allObjects isEqualToArray:objects])))
   {
      n = self->_nSelectedObjects;
      _setObjectArray( self, objects);

      selectFirst = ! n && self->_flags.selectsFirstObjectAfterFetch;

      //
      // do not destroy old selection if possible
      // first try nicely, if selection is present
      [self _updateDisplayedObjects:! selectFirst];
   }
   
   //
   // then be hard on it, if we must (otherwise no selection)
   //
   if( ! self->_nSelectedObjects && self->_flags.selectsFirstObjectAfterFetch)
   {
      if( self->_nDisplayedObjects > 0)
      {
         displayed = [self displayedObjects];
         obj       = [displayed objectAtIndex:0];
         [self selectObject:obj];
      }
   }
   
   NSParameterAssert( ! self->_flags.selectsFirstObjectAfterFetch || 
                      self->_nSelectedObjects || 
                      ! self->_nDisplayedObjects);
}


- (void) setObjectArray:(NSArray *) objects
{
   NSParameterAssert( ! objects|| [objects isKindOfClass:[NSArray class]]);
#if DEBUG_VERBOSE
   NSLog( @"%s %@ %d objects", __PRETTY_FUNCTION__, self, [objects count]);
#endif   
   enterCall( self);
   
   [self endEditingOrRaise];

   [self _setObjectArray:objects];

   leaveCall( self);
}

@end
