/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF
 *  Copyright (C) 2006 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 *
 */
#import "EODisplayGroup.h"


@interface EODisplayGroup ( Private)

- (void) _assertSanity;

@end

#ifdef SELF
#undef SELF
#endif
#define SELF   ((struct { @defs( EODisplayGroup) } *) self)


#if DEBUG_FLUSH

#define enterCall( self)   _enterCall( self, (char *) __PRETTY_FUNCTION__)
#define resetCall( self)   _resetCall( self, (char *) __PRETTY_FUNCTION__)
#define resetCallAndCheck( self)   _resetCallAndCheck( self, (char *) __PRETTY_FUNCTION__)
#define undoCall( self)    _undoCall( self, (char *) __PRETTY_FUNCTION__)
#define leaveCall( self)   _leaveCall( self, (char *) __PRETTY_FUNCTION__)

static inline void   _enterCall( EODisplayGroup *self, char *caller)
{
   NSLog( @"enterCall( %@) flush was %d before (now +1) (called from %s)", self, self->_flushCount, caller);
   ++self->_flushCount;
}


static inline void   _undoCall( EODisplayGroup *self, char *caller)
{
   NSLog( @"undoCall(  %@) flush %d before (now -1) (called from %s)", self, self->_flushCount, caller);
   NSCAssert( self->_flushCount, @"Flush count decremented too much");
   --self->_flushCount;
}


static inline void   _leaveCall( EODisplayGroup *self, char *caller)
{
   NSLog( @"leaveCall( %@) flush was %d before (now -1) (called from %s)", self, self->_flushCount, caller);
   NSCAssert1( self->_flushCount, @"Flush count decremented too much: %d", self->_flushCount);
   if( ! --self->_flushCount)
      [self willChange];
}


static inline void   _resetCall( EODisplayGroup *self, char *caller)
{
   NSLog( @"resetCall(  %@) flush was %d before (now 0) (called from %s)", self, self->_flushCount, caller);
   self->_flushCount = 0;
}

static inline void   _resetCallAndCheck( EODisplayGroup *self, char *caller)
{
   NSCAssert1( self->_flushCount == 0, @"Flush count %d wasn't decremented properly", self->_flushCount);
   NSLog( @"resetCall(  %@) flush was %d before (now 0) (called from %s)", self, self->_flushCount, caller);
   self->_flushCount = 0;
}


#else

static inline void   enterCall( EODisplayGroup *self)
{
   ++SELF->_flushCount;
#if DEBUG   
   [self _assertSanity];
#endif
}


static inline void   undoCall( EODisplayGroup *self)
{
#if DEBUG   
   [self _assertSanity];
#endif
   --SELF->_flushCount;
}


static inline void   leaveCall( EODisplayGroup *self)
{
   extern void   mulle_eo_display_group_assert_sanity();

#if DEBUG   
   [self _assertSanity];
#endif
   if( ! --SELF->_flushCount)
      [self willChange];
}


static inline void   resetCall( EODisplayGroup *self)
{
   SELF->_flushCount = 0;
}


static inline void   resetCallAndCheck( EODisplayGroup *self)
{
   resetCall( self);
}


#endif


static inline void   displayedObjectsHaveChanged( EODisplayGroup  *self)
{
   SELF->_flags.justChangedContents = YES;
}


static inline void   setSelectedIndexes( EODisplayGroup  *self, NSArray *indexes)
{
   NSCParameterAssert( SELF->_selection != indexes);
   
   [SELF->_selection release];
   SELF->_selection = [indexes retain];
}


static inline void   invalidateSelectedIndexes( EODisplayGroup  *self)
{
   [SELF->_selection release];
   SELF->_selection = nil;
}


static inline void   selectionHasChanged( EODisplayGroup  *self)
{
   SELF->_flags.justChangedSelection = YES;
}



