/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF
 *  Copyright (C) 2006 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 *
 */
#import "Defines.h"

#import "EODisplayGroup+Selection.h"

#import "EODisplayGroup+Associations.h"
#import "EODisplayGroup+Private.h"


@implementation EODisplayGroup ( Selection)

static BOOL   _setSelectionIndexes( EODisplayGroup *self, NSArray *selection);

- (NSMutableArray *) _keepableSelectionForObjects:(NSArray *) objects
{
   NSMutableArray   *array;
   unsigned int     i, n;
   id               obj;
   
   n = self->_nSelectedObjects;

   array = [[[NSMutableArray allocWithZone:self->_zone] initWithCapacity:n] autorelease];
   
   //
   // could set up indices here as well, but lets
   // assume lazy is better more often
   //
   for( i = 0; i < n; i++)
   {
      obj = [self->_selectedObjects objectAtIndex:i];
      if( [objects indexOfObjectIdenticalTo:obj] != NSNotFound)
         [array addObject:obj];
   }
   return( array);
}


- (NSArray *) selectedObjects
{
   return( [[_selectedObjects retain] autorelease]);
}


//
// premature optimization for this
//
#warning (nat) make selectionObjects lazy as well

static int  uint_compare( unsigned int *a, unsigned int *b)
{
   return( *a - *b);
}


- (unsigned int *) _selectionIndexes:(unsigned int *) len
                          forObjects:(NSArray *) objects
{
   unsigned int    i, n;
   unsigned int    j;
   unsigned int    index;
   IMP             IMPindexOfObjectIdenticalTo;
   IMP             IMPobjectAtIndex;
   id              obj;
   size_t          bytes;
   unsigned int    *buf;
   
   n  = [objects count];
   if( ! n)
   {
      *len = n;
      return( NULL);
   }
   
   bytes = sizeof( unsigned int) * n;
   buf   = (unsigned int *) malloc( bytes);
   if( ! buf)
      [NSException raise:NSMallocException
                  format:@"Failed to allocate %ld bytes", bytes];
   
   IMPindexOfObjectIdenticalTo = [_displayedObjects methodForSelector:@selector( indexOfObjectIdenticalTo:)];
   IMPobjectAtIndex            = [_selectedObjects methodForSelector:@selector( objectAtIndex:)];
   
   for( j = i = 0; i < n; i++)
   {
      obj   = IMPobjectAtIndex( objects, @selector( objectAtIndex:), i);
      index = (unsigned int) IMPindexOfObjectIdenticalTo( _displayedObjects, 
                                                          @selector( indexOfObjectIndenticalTo:), 
                                                          obj);
      if( index != NSNotFound)
         buf[ j++] = index;
   }
   
   // have real indexes sorted
   qsort( buf, j, sizeof( unsigned int), (void *) uint_compare);
   
   *len = j; 
   return( buf);
}


- (NSArray *) _selectionIndexesForObjects:(NSArray *) objects
{
   unsigned int    i, n;
   IMP             IMPallocWithZone;
   IMP             IMPinitWithUnsignedInt;
   IMP             IMPaddObject;
   NSNumber        *number;
   Class           numberClass;
   NSMutableArray  *array;
   unsigned int    *buf;
   
   buf = [self _selectionIndexes:&n
                      forObjects:objects];

   array = [[[NSMutableArray alloc] initWithCapacity:n] autorelease];
   if( ! buf)
      return( array);
   
   numberClass = [NSNumber class];
   
   IMPaddObject = [array methodForSelector:@selector( addObject:)];
   
   IMPallocWithZone       = [NSNumber methodForSelector:@selector( allocWithZone:)];
   number                 = IMPallocWithZone( numberClass, @selector( allocWithZone:), _zone);
   IMPinitWithUnsignedInt = [number methodForSelector:@selector( initWithUnsignedInt:)];
   
   for( i = 0; i < n; i++)
   {
      number = IMPinitWithUnsignedInt( number, @selector( initWithUnsignedInt:), buf[ i]);
      (*IMPaddObject)( array, @selector( addObject:), number);
      [number release];
      number = IMPallocWithZone( numberClass, @selector( allocWithZone:), _zone);
   }
   [number release];
   
   return( array);
}


- (NSArray *) selectionIndexes
{
   NSArray   *array;

   
   if( _selection)
      return( [[_selection retain] autorelease]);

   array      = [self _selectionIndexesForObjects:[self selectedObjects]];
   _selection = [array retain];

   return( array);
}


- (BOOL) _setSelectedObjects:(NSMutableArray *) objects
                    verified:(BOOL) verified
{
   id             old;
   NSArray        *oldIndexes;
   NSArray        *newIndexes;
   BOOL           objectsSame;
   BOOL           selectionSame;
   unsigned int   i, n;
   id             obj;
   
#if DEBUG_VERBOSE
   NSLog( @"%s %@ %@", __PRETTY_FUNCTION__, self, objects);
#endif   
   
   // identical ? nothing changed
   objectsSame = [self->_selectedObjects isEqualToArray:objects];
   if( ! objectsSame && ! verified)
   {
      n = [objects count];
      for( i = 0; i < n; i++)
      {
         obj = [objects objectAtIndex:i];
         if( [_displayedObjects indexOfObjectIdenticalTo:obj] == NSNotFound)
            return( NO);
      }
   }
   
   oldIndexes = [self selectionIndexes];
   newIndexes = [self _selectionIndexesForObjects:objects];
   
   selectionSame = [oldIndexes isEqualToArray:newIndexes];
   if( selectionSame)
   {
      if( objectsSame)
         return( YES);
   }
   else
      if( self->_impDelegateDisplayGroupShouldChangeSelectionToIndexes)
         if( ! self->_impDelegateDisplayGroupShouldChangeSelectionToIndexes( self->_delegate, @selector( displayGroup:shouldChangeSelectionToIndexes:), self,  newIndexes))
            return( NO);
   
   selectionHasChanged( self);  
   if( ! objectsSame)
   {
      old = self->_selectedObjects;
      self->_selectedObjects  = [objects retain];  // we know its private!
      self->_nSelectedObjects = [objects count];
      [old release];
   }
   
   setSelectedIndexes( self, newIndexes);
   
   //
   // hackish, we can't invalidate the changesFlag yet, because we
   // do not know! We need the selection for that 
   //
   //
   // this is called, when the selection indices have changed
   //
   if( ! selectionSame || ! objectsSame)  // seems to be more compatible
      if( self->_impDelegateDisplayGroupDidChangeSelection)
         (*self->_impDelegateDisplayGroupDidChangeSelection)( self->_delegate, 
                                                              @selector( displayGroupDidChangeSelection:), 
                                                              self);
   //
   // this is called, when the selection indices stays the same
   // but the selected objects have changed
   //
   if( ! objectsSame)
      if( self->_impDelegateDisplayGroupDidChangeSelectedObjects)
         (*self->_impDelegateDisplayGroupDidChangeSelectedObjects)( self->_delegate, 
                                                                    @selector( displayGroupDidChangeSelectedObjects:), 
                                                                    self);
   return( YES);
}


static BOOL   _setSelectionIndexes( EODisplayGroup *self, NSArray *selection)
{
   NSMutableArray  *array;
   unsigned int    i, n, max;
   unsigned int    index;
   id              obj;
   BOOL            flag;
   
   if( self->_flags._shutdown)
      return( NO);
   
   selection = [selection sortedArrayUsingSelector:@selector( compare:)];
   if( [selection isEqualToArray:[self selectionIndexes]])
      return( YES);
   
   //
   // we always set selectionHasChanged: the reason is, that even if in the
   // end we don't, we still want the UI to be in sync and reflect our state
   //
   selectionHasChanged( self);  

   n     = [selection count];
   max   = self->_nDisplayedObjects;
   array = [[[NSMutableArray alloc] initWithCapacity:n] autorelease];
   
   for( i = 0; i < n; i++)
   {
      index = [[selection objectAtIndex:i] unsignedIntValue];
      if( index >= max)
      {
         [array release];
         return( NO);
      }
      obj = [self->_displayedObjects objectAtIndex:index];
      [array addObject:obj];
   }
   
   if( self->_impDelegateDisplayGroupShouldChangeSelectionToIndexes)
      if( ! self->_impDelegateDisplayGroupShouldChangeSelectionToIndexes( self->_delegate, @selector( displayGroup:shouldChangeSelectionToIndexes:), self, selection))
         return( NO);
   
   flag = [self _setSelectedObjects:array
                           verified:YES];
   
   return( flag);
}


- (BOOL) setSelectionIndexes:(NSArray *) indexes
{
   BOOL   flag;

   NSParameterAssert( ! indexes|| [indexes isKindOfClass:[NSArray class]]);

   if( ! [self endEditing])
      return( NO);
   
   enterCall( self);
   flag = _setSelectionIndexes( self, indexes);
   leaveCall( self);
   return( flag);
}


- (id) selectedObject
{
   if( ! _nSelectedObjects)
      return( nil);
   return( [_selectedObjects objectAtIndex:0]);
}


- (BOOL) selectObject:(id) obj
{
   NSMutableArray   *array;
   BOOL             flag;
   
   enterCall( self);
   if( [_displayedObjects indexOfObjectIdenticalTo:obj] == NSNotFound)
   {
      [self clearSelection];
      leaveCall( self);
      return( NO);
   }
   
   array = [[[NSMutableArray alloc] initWithObjects:obj, nil] autorelease];
   flag  = [self _setSelectedObjects:array
                            verified:YES];

   leaveCall( self);
   return( flag);
}


- (BOOL) selectObjectsIdenticalTo:(NSArray *) objectSelection
{
   NSMutableArray  *array;
   BOOL            flag;
   
   NSParameterAssert( ! objectSelection|| [objectSelection isKindOfClass:[NSArray class]]);

   enterCall( self);

   array = [[[NSMutableArray allocWithZone:_zone] initWithArray:objectSelection] autorelease];
   flag  = [self _setSelectedObjects:array
                            verified:NO];
   
   leaveCall( self);
   return( flag);
}


- (BOOL) selectObjectsIdenticalTo:(NSArray *) objectSelection
             selectFirstOnNoMatch:(BOOL) selectFirstOnNoMatch
{
   BOOL  flag;

   NSParameterAssert( ! objectSelection|| [objectSelection isKindOfClass:[NSArray class]]);

   enterCall( self);
   flag = [self selectObjectsIdenticalTo:objectSelection];
   if( flag && _nSelectedObjects == 0)
   {
      if( _nDisplayedObjects)
         [self setSelectedObject:[_displayedObjects objectAtIndex:0]];
      flag = NO;
   }
   leaveCall( self);
   return( YES);
}


static BOOL   _selectNext( EODisplayGroup *self)
{
   unsigned int  n;
   unsigned int  nSelected;
   id            obj;
   unsigned int  index;

   n = self->_nDisplayedObjects;
   if( ! n)
      return( NO);
   
   nSelected = self->_nSelectedObjects;
   if( ! nSelected)
   {
      [self setSelectedObject:[self->_displayedObjects objectAtIndex:0]];
      return( YES);
   }
   
   obj    = [self->_selectedObjects objectAtIndex:0];
   index  = [self->_displayedObjects indexOfObjectIdenticalTo:obj];
   NSCAssert( index != NSNotFound, @"Selection and Displayes Objects outta sync");
   ++index;
   if( index == n)
      index = 0;
   [self setSelectedObject:[self->_displayedObjects objectAtIndex:index]];
   return( YES);
}


- (BOOL) selectNext
{
   BOOL  flag;
   
   enterCall( self);
   flag = _selectNext( self);
   leaveCall( self);
   return( flag);
}


/*
 * If there are no objects, does nothing and returns NO.
 * If there's no selection, selects the object at index zero and returns YES.
 * If the first selected object is at index zero, selects the last object and returns YES.
 * Otherwise selects the object before the first selected object.
 */
static BOOL   _selectPrevious( EODisplayGroup *self)
{
   unsigned int  n;
   unsigned int  nSelected;
   id            obj;
   unsigned int  index;

   n = self->_nDisplayedObjects;
   if( ! n)
      return( NO);
   nSelected = self->_nSelectedObjects;
   if( ! nSelected)
   {
      [self setSelectedObject:[self->_displayedObjects objectAtIndex:0]];
      return( YES);
   }
   
   obj    = [self->_selectedObjects objectAtIndex:0];
   index  = [self->_displayedObjects indexOfObjectIdenticalTo:obj];
   NSCAssert( index != NSNotFound, @"Selection and Displayes Objects outta sync");
   if( index == 0)
      index = n;
   --index;
   [self setSelectedObject:[self->_displayedObjects objectAtIndex:index]];
   return( YES);
}


- (BOOL) selectPrevious
{
   BOOL  flag;

   enterCall( self);
   flag = _selectPrevious( self);
   leaveCall( self);
   
   return( flag);
}


- (BOOL) clearSelection
{

   BOOL   flag;
   
   enterCall( self);
   flag = [self _setSelectedObjects:[NSMutableArray array]
                           verified:YES];
   leaveCall( self);
   
   return( flag);
}


/*
 * WHERE DOES THIS API COME FROM ????
 */
- (void) setSelectedObject:(id) anObject
{
   NSMutableArray   *array;
   
   enterCall( self);
   
   array = [[[NSMutableArray alloc] initWithObjects:anObject, nil] autorelease];
   [self _setSelectedObjects:array
                    verified:NO];
   
   leaveCall( self);
}


/*
 * WHERE DOES THIS API COME FROM ????
 */
- (void) setSelectedObjects:(NSArray *) objects
{
   NSMutableArray   *array;
   
   enterCall( self);
   array = [[[NSMutableArray allocWithZone:_zone] initWithArray:objects] autorelease];
   [self _setSelectedObjects:array
                    verified:NO];
   leaveCall( self);
}


@end
