/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF
 *  Copyright (C) 2002,2005 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.2  2002/04/22 21:56:29  nat
 * EODisplayGroup mit code zugemuellt
 *
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#ifndef _EODISPLAYGROUP_H
#define _EODISPLAYGROUP_H

#import <Foundation/Foundation.h>
#import "MulleEOInterfaceParentIncludes.h"
#import <AppKit/NSNibDeclarations.h>

@class EOAssociation;


extern NSString  *EODisplayGroupException; // = @"EODisplayGroupException";
extern NSString  *EODisplayGroupLikeMatchOperator;
extern NSString  *EODisplayGroupCaseInsensitiveLikeMatchOperator;
extern NSString  *EODisplayGroupShutdownNotification;  
extern NSString  *EODisplayGroupWillFetchNotification;

//
// some public variables defined for the benefit of some associations
//
@interface EODisplayGroup : EODelayedObserver <NSCoding>
{
    EODataSource          *_dataSource;
    NSMutableArray        *_allObjects;

@private    
    NSMutableArray        *_displayedObjects;

@protected    
    id 		          _delegate;
    NSArray               *_selection;

    NSArray               *_sortOrdering;
    EOQualifier           *_qualifier;
    NSArray               *_localKeys;  // For IB inspector
@private    
    NSMutableArray        *_selectedObjects;

@protected    
    id                    _observerNotificationBeginProxy;
    id                    _observerNotificationEndProxy;
    int                   _updatedObjectIndex;
    NSDictionary          *_insertedObjectDefaultValues;
    NSMutableArray        *_savedAllObjects;
    NSMutableDictionary   *_queryMatch;
    NSMutableDictionary   *_queryMin;
    NSMutableDictionary   *_queryMax;
    NSMutableDictionary   *_queryOperator;
    NSString              *_defaultStringMatchOperator;
    NSString              *_defaultStringMatchFormat;
    NSMutableDictionary   *_queryBindings;
    void                  *_reserved;
    int                   _flushCount;
    struct
    {
        unsigned   selectsFirstObjectAfterFetch:1;
        unsigned   didChangeContents:1;		
        unsigned   didChangeSelection:1;
        unsigned   didChangeDisplayedObjects:1;
        unsigned   autoFetch:1;
        unsigned   haveFetched:1;
        unsigned   validateImmediately:1;
        unsigned   queryMode:1;
        unsigned   optimisticRefresh:1;
        unsigned   fetchAll:1;
        unsigned   _initialized:1;
        unsigned   _shutdown:1;
        unsigned   selectionObjectsAreOK:1;
        unsigned   memoSelectsFirstObjectAfterFetch:1;
        unsigned   memoHaveFetched:1;
        unsigned   verifySelectedObjects:1;   // unused
        unsigned   hookedIntoEditingContext:1;
        unsigned   displayedObjectsAreOK:1;
        unsigned   justChangedContents:1;     // new addition
        unsigned   justChangedSelection:1;    // new addition
        unsigned   dataSourceSupportsKeyPaths:1;
        unsigned   shouldFetchKeyPaths:1;
        unsigned   _reserved:11;
    } _flags;
    EOAssociation      *_editingAssociation;
    EOEditingContext   *_editorOfContext;     //  hacque, don't use it (unretained)
    // much more private
@private    
   unsigned int   _nDisplayedObjects;
   unsigned int   _nSelectedObjects;
    
@private
   IMP   _impDelegateDisplayGroupCreateObjectFailedForDataSource;
   IMP   _impDelegateDisplayGroupDidChangeDataSource;
   IMP   _impDelegateDisplayGroupDidChangeSelectedObjects;
   IMP   _impDelegateDisplayGroupDidChangeSelection;
   IMP   _impDelegateDisplayGroupDidDeleteObject;
   IMP   _impDelegateDisplayGroupEditFailed;                   // addition!!
   IMP   _impDelegateDisplayGroupDidFetchObjects;
   IMP   _impDelegateDisplayGroupDidInsertObject;
   IMP   _impDelegateDisplayGroupDidSetValueForObjectKey;
   IMP   _impDelegateDisplayGroupDisplayArrayForObjects;
   IMP   _impDelegateDisplayGroupShouldChangeSelectionToIndexes;
   IMP   _impDelegateDisplayGroupShouldDeleteObject;
   IMP   _impDelegateDisplayGroupShouldDisplayAlertWithTitleMessage;
   IMP	 _impDelegateDisplayGroupShouldFetch;
   IMP	 _impDelegateDisplayGroupShouldFetchWithKeyPaths;
   IMP	 _impDelegateDisplayGroupShouldInsertObjectAtIndex;
   IMP	 _impDelegateDisplayGroupShouldRedisplayForChangesInEditingContextNotification;
   IMP	 _impDelegateDisplayGroupShouldRefetchForInvalidatedAllObjectsNotification;
   
   NSZone         *_zone;
   EODataSource   *_memoDataSource;
   int            _allObjectsChangeCount;
}

+ (NSString *) globalDefaultStringMatchOperator;
+ (void) setGlobalDefaultStringMatchOperator:(NSString *) op;

+ (NSString *) globalDefaultStringMatchFormat;
+ (void) setGlobalDefaultStringMatchFormat:(NSString *) format;

+ (BOOL) globalDefaultForValidatesChangesImmediately;
+ (void) setGlobalDefaultForValidatesChangesImmediately:(BOOL) flag;

- (id) init;

- (void) setDataSource:(EODataSource *) dataSource;
- (EODataSource *) dataSource;

- (void) setDelegate:(id) aDelegate;
- (id) delegate;

- (void) redisplay;
- (BOOL) fetch;

- (void) setInQueryMode:(BOOL) flag;
- (void) setSelectsFirstObjectAfterFetch:(BOOL) flag;
- (void) setValidatesChangesImmediately:(BOOL) flag;
- (void) setFetchesOnLoad:(BOOL) flag;
- (void) setShouldFetchKeyPaths:(BOOL) flag;
- (void) setGreaterThanQueryValues:(NSDictionary *) values;
- (void) setLessThanQueryValues:(NSDictionary *) values;
- (void) setEqualToQueryValues:(NSDictionary *) values;
- (void) setQueryOperatorValues:(NSDictionary *) values;
- (void) setQueryBindingValues:(NSDictionary *) values;
- (void) setDefaultStringMatchOperator:(NSString *) anOperator;

- (NSDictionary *) greaterThanQueryValues;
- (NSDictionary *) lessThanQueryValues;
- (NSDictionary *) equalToQueryValues;
- (NSDictionary *) queryOperatorValues;
- (NSDictionary*) queryBindingValues;
- (NSString *) defaultStringMatchOperator;
- (NSString *) defaultStringMatchFormat;
- (NSArray *) localKeys;
- (NSArray *) sortOrderings;
- (EOQualifier *) qualifier;
- (EOQualifier *) qualifierFromQueryValues;

- (void) setSortOrderings:(NSArray *) keySortOrderArray;
- (void) setQualifier:(EOQualifier *) qualifier;
- (void) setDefaultStringMatchFormat:(NSString *) format;
- (void) setLocalKeys:(NSArray *) newKeySet;
- (void) setUsesOptimisticRefresh:(BOOL) flag;
- (void) qualifyDisplayGroup;
- (void) qualifyDataSource;

- (BOOL) inQueryMode;
- (BOOL) selectsFirstObjectAfterFetch;
- (BOOL) validatesChangesImmediately;
- (BOOL) fetchesOnLoad;
- (BOOL) shouldFetchKeyPaths;
- (BOOL) fetchesKeyPaths;
- (BOOL) usesOptimisticRefresh;

- (BOOL) selectionChanged;
- (BOOL) contentsChanged;
- (BOOL) selectionOrContentsChanged;	// addition

- (BOOL) enabledToSetSelectedObjectValueForKey:(NSString *) key;

//
// small bonus method, if contents changed and
// displayedObjects did not change then you don't
// need to refetch (for _dataSources f.e.)
//

- (int) allObjectsChangeCount;

- (void) shutdown;  // to kill the DG pretty much without deallocing

#if 0
- (NSData *) dragNDropDataForObjects:(NSArray *) objects;
- (unsigned int) objectsCountFromNDropData:(NSData *) data;
- (NSArray *) objectsFromDragNDropData:(NSData *) data;
#endif

@end


// keys and values used in editFail: dictionaries
extern NSString   *EODisplayGroupObjectKey;   
extern NSString   *EODisplayGroupIndexKey;   
extern NSString   *EODisplayGroupOperationKey;    
extern NSString   *EODisplayGroupExceptionKey;    
extern NSString   *EODisplayGroupValueKey;    
extern NSString   *EODisplayGroupKeyKey;    

extern NSString   *EODisplayGroupOperationInsertValue;    
extern NSString   *EODisplayGroupOperationUpdateValue;    
extern NSString   *EODisplayGroupOperationDeleteValue;    


@interface NSObject ( EODisplayGroupDelegate)

- (BOOL)                                displayGroup:(EODisplayGroup *) displayGroup
  shouldRedisplayForEditingContextChangeNotification:(NSNotification *) notification;

- (BOOL)                                displayGroup:(EODisplayGroup *) displayGroup
   shouldRefetchForInvalidatedAllObjectsNotification:(NSNotification *) notification;

- (BOOL)                                displayGroup:(EODisplayGroup *) displayGroup
                      shouldChangeSelectionToIndexes:(NSArray *) newIndexes;

- (void) displayGroupDidChangeSelection:(EODisplayGroup *) displayGroup;
- (void) displayGroupDidChangeSelectedObjects:(EODisplayGroup *) displayGroup;
- (BOOL) displayGroupShouldFetch:(EODisplayGroup *) displayGroup;
- (void) displayGroup:(EODisplayGroup *) displayGroup
      didFetchObjects:(NSArray *) objects;

- (NSArray *) displayGroup:(EODisplayGroup *) displayGroup
    displayArrayForObjects:(NSArray *) objects;

- (void)       displayGroup:(EODisplayGroup *) displayGroup
                didSetValue:(id) value
                  forObject:(id) eo
                        key:(NSString *) key;

- (void)             displayGroup:(EODisplayGroup *) displayGroup
  createObjectFailedForDataSource:(EODataSource *) dataSource;

- (BOOL)             displayGroup:(EODisplayGroup *) displayGroup
               shouldInsertObject:object
                          atIndex:(unsigned)newIndex;

- (void)             displayGroup:(EODisplayGroup *) displayGroup
                  didInsertObject:object;

- (BOOL)             displayGroup:(EODisplayGroup *) displayGroup
               shouldDeleteObject:object;

- (void)             displayGroup:(EODisplayGroup *) displayGroup
                  didDeleteObject:object;

- (void) displayGroupDidChangeDataSource:(EODisplayGroup *) displayGroup;

- (BOOL)             displayGroup:(EODisplayGroup *) displayGroup
      shouldDisplayAlertWithTitle:(NSString *) title
                          message:(NSString *) message;

@end


@interface EODataSource ( EODisplayGroupBetterDataSource)

- (void) setQualifierBindings:(NSDictionary *) bindings;
- (void) setAuxiliaryQualifier:(EOQualifier *) qualifier;

@end

/*
@interface EODisplayGroup ( EOEditors)

- (BOOL) editorHasChangesForEditingContext:(EOEditingContext *) editingContext;
- (void) editingContextWillSaveChanges:(EOEditingContext *) editingContext;

@end


@interface EODisplayGroup ( EOMessageHandlers)

- (void)  editingContext:(EOEditingContext *) editingContext
     presentErrorMessage:(NSString *) message;

@end
*/
#endif
