/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2002,2005 Mulle kybernetiK. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 *
 * $Log$
 * Revision 1.3  2002/04/22 22:16:36  nat
 * Noch ein bischen mehr code, genug fuer heute
 *
 * Revision 1.2  2002/04/22 21:56:29  nat
 * EODisplayGroup mit code zugemuellt
 *
 * Revision 1.1.1.1  2002/04/21 23:58:37  nat
 * Mercyful Release - Diese Version kann noch gaaaaa nix
 *
 */
#import "Defines.h"

#import "EODisplayGroup.h"

#import <AppKit/AppKit.h>

#import "EODisplayGroup+Associations.h"
#import "EODisplayGroup+InsertUpdateDelete.h"
#import "EODisplayGroup+Objects.h"
#import "EODisplayGroup+Private.h"
#import "EODisplayGroup+Selection.h"
#import "EOAssociation.h"
#import "EODictionaryDataSource.h"
#import "NSArray+MulleFastRemove.h"
#import "NSMutableArray+MulleFastRemove.h"
#import "NSArray+MulleCommonMembers.h"


NSString  *EODisplayGroupException = @"EODisplayGroupException";
NSString  *EODisplayGroupLikeMatchOperator  = @"like";
NSString  *EODisplayGroupCaseInsensitiveLikeMatchOperator = @"caseInsensitiveLike";

NSString   *EODisplayGroupWillFetchNotification = @"EODisplayGroupWillFetchNotification";

NSString   *EODisplayGroupObjectKey    = @"object";  
NSString   *EODisplayGroupIndexKey     = @"index";  
NSString   *EODisplayGroupOperationKey = @"operation"; 
NSString   *EODisplayGroupExceptionKey = @"exception";    
NSString   *EODisplayGroupValueKey     = @"value";
NSString   *EODisplayGroupKeyKey       = @"key";

NSString   *EODisplayGroupOperationInsertValue = @"insert";   
NSString   *EODisplayGroupOperationUpdateValue = @"update";
NSString   *EODisplayGroupOperationDeleteValue = @"delete";    

NSString   *EODisplayGroupShutdownNotification = @"EODisplayGroupShutdown";  

#define ENCODING_VERSION   270


@interface EODisplayGroup ( _EOForwardDeclarations)

- (BOOL) _setInQueryMode:(BOOL) flag;
- (void) addBoundKeysForDisplayedObjectsToSet:(NSMutableSet *) set;

@end


@implementation EODisplayGroup


static NSString   *globalDefaultStringMatchOperator = @"caseInsensitiveLike";  
static NSString   *globalDefaultStringMatchFormat   = @"%@*";    // Keine Ahnung
static BOOL       globalDefaultForValidatesChangesImmediately = NO;


+ (NSString *) globalDefaultStringMatchOperator
{
   return( globalDefaultStringMatchOperator);
}


+ (void) setGlobalDefaultStringMatchOperator:(NSString *) op
{
   id   old = globalDefaultStringMatchOperator;
   globalDefaultStringMatchOperator = [op copy];
   [old release];
}


+ (NSString *) globalDefaultStringMatchFormat
{
   return( globalDefaultStringMatchFormat);
}


+ (void) setGlobalDefaultStringMatchFormat:(NSString *) format
{
   id   old = globalDefaultStringMatchFormat;
   globalDefaultStringMatchFormat = [format copy];
   [old release];
}


+ (BOOL) globalDefaultForValidatesChangesImmediately
{
   return( globalDefaultForValidatesChangesImmediately);

}

+ (void) setGlobalDefaultForValidatesChangesImmediately:(BOOL) flag
{
   globalDefaultForValidatesChangesImmediately = flag;
}


- (void) willChange
{
   if( self->_flags._shutdown)
      return;
   
   if( self->_flags.justChangedContents || self->_flags.justChangedSelection)
   {
      self->_flags.didChangeContents  |= self->_flags.justChangedContents;
      self->_flags.didChangeSelection |= self->_flags.justChangedSelection;
      self->_flags.justChangedContents  = NO;
      self->_flags.justChangedSelection = NO;
#if DEBUG_VERBOSE
      NSLog( @"%s %@%s%s", __PRETTY_FUNCTION__, self,
             self->_flags.didChangeContents ? " Contents Changed" : "",
             self->_flags.didChangeSelection ? " Selection Changed" : "");
#endif
      [super willChange];
      [EOObserverCenter notifyObserversObjectWillChange:nil];
   }
}


+ (void) initialize
{
   [EODisplayGroup setVersion:ENCODING_VERSION];
}


static void   wipe( EODisplayGroup *self)
{
   [self->_allObjects release];
   self->_allObjects = [NSMutableArray new];
   [self->_displayedObjects release];
   self->_displayedObjects  = [NSMutableArray new];
   self->_nDisplayedObjects = 0;
   [self->_selection release];
   self->_selection = [NSMutableArray new];
   [self->_selectedObjects release];
   self->_selectedObjects = [NSMutableArray new];
   self->_nSelectedObjects  = 0;
   self->_flags.haveFetched = NO;
}


#if DEBUG
unsigned int   NumberOfDisplayGroupsAlive;
NSHashTable    *aliveDisplayGroups;


static void    addAliveDisplayGroup( EODisplayGroup *group)
{
   if( ! aliveDisplayGroups)
      aliveDisplayGroups = NSCreateHashTable( NSNonRetainedObjectHashCallBacks, 10);
   NSHashInsertKnownAbsent( aliveDisplayGroups, group);
}

static void    removeAliveDisplayGroup( EODisplayGroup *group)
{
   if( ! aliveDisplayGroups)
      return;
   
   NSHashRemove( aliveDisplayGroups, group);
   if( ! NSCountHashTable( aliveDisplayGroups))
   {
      NSFreeHashTable( aliveDisplayGroups);
      aliveDisplayGroups = NULL;
   }
}


//
// convenient sometimes...
//
void    clearAliveDisplayGroupsMemory()
{
   if( aliveDisplayGroups)
   {
      NSFreeHashTable( aliveDisplayGroups);
      aliveDisplayGroups = NULL;
   }
}


static void   dumpAliveDisplayGroups()
{
   NSHashEnumerator   rover;
   EODisplayGroup     *group;
   
   if( ! aliveDisplayGroups)
      return;

   rover = NSEnumerateHashTable( aliveDisplayGroups);
   while( group = NSNextHashEnumeratorItem( &rover))
   {
      printf( "%p +%u\n", group, [group retainCount]);
   }
   NSEndHashTableEnumeration( &rover);
}


+ (id) alloc
{
   return( [self allocWithZone:NULL]);
}


+ (id) allocWithZone:(NSZone *) zone
{
   id   p;
   
   NumberOfDisplayGroupsAlive++;
   p = [super allocWithZone:zone];
   addAliveDisplayGroup( p);
   return( p);
}

#endif


- init
{
   [super init];
   
   _updatedObjectIndex = -1;
   _zone = [self zone];
   
   _defaultStringMatchOperator = [globalDefaultStringMatchOperator retain];	// one retain too much doesn't hurt
   _defaultStringMatchFormat   = [globalDefaultStringMatchFormat retain];

#if ENABLE_KEYPATHS   
   _flags.shouldFetchKeyPaths = YES;
#endif   
   // needed to reset the contentsChanged and selectionChanged flag
   // after all other observers are notified.  Thus, EODisplayGroups
   // have a high (6th) delayed observer priority.
   [EOObserverCenter addObserver:self
                       forObject:self];
   
   wipe( self);
   
   return( self);
}


#if DEBUG
- (NSString *) description
{
   return( [NSString stringWithFormat:@"<%@ %p: (%@)%s%s>", isa, self,
         [self dataSource],
         [self contentsChanged] ? " C" : "", 
         [self selectionChanged] ? " S" : ""]);
}
#endif


- (id) nibInstantiate
{
   return( self);
}


- (void) shutdown
{
   NSNotificationCenter   *center;
   
   if( self->_flags._shutdown)
      return;
   
   enterCall( self);
   
#if DEBUG_VERBOSE
   fprintf( stderr, "%s %p\n", __PRETTY_FUNCTION__, self);
#endif
   center = [NSNotificationCenter defaultCenter];
   [center postNotificationName:EODisplayGroupShutdownNotification
                         object:self];
#warning (nat) need to add some stuff here   
   // this MUST be done before asking about observingAssociations!
   self->_flags._shutdown = YES;
   
   [center removeObserver:self];
   [EOObserverCenter removeObserver:self
                          forObject:self];
   [[self observerQueue] dequeueObserver:self];
   [self setDelegate:nil];
   
   //
   // this is a novel idea I think. When the displayGroup goes down
   // it actively breaks the connections of the associations, thereby
   // releasing them implicitly
   // doesn't work if DF is still observing (would get retained again)
   //
   [[self observingAssociations] makeObjectsPerformSelector:@selector( breakConnection)];
   
   [self _setInQueryMode:NO];  // get rid of memo datasource
   [self setDataSource:nil];   // now remove it from EC too!
   [_allObjects release];
   _allObjects = nil;
   [_displayedObjects release];
   _displayedObjects = nil;
   _nDisplayedObjects = 0;
   [_selection release];
   _selection = nil;
   [_selectedObjects release];
   _selectedObjects = nil;
   _nSelectedObjects = 0;
   
   undoCall( self);
}


- (void) dealloc
{
   [self shutdown];
   
#if DEBUG_VERBOSE
   fprintf( stderr, "%s %p\n", __PRETTY_FUNCTION__, self);
#endif
   [_selection release];
   [_sortOrdering release];
   [_qualifier release];
   [_localKeys release];
   [_savedAllObjects release];
   [_insertedObjectDefaultValues release];
   [_queryMatch release];
   [_queryMin release];
   [_queryMax release];
   [_queryOperator release];
   [_defaultStringMatchOperator release];
   [_defaultStringMatchFormat release];
   [_queryBindings release];
   
#if DEBUG
   removeAliveDisplayGroup( self);
   NumberOfDisplayGroupsAlive--;
#endif   
   [super dealloc];
}


#if DEBUG

- (id) retain
{
   return( [super retain]);
}


- (void) release
{
   [super release];
}


- (id) autorelease
{
   return( [super autorelease]);
}

#endif


- (id) initWithCoder:(NSCoder *) decoder
{
   BOOL  bits;
   int   version;
   id    unused;
   
   [self init];

   version = [decoder versionForClassName:@"EODisplayGroup"];
#if DEBUG_CODER   
   NSLog( @"<<>> %@ %p Version: %d", isa, self, version);
#endif   
   _dataSource = [[decoder decodeObject] retain];

   unused = [decoder decodeObject];
   NSAssert1( unused == nil, @"interesting, what did we find here ? %@", unused);
   unused = [decoder decodeObject];
   NSAssert1( unused == nil, @"how interesting, what did we find here ? %@", unused);
   
   _sortOrdering = [[decoder decodeObject] retain];
   _qualifier    = [[decoder decodeObject] retain];

   bits = 0;
   [decoder decodeValueOfObjCType:@encode( BOOL)
                               at:&bits];
   _flags.autoFetch = bits;

   [decoder decodeValueOfObjCType:@encode( BOOL)
                               at:&bits];
   _flags.validateImmediately = bits;

#warning (guess)   
   [decoder decodeValueOfObjCType:@encode( BOOL)
                               at:&bits];
   _flags.selectsFirstObjectAfterFetch = bits;	  

   [decoder decodeValueOfObjCType:@encode( BOOL)
                               at:&bits];
   _flags.optimisticRefresh = bits;  // used to be inverted, because of bug in palette
   
   _localKeys = [[decoder decodeObject] retain];

#if DEBUG_CODER   
   NSLog( @"Done %@", self);
#endif   
   
   return( self);
}


- (void) encodeWithCoder:(NSCoder *) encoder
{
   BOOL  bits;

#if DEBUG_CODER
   NSLog( @"%@ %s Version=%d (%d)", self, __PRETTY_FUNCTION__, [EODisplayGroup version], [isa version]);
#endif   
   [encoder encodeObject:[self dataSource]];  // encode "real" data source
   [encoder encodeObject:nil];
   [encoder encodeObject:nil];
   [encoder encodeObject:_sortOrdering];
   [encoder encodeObject:_qualifier];

   bits = _flags.autoFetch;
   [encoder encodeValueOfObjCType:@encode( BOOL)
                               at:&bits];
   bits = _flags.validateImmediately;
   [encoder encodeValueOfObjCType:@encode( BOOL)
                               at:&bits];
   bits = _flags.selectsFirstObjectAfterFetch;
   [encoder encodeValueOfObjCType:@encode( BOOL)
                               at:&bits];
   bits = _flags.optimisticRefresh;
   [encoder encodeValueOfObjCType:@encode( BOOL)
                               at:&bits];
   
   [encoder encodeObject:_localKeys];
#if DEBUG_CODER   
   NSLog( @"Done %@", self);
#endif   
}


- (EODataSource *) dataSource
{
#warning (nat) freestyle code here, is this EODictionaryDataSource a good idea ??
   //
   // this is obsure, but seems to work better with my old code, it appears to expect this
   // this is also "exploited" in the encoding code...
   //
   return( _flags.queryMode ? _memoDataSource : _dataSource);
}


- (id) delegate
{
   return( _delegate);
}


- (int) allObjectsChangeCount
{
   return( _allObjectsChangeCount);
}


static IMP  imp__( id obj, SEL sel)
{
   return( [obj respondsToSelector:sel] ? [obj methodForSelector:sel] : 0);
}


- (void) setDelegate:(id) aDelegate
{
   _delegate = aDelegate;
   _impDelegateDisplayGroupCreateObjectFailedForDataSource = imp__( _delegate, @selector( displayGroup:createObjectFailedForDataSource:));
   _impDelegateDisplayGroupDidChangeDataSource             = imp__( _delegate, @selector( displayGroupDidChangeDataSource:));
   _impDelegateDisplayGroupDidChangeSelectedObjects        = imp__( _delegate, @selector( displayGroupDidChangeSelectedObjects:));
   _impDelegateDisplayGroupDidChangeSelection              = imp__( _delegate, @selector( displayGroupDidChangeSelection:));
   _impDelegateDisplayGroupDidDeleteObject                 = imp__( _delegate, @selector( displayGroup:didDeleteObject:));
   _impDelegateDisplayGroupEditFailed                      = imp__( _delegate, @selector( displayGroup:editFailed:));
   _impDelegateDisplayGroupDidFetchObjects                 = imp__( _delegate, @selector( displayGroup:didFetchObjects:));
   _impDelegateDisplayGroupDidInsertObject                 = imp__( _delegate, @selector( displayGroup:didInsertObject:));
   _impDelegateDisplayGroupDidSetValueForObjectKey         = imp__( _delegate, @selector( displayGroup:didSetValue:forObject:key:));
   _impDelegateDisplayGroupDisplayArrayForObjects          = imp__( _delegate, @selector( displayGroup:displayArrayForObjects:));
   _impDelegateDisplayGroupShouldChangeSelectionToIndexes  = imp__( _delegate, @selector( displayGroup:shouldChangeSelectionToIndexes:));
   _impDelegateDisplayGroupShouldDeleteObject              = imp__( _delegate, @selector( displayGroup:shouldDeleteObject:));

   _impDelegateDisplayGroupShouldDisplayAlertWithTitleMessage  = imp__( _delegate, @selector( displayGroup:shouldDisplayAlertWithTitle:message:));
   _impDelegateDisplayGroupShouldFetch                         = imp__( _delegate, @selector( displayGroupShouldFetch:));
   _impDelegateDisplayGroupShouldFetchWithKeyPaths             = imp__( _delegate, @selector( displayGroup:shouldFetchWithKeyPaths:));
   _impDelegateDisplayGroupShouldInsertObjectAtIndex           = imp__( _delegate, @selector( displayGroup:shouldInsertObject:atIndex:));
   _impDelegateDisplayGroupShouldRedisplayForChangesInEditingContextNotification = imp__( _delegate, @selector( displayGroup:shouldRedisplayForEditingContextChangeNotification:));
   _impDelegateDisplayGroupShouldRefetchForInvalidatedAllObjectsNotification = imp__( _delegate, @selector( displayGroup:shouldRefetchForInvalidatedAllObjectsNotification:));
}


- (void) subjectChanged
{
   _flags.didChangeContents  = NO;
   _flags.didChangeSelection = NO;
   _updatedObjectIndex       = -1;
#if DEBUG_VERBOSE
   NSLog( @"%@ %s (Leaving...)", self, __PRETTY_FUNCTION__);
#endif   
}


- (EOObserverPriority) priority
{
   return( EOObserverPrioritySixth);
}


- (void) objectsInvalidatedInEditingContext:(NSNotification *) notification
{
   BOOL   flag;
   
   
   flag = YES;
   if( _impDelegateDisplayGroupShouldRefetchForInvalidatedAllObjectsNotification)
      flag = (* (BOOL (*)()) _impDelegateDisplayGroupShouldRefetchForInvalidatedAllObjectsNotification)( _delegate, @selector( displayGroup:shouldRefetchForInvalidatedAllObjectsNotification:), self, notification);

   // lots of objects could be gone, we refetch regardless
   if( flag)
      [self fetch];
   
   // hmmm, when do we fetch then ??
}



- (void) _objectsChangedInEditingContext:(NSNotification *) notification
{
   NSDictionary     *userInfo;
   NSArray          *updated;
   NSArray          *deleted;
   NSArray          *inserted;
   NSArray          *invalidated;
   BOOL             flag;
   unsigned int     n;
#if 0   
   NSSet            *allSet;
   NSMutableArray   *array;
   unsigned int     i;
   id               obj;
#endif    
//   Class          myObjectsClass;
   
   userInfo = [notification userInfo];

   updated     = [userInfo objectForKey:@"updated"];
   deleted     = [userInfo objectForKey:@"deleted"];
   inserted    = [userInfo objectForKey:@"inserted"];
   invalidated = [userInfo objectForKey:@"invalidated"];

#if DEBUG_VERBOSE
   NSLog( @"%@ %s (U:%@ D:%@ I:%u V:%u)", self, __PRETTY_FUNCTION__,
          updated, deleted, [inserted count], [invalidated count]);
#endif

   //
   // if invalidated, then we have a lot of faults here
   // which we don't want to fire, so we say YES
   //
   if( [invalidated count])
   {
      displayedObjectsHaveChanged( self);
      return;
   }
   
   //
   // if the DG is really large, it makes no sense to be smart about it
   // as the smartness is too expensive
   //
   if( [[self displayedObjects] count] > 1000000)
   {
      displayedObjectsHaveChanged( self);
      return;
   }
   
   // Grunz: hier krieg ich die klasse nicht...
   //
   //   myObjectsClass = [_dataSource classDescriptionForObjects]:

   //
   // remove all known deleted objects. If we have deletions
   // in the end (still) they can be considered changes to
   // instance variables...
   //
   
#if 0
   allSet = nil;
   
   if( n = [deleted count])
   {
      allSet = [NSSet setWithArray:_allObjects];

      flag = NO;
      for( i = n; i > 0;)
      {
         --i;
         obj = [deleted objectAtIndex:i];
         if( [allSet containsObject:obj])
         {
            [_allObjects removeObjectIdenticalTo:obj];
            [_displayedObjects removeObjectIdenticalTo:obj];
            self->_nDisplayedObjects = [_displayedObjects count];
            [_selectedObjects removeObjectIdenticalTo:obj];
            self->_nSelectedObjects = [_selectedObjects count];
            flag = YES;
         }
      }
      
      if( flag)
      {
         displayedObjectsHaveChanged( self);
         selectionHasChanged( self);
         invalidateSelectedIndexes( self);
         return;  // that's quite enough, thank you
     }
   }

   //
   // check all inserted objects, to be sure we don't already
   // have them, forcing needless screen updates
   //
   if( n = [inserted count])
   {
      if( ! allSet)
         allSet = [NSSet setWithArray:_allObjects];

      array = nil;
      for( i = n; i > 0;)
      {
         --i;
         obj = [inserted objectAtIndex:i];
         if( [allSet containsObject:obj])
         {
            if( ! array)
               array = [NSMutableArray arrayWithArray:inserted];
            [array removeObjectAtIndex:i];
         }
      }
      inserted = array ? array : inserted;
   }
#endif
   
   n = [deleted count];
   if( n)
   {
      // if that many are deleted, just assume we are changed
      if( n >= 10000 || [_displayedObjects mulleHasCommonMembersWithArray:deleted])
      {
         Class         aClass;
         EODataSource  *source;
         NSArray       *array;
         
         // --- DESIGN WART ---
         // ungh, so something has been deleted what we are displaying
         // if this is a EODetailDataSource then the association is 
         // going to refetch anyway...  
         // if we have an EOArrayDataSource then we should refetch really
         // if we have some other, then we remove the objects from the
         // array. 
         // -- this should be configurable by the outside --
         // or maybe one should ask the dataSource
         //
         source = [self dataSource];
         aClass = [source class];
         
         if( aClass == [EODetailDataSource class])
         {
            displayedObjectsHaveChanged( self);
            return;
         }
         else
            if( aClass == [EOArrayDataSource class])
            {
               [self fetch];
               return;
            }
         
         // other classes, lets change the "objectArray"
         array = [[self allObjects] mulleArrayByRemovingObjectsInArray:deleted];
         [self setObjectArray:array];
         return;
      }
   }

   //
   // Do the "official" optimistic check now
   //
   if( _flags.optimisticRefresh)
   {
      n = [updated count];
      if( n < 10000)
      {
         flag = [_displayedObjects mulleHasCommonMembersWithArray:updated];
         if( ! flag)
            return;
      }
   }
 
   // FUTURE.
   // Make this smarter!! Think hard about what it means to have an
   // insert of which type where!! Consider the datasource and the
   // qualifiers etc. [I think this can't be made to work] refetching
   // here is not what the user wants. It would be nicest to filter the
   // objects through the datasources fetchspecifications and qualifiers
   // and then put it into our allObjects. Finally, reupdate and reselect
   // 
   if( [inserted count] + [updated count] + [deleted count])
      displayedObjectsHaveChanged( self);

   return;
}


- (void) objectsChangedInEditingContext:(NSNotification *) notification
{
   BOOL   flag;
   
//   resetCall( self);  // DO NOT RESET ON NOTIFICATION CALLS!!
   enterCall( self);
   flag = YES;
   if( _impDelegateDisplayGroupShouldRedisplayForChangesInEditingContextNotification)
      flag = (* (BOOL (*)()) _impDelegateDisplayGroupShouldRedisplayForChangesInEditingContextNotification)( _delegate, @selector( displayGroup:shouldRedisplayForEditingContextChangeNotification:), self, notification);
   if( flag)
      [self _objectsChangedInEditingContext:notification];
   leaveCall( self);
}


static void   hookIntoEditingContext( EODisplayGroup *self, EOEditingContext *context)
{
   NSNotificationCenter   *center;

  if( ! context || self->_flags.hookedIntoEditingContext)
      return;
   
   center = [NSNotificationCenter defaultCenter];
   [context addEditor:self];
   if( [context messageHandler] == nil)
      [context setMessageHandler:self];
      
   [center addObserver:self
              selector:@selector( objectsInvalidatedInEditingContext:)
                  name:EOInvalidatedAllObjectsInStoreNotification
                object:context];
   [center addObserver:self
              selector:@selector( objectsChangedInEditingContext:)
                  name:EOObjectsChangedInEditingContextNotification
                object:context];
   self->_flags.hookedIntoEditingContext = YES;
   self->_editorOfContext                = context;
#if DEBUG_VERBOSE
   NSLog( @"<<>> Added %p to context %p as editor", self, context);
#endif      
}


static void   unhookFromEditingContext( EODisplayGroup *self, EOEditingContext *context)
{
   NSNotificationCenter   *center;
   
   if( context)
   {
   // doing this twice doesn't hurt
      center = [NSNotificationCenter defaultCenter];
      if( [context messageHandler] == self)
         [context setMessageHandler:nil];
      [context removeEditor:self];
      [center removeObserver:self
                        name:EOObjectsChangedInEditingContextNotification
                      object:context];
      [center removeObserver:self
                        name:EOInvalidatedAllObjectsInStoreNotification
                      object:context];
   }
   self->_flags.hookedIntoEditingContext = NO;
   self->_editorOfContext                = nil;
}


// how much of that stuff do I need to do in awakeFromNib ??
- (void) setDataSource:(EODataSource *) dataSource
{
   EOEditingContext       *context;
   id                     old;

   NSParameterAssert( ! dataSource|| [dataSource isKindOfClass:[EODataSource class]]);

   enterCall( self);

   // nib loading calls this extraneously apparently.
   // (see initWithCoder and awakeWithNib)
#if 1   
   if( dataSource == _dataSource)
   {
      leaveCall( self);
      return;
   }
#endif
   
NS_DURING
   old  = _dataSource;
   _dataSource = [dataSource retain];

   _flags.dataSourceSupportsKeyPaths = [_dataSource respondsToSelector:@selector( setKeyPaths:)];
   
   //
   // we need to keep _editorOfContext around, because especially
   // detail datasources have a tendency to lose their editingContext
   // fetch will hook it up
   //
   context = [dataSource editingContext];
   if( _editorOfContext != context)
   {
      unhookFromEditingContext( self, _editorOfContext);
      hookIntoEditingContext( self, context);
   }   
   // especially first time around this generates too much delegate
   // traffic, that upsets applications

   // now release old, which may release the editingContext also
   [old release];
   
   if( _impDelegateDisplayGroupDidChangeDataSource)
      _impDelegateDisplayGroupDidChangeDataSource( _delegate, @selector( displayGroupDidChangeDataSource:), self);
NS_HANDLER
#if DEBUG
   NSLog( @"******EXCEPTION: %@", localException);
#endif   
   resetCall( self);
   [localException raise];
NS_ENDHANDLER
   leaveCall( self);
}


- (void) redisplay
{
   // NEED to do this,
   // they also set the changedSelection
   // and changedContents flags!
   //
#if DEBUG_VERBOSE
   NSLog( @"%@ %s", self, __PRETTY_FUNCTION__);
#endif
   
   displayedObjectsHaveChanged( self);
   selectionHasChanged( self);
   invalidateSelectedIndexes( self);
   
   [self willChange];
}


//
// Ref: The Crown - "Under The Whip"
//
static NSException   *_fetch( EODisplayGroup *self)
{
   NSMutableSet   *set;
   NSArray        *array;
   NSArray        *keyPaths;
   NSException    *exception;
   NSDictionary   *userInfo;
   
   if( self->_impDelegateDisplayGroupShouldFetch && ! self->_flags.queryMode)
      if( ! self->_impDelegateDisplayGroupShouldFetch( self->_delegate, @selector( displayGroupShouldFetch:), self))
         return( nil);

   array     = nil;
   exception = nil;
NS_DURING      // ensure that wrapping delegate method is called
   //
   // ok: assume that nil means failure of some sort ??
   //
   if( [self fetchesKeyPaths])
   {
      keyPaths = nil;
      
      if( ! self->_flags.queryMode)
      {
         set = [NSMutableSet set];
         [self addBoundKeysForDisplayedObjectsToSet:set];
         keyPaths = [set allObjects];
      }

      if( keyPaths && self->_impDelegateDisplayGroupShouldFetchWithKeyPaths)
         keyPaths = (*self->_impDelegateDisplayGroupShouldFetchWithKeyPaths)( self->_delegate, @selector( displayGroup:                    shouldFetchWithKeyPaths:), self, keyPaths);                         
      [self->_dataSource setKeyPaths:keyPaths];
   }
   
   array = [self->_dataSource fetchObjects];

   //
   // it's more convenient for me to do it before setting the object array,
   // because that may trigger more fetched by faults...
   //
   if( self->_impDelegateDisplayGroupDidFetchObjects && ! self->_flags.queryMode)
      self->_impDelegateDisplayGroupDidFetchObjects( self->_delegate, @selector( displayGroup:didFetchObjects:), self, array);
   
   if( array)   // could be empty, but nil is an error
   {
      self->_flags.haveFetched = YES;
      [self setObjectArray:array];
   }
   else
      if( self->_dataSource)
      {
         userInfo  = [NSDictionary dictionaryWithObjectsAndKeys:self, @"displayGroup", nil];
         exception = [[[NSException alloc] initWithName:NSGenericException
                                                 reason:@"Datasource failed to fetch"
                                               userInfo:userInfo] autorelease];
      }
NS_HANDLER
   exception = [localException retain];
NS_ENDHANDLER
   return( exception); 
}


- (BOOL) fetch
{
   NSException        *exception;
   EOEditingContext   *context;
   
#if DEBUG_VERBOSE
   NSLog( @"%@ %s", self, __PRETTY_FUNCTION__);
#endif   
   enterCall( self);
   
   NSParameterAssert( [self inQueryMode] == NO);  // is this ever useful ???
   
   if( ! [self endEditing])
   {
      leaveCall( self);
      return( NO);
   }
   
   [[NSNotificationCenter defaultCenter] postNotificationName:EODisplayGroupWillFetchNotification
                                                       object:self];
   //
   // during fetch we can register ourselves as editors
   // stupid detail displayGroups...
   if( ! _editorOfContext)
   {
      context = [_dataSource editingContext];
      hookIntoEditingContext( self, context);
   }
   
   exception = _fetch( self);
   NSParameterAssert( ! self->_flags.selectsFirstObjectAfterFetch || 
                      self->_nSelectedObjects || 
                      ! self->_nDisplayedObjects);

#if DEBUG
   if( exception)
      NSLog( @"*** exception: %@", exception);
#endif   
   if( exception)
   {
      resetCall( self);
      [exception raise];
      return( NO);
   }
   
   leaveCall( self);
   return( YES);
}


- (void) setSortOrderings:(NSArray *) keySortOrderArray
{
   id   old;

   NSParameterAssert( ! keySortOrderArray|| [keySortOrderArray isKindOfClass:[NSArray class]]);

   old = _sortOrdering;
   _sortOrdering = [keySortOrderArray copyWithZone:_zone];
   [old release];
}


- (NSArray *) sortOrderings
{
   return( _sortOrdering);
}


- (void) setQualifier:(EOQualifier *) q
{
   id   old;
   
   NSParameterAssert( ! q|| [q isKindOfClass:[EOQualifier class]]);

   old = _qualifier;
   _qualifier = [q retain];
   [old release];
}


- (EOQualifier *) qualifier
{
   return( _qualifier);
}


static void   addQualifiersFromMatchDictionary( NSMutableArray *array,
                                                SEL defaultOp,
                                                NSDictionary *matches,
                                                NSDictionary *operators,
                                                NSString *defaultMatch)
{
   NSEnumerator          *rover;
   NSString              *key;
   EOKeyValueQualifier   *q;
   NSString   		 *string;
   SEL                   op;
   id                    value;
   Class                 nsStringClass;
   
   nsStringClass = [NSString class];
   
   rover = [matches keyEnumerator];
   while( key = [rover nextObject])
   {
      op     = defaultOp;
      value  = [matches objectForKey:key];
      string = [operators objectForKey:key];
      if( string)
         op = [EOQualifier operatorSelectorForString:string];

      // actually according to the dox, this should be
      // not if (! value) but if( ! operator)
      if( [value isKindOfClass:nsStringClass])
      {
         if( [value length] == 0)
            continue;
         if( defaultMatch)
            value = [NSString stringWithFormat:defaultMatch, value];
      }
      else
         if( op == EOQualifierOperatorLike || op == EOQualifierOperatorCaseInsensitiveLike)
            op = EOQualifierOperatorEqual;
      
      NSCAssert1( op != 0, @"Operator %@ unknown in QueryOperator dictionary", string);
      q  = [[EOKeyValueQualifier alloc] initWithKey:key
                                   operatorSelector:op
                                              value:value];
      [array addObject:q];
      [q release];
   }
}


- (EOQualifier *) qualifierFromQueryValues
{
   NSMutableArray  *array;
   SEL             sel;
   
   array = [NSMutableArray array];
   // why is it called a "default" ??
   
   sel = EOQualifierOperatorCaseInsensitiveLike;
   if( [_defaultStringMatchOperator isEqualToString:EODisplayGroupLikeMatchOperator])
      sel = EOQualifierOperatorLike;
   
   addQualifiersFromMatchDictionary( array, sel,
                                     _queryMatch, _queryOperator,
                                     _defaultStringMatchFormat);
   addQualifiersFromMatchDictionary( array, EOQualifierOperatorLessThan,
                                     _queryMin, nil,
                                     _defaultStringMatchFormat);
   addQualifiersFromMatchDictionary( array, EOQualifierOperatorGreaterThan,
                                     _queryMax, nil,
                                     _defaultStringMatchFormat);

   switch( [array count])
   {
   case 0 :
      return( nil);
      
   case 1 :
      return( [array objectAtIndex:0]);
   }
   return( [[[EOAndQualifier alloc] initWithQualifierArray:array] autorelease]);
}


- (void) qualifyDisplayGroup
{
   enterCall( self);
   
   [self endEditingOrRaise];
   
   [self setQualifier:[self qualifierFromQueryValues]];
   if( [self inQueryMode])
      [self setInQueryMode:NO];
   
   [self updateDisplayedObjects];
   leaveCall( self);
}


- (void) qualifyDataSource
{
   EOQualifier    *q;
   NSDictionary   *bindings;
   
   enterCall( self);

   [self endEditingOrRaise];

   q        = [self qualifierFromQueryValues];
   bindings = [self queryBindingValues];
   
   if( [self inQueryMode])
      [self setInQueryMode:NO];
   
   // now datasource back to normal
   if( [_dataSource respondsToSelector:@selector( setQualifierBindings:)])
      [(id) _dataSource setQualifierBindings:bindings];
   if( [_dataSource respondsToSelector:@selector( setAuxiliaryQualifier:)])
      [(id) _dataSource setAuxiliaryQualifier:q];
   // else complain ??
   
   [self fetch];
   leaveCall( self);
}


- (void)   _resetQueryDictionariesIfNeeded
{
   if( ! _queryMatch)
      _queryMatch = [[NSMutableDictionary allocWithZone:_zone] init];
   
   if( ! _queryMin)
      _queryMin = [[NSMutableDictionary allocWithZone:_zone] init];
   
   if( ! _queryMax)
      _queryMax = [[NSMutableDictionary allocWithZone:_zone] init];
}


- (BOOL) _setInQueryMode:(BOOL) flag
{
   if( _flags.queryMode == flag)
      return( NO);
   
   wipe( self);
   _flags.queryMode = flag;
   if( flag)  //
   {
      _memoDataSource = _dataSource;
      
      _flags.memoSelectsFirstObjectAfterFetch = _flags.selectsFirstObjectAfterFetch;
      _flags.memoHaveFetched                  = _flags.haveFetched;
      [self _resetQueryDictionariesIfNeeded];
      _flags.selectsFirstObjectAfterFetch     = YES;
      _flags.haveFetched                      = NO;
      
      // hmm only queryMatch ???
      _dataSource = [[EODictionaryDataSource allocWithZone:_zone]    
                                    initWithMutableDictionary:_queryMatch];
      _fetch( self);  // must do this app.
      return( YES);
   }
   
   NSParameterAssert( _memoDataSource);
      
   _flags.selectsFirstObjectAfterFetch = _flags.memoSelectsFirstObjectAfterFetch;
   _flags.haveFetched                  = _flags.memoHaveFetched;
      
   [_dataSource release];
   _dataSource     = _memoDataSource;
   _memoDataSource = nil;
   
   return( YES);
}


- (void) setInQueryMode:(BOOL) flag
{
   BOOL   changed;
   
   enterCall( self);

   changed = NO;
NS_DURING
   changed = [self _setInQueryMode:flag];
      
NS_HANDLER
#if DEBUG
   NSLog( @"******EXCEPTION: %@", localException);
#endif   
   resetCall( self);
   [localException raise];
NS_ENDHANDLER   
   
   if( changed)
      displayedObjectsHaveChanged( self);
   leaveCall( self);
}


- (BOOL) inQueryMode
{
   return( _flags.queryMode);
}


- (void) setSelectsFirstObjectAfterFetch:(BOOL) flag
{
   _flags.selectsFirstObjectAfterFetch = flag;
}


- (BOOL) selectsFirstObjectAfterFetch
{
   return( _flags.selectsFirstObjectAfterFetch);
}


- (void) setValidatesChangesImmediately:(BOOL) flag
{
   _flags.validateImmediately = flag;
}


- (BOOL) validatesChangesImmediately
{
   return( _flags.validateImmediately);
}


- (void) setFetchesOnLoad:(BOOL) flag
{
   _flags.autoFetch = flag;
}


- (void) setShouldFetchKeyPaths:(BOOL) flag
{
   _flags.shouldFetchKeyPaths = flag;
}


- (BOOL) fetchesOnLoad
{
   return( _flags.autoFetch);
}


- (BOOL) shouldFetchKeyPaths
{
   return( _flags.shouldFetchKeyPaths);
}


- (BOOL) fetchesKeyPaths
{
   return( _flags.shouldFetchKeyPaths && _flags.dataSourceSupportsKeyPaths);
}


- (NSDictionary *) greaterThanQueryValues
{
   return( _queryMax);
}


- (void) setGreaterThanQueryValues:(NSDictionary *) values
{
   id   old;
   
   NSParameterAssert( ! values || [values isKindOfClass:[NSDictionary class]]);
   old = _queryMax;
   _queryMax = [values mutableCopyWithZone:_zone];
   [old release];
}


- (NSDictionary *) lessThanQueryValues
{
   return( _queryMin);
}


- (void) setLessThanQueryValues:(NSDictionary *) values
{
   id   old;
   
   NSParameterAssert( ! values || [values isKindOfClass:[NSDictionary class]]);
   old = _queryMin;
   _queryMin = [values mutableCopyWithZone:_zone];
   [old release];
}


- (NSDictionary *) equalToQueryValues
{
   return( _queryMatch);
}


- (void) setEqualToQueryValues:(NSDictionary *) values
{
   id   old;
   
   NSParameterAssert( ! values || [values isKindOfClass:[NSDictionary class]]);
   old = _queryMatch;
   _queryMatch = [values mutableCopyWithZone:_zone];
   [old release];
}


- (NSDictionary *) queryOperatorValues
{
   return( _queryOperator);
}


- (void) setQueryOperatorValues:(NSDictionary *) values
{
   id   old;
   
   NSParameterAssert( ! values || [values isKindOfClass:[NSDictionary class]]);
   old = _queryOperator;
   _queryOperator = [values copyWithZone:_zone];
   [old release];
}


- (NSDictionary *) queryBindingValues
{
   return( _queryBindings);
}


- (void) setQueryBindingValues:(NSDictionary *) values
{
   id   old;
   
   NSParameterAssert( ! values || [values isKindOfClass:[NSDictionary class]]);
   old = _queryBindings;
   _queryBindings = [values mutableCopyWithZone:_zone];
   [old release];
}


- (NSString *) defaultStringMatchOperator
{
   return( _defaultStringMatchOperator);
}


- (void) setDefaultStringMatchOperator:(NSString *) s
{
   id   old;
   
   NSAssert( [s isEqualToString:EODisplayGroupLikeMatchOperator] || [s isEqualToString:EODisplayGroupCaseInsensitiveLikeMatchOperator],
             @"can only deal with like and caseInsensitiveLike");

   old = _defaultStringMatchOperator;
   _defaultStringMatchOperator = [s copyWithZone:_zone];
   [old release];
}


- (NSString *) defaultStringMatchFormat
{
   return( _defaultStringMatchFormat);
}


- (void) setDefaultStringMatchFormat:(NSString *) format
{
   id   old;
   
   NSParameterAssert( ! format || [format isKindOfClass:[NSString class]]);
   old = _defaultStringMatchFormat;
   _defaultStringMatchFormat = [format copyWithZone:_zone];
   [old release];
}


- (NSArray *) localKeys
{
   return( _localKeys);
}


- (void) setLocalKeys:(NSArray *) newKeySet
{
   id   old;
   
   NSParameterAssert( ! newKeySet || [newKeySet isKindOfClass:[NSArray class]]);
   old = _localKeys;
   _localKeys = [newKeySet copyWithZone:_zone];
   [old release];
}


- (void) setUsesOptimisticRefresh:(BOOL) flag
{
   _flags.optimisticRefresh = flag;
}


- (BOOL) usesOptimisticRefresh
{
   return( _flags.optimisticRefresh);
}


- (BOOL) selectionOrContentsChanged
{
   return( _flags.didChangeSelection | _flags.didChangeContents | _flags.justChangedSelection | _flags.justChangedContents);
}


- (BOOL) selectionChanged
{
   return( _flags.didChangeSelection | _flags.justChangedSelection);
}


- (BOOL) contentsChanged
{
   return( _flags.didChangeContents | _flags.justChangedContents);
}


//
// OK this stuff, i have no idea what's the use
// just implemented according to the dox
//
#warning (nat) need to use this in associations for enable/disable code
- (BOOL) enabledToSetSelectedObjectValueForKey:(NSString *) key
{
   if( [key hasPrefix:@"@"])
      return( YES);

   return( _nSelectedObjects > 0);
}


- (void) awakeFromNib
{
#if DEBUG_VERBOSE
   NSLog( @"<<>> %@ %p %s", isa, self, __PRETTY_FUNCTION__);
#endif   
   if( ! _flags.haveFetched)
      wipe( self);
   
   [EOObserverCenter addObserver:self
                       forObject:self];

//#if DEBUG_VERBOSE
//   if( [_dataSource isKindOfClass:[EODetailDataSource class]])
//      NSLog( @"Ha");
//#endif
   
   // reestablish notifications, and editors relation
   hookIntoEditingContext( self, [_dataSource editingContext]);
   
   if( _flags.autoFetch)
      [self fetch];
}


- (BOOL) editorHasChangesForEditingContext:(EOEditingContext *) anEditingContext
{
   NSParameterAssert( [anEditingContext isKindOfClass:[EOEditingContext class]]);

   if( _editingAssociation)
      return( YES);
   return( NO);
}


#ifdef GNUSTEP
#define NSObjectNotAvailableException  EOGenericException
#endif

- (void) editingContextWillSaveChanges:(EOEditingContext *) anEditingContext
{
   NSParameterAssert( [anEditingContext isKindOfClass:[EOEditingContext class]]);

   if( _editingAssociation)
   {
      if( ! [_editingAssociation endEditing])
         [NSException raise:NSObjectNotAvailableException
                     format:@"Save aborted, because user is still editing"];
   }
}


- (void) editingContext:(EOEditingContext *) anEditingContext
    presentErrorMessage:(NSString *) errorMessage
{

   // how obsolete...
   // 
   NSRunCriticalAlertPanel( @"Database Access",
                            errorMessage,
                            @"OK", nil, nil);
   
   // next version, maybe tell the delegate about it...
}


- (NSString *) debugDescription
{
   NSString       *s;
   unsigned int   n;
   
   s = @"(0)";
   n = [[self allObjects] count];
   if( n)
      s = [NSString stringWithFormat:@"(%u D:%u S:%u)", n, 
         [[self displayedObjects] count],
         [[self selectedObjects] count]];
         
   return( [NSString stringWithFormat:@"<%@ %p (%@) %@>", isa, self, [_dataSource debugDescription], s]);
}

@end


