/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2006 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 * $Log$
 *
 */
#import "NSArray+MulleCommonMembers.h"


@implementation NSArray ( MulleCommonMembers)

//
// just use hashtables here, because sets use 'isEqual' and that
// will trigger faults, which we'd want to avoid
// make it a category on NSArray sometime
//
- (BOOL) mulleHasCommonMembersWithArray:(NSArray *) bArray
{
   NSArray          *aArray;
   BOOL             flag;
   unsigned int     i, n;
   unsigned int     aCount;
   unsigned int     bCount;
   unsigned int     offset;
   NSHashTable      *hashTable;
   id               buf[ 64];
   
   aArray = self;
   
   bCount = [bArray count];
   if( ! bCount)
      return( NO);
   
   aCount = [aArray count];
   if( ! aCount)
      return( NO);
   
   if( bCount > aCount)
   {
      unsigned int   tmpCount;
      NSArray        *tmpArray;
      
      tmpCount = aCount;
      aCount   = bCount;
      bCount   = tmpCount;
      
      tmpArray = bArray;
      aArray   = bArray;
      bArray   = tmpArray;
   }
   
   //
   // build up a hash table from array a (the smaller)
   //
   
   hashTable = NSCreateHashTable( NSNonOwnedPointerHashCallBacks, aCount);
   for( offset = 0; offset < aCount; offset += 64)
   {
      n = aCount - offset;
      if( n > 64)
         n = 64;
      [aArray getObjects:buf
                   range:NSMakeRange( offset, n)];
      for( i = 0; i < n; i++)
         NSHashInsert( hashTable, buf[ i]);
   }
   
   //
   // run thru b and look for a hit
   //
   flag = NO;
   for( offset = 0; offset < bCount; offset +=64)
   {
      n = bCount - offset;
      if( n > 64)
         n = 64;
      [bArray getObjects:buf
                   range:NSMakeRange( offset, n)];
      for( i = 0; i < n; i++)
         if( NSHashGet( hashTable, buf[ i]))
         {
            flag = YES;
            break;
         }
   }
      
   NSFreeHashTable( hashTable);
      
   return( flag);
}


@end
