/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF
 *  Copyright (C) 2002 Mulle kybernetiK. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *
 *  Created by Nat! on Sun Apr 28 2002.
 *
 * $Id$
 *
 *
 * $Log$
 *
 */
#import "Defines.h"

#import "EOAspectConnector+IBConnectors.h"

#import "EOAssociation.h"


@implementation EOAspectConnector ( IBConnectors)

- (id) source
{
#if DEBUG_ASPECT_CONNECTOR_VERBOSE
   NSLog( @"%s %@ %s (%@)", __PRETTY_FUNCTION__, self, [isa conformsToProtocol:@protocol( IBConnectors)] ? "CONFORMS" : "DOES NOT CONFORM", [association object]);
#endif
   return( [association object]);
}


- (id) destination
{
#if DEBUG_ASPECT_CONNECTOR_VERBOSE
   NSLog( @"%s %@", __PRETTY_FUNCTION__, self);
#endif   
   return( [association displayGroupForAspect:aspectName]);
}


- (NSString *) label
{
#if DEBUG_ASPECT_CONNECTOR_VERBOSE
   NSLog( @"%s %@", __PRETTY_FUNCTION__, self);
#endif   
   return( [association displayGroupKeyForAspect:aspectName]);
}


- (void) replaceObject:(id) oldObject
            withObject:(id) newObject
{
#if DEBUG_ASPECT_CONNECTOR
   NSLog( @"%s %@", __PRETTY_FUNCTION__, self);
#endif   
   if( ! oldObject || ! newObject)
   {
      NSLog( @"%s %@ %@ called with nil object", __PRETTY_FUNCTION__, oldObject, newObject);
      return;
   }
   
   if( oldObject == [association object])
   {
      id  old;

      old = association;
      association = [[[old class] alloc] initWithObject:newObject];
      [association bindAspect:aspectName
                 displayGroup:[old displayGroupForAspect:aspectName]
                          key:[old displayGroupKeyForAspect:aspectName]];
      [old release];
      return;
   }
   
   if( oldObject == [association displayGroupForAspect:aspectName])
   {
      [association bindAspect:aspectName
                 displayGroup:newObject
                          key:[association displayGroupKeyForAspect:aspectName]];
      return;
   }
   
   NSLog( @"%s %@ %@ failed to find object to replace", __PRETTY_FUNCTION__, oldObject, newObject);
}


- (id) nibInstantiate
{
#if DEBUG_ASPECT_CONNECTOR
   NSLog( @"%s %@", __PRETTY_FUNCTION__, self);
#endif
/*   association = [association nibInstantiate];
*/
   return( self);  // hein ??
}


- (void) establishConnection
{
#if DEBUG_ASPECT_CONNECTOR
   NSLog( @"%s %@", __PRETTY_FUNCTION__, self);
#endif
   if( ! [association isConnected])
   {
      [association establishConnection];  // should retain itself now

      //
      // associations aren't top level objects currently
      // so they dangle around. After the first connection
      // they have retained themselves. What we do then is
      // autorelease it here once. People breaking connections
      // should be sure to have retained it
      //
      // doesn't work, for some reason ??
      // Doesn't work, because non-toplevel objects get a release
      // anyway. So it's not needed.
      //
      // [association autorelease];
   }
}

@end
