/*
 *  MulleEOInterface - Replacement EOInterface Framework for EOF 
 *  Copyright (C) 2006 Mulle kybernetiK. All rights reserved.
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; specifically
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  THIS IS EXPERIMENTAL SOFTWARE AND IT IS KNOWN TO HAVE BUGS, SOME OF
 *  WHICH MAY HAVE SERIOUS CONSEQUENCES. THE COPYRIGHT HOLDER ALLOWS FREE
 *  USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. THE COPYRIGHT HOLDER
 *  DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 *  RESULTING DIRECTLY OR INDIRECTLY FROM THE USE OF THIS SOFTWARE OR OF
 *  ANY DERIVATIVE WORK.
 *
 *  Coded by Nat!
 *
 * $Id$
 *
 * $Log$
 *
 */
#import "Defines.h"

#import "EOWindowStateReflector.h"

#import "MulleEOInterfaceParentIncludes.h"


@implementation EOWindowStateReflector

- (void) editingContextChanged:(NSNotification *) notification
{
   EOEditingContext   *context;
   
   context = [notification object];
   [window setDocumentEdited:[context hasChanges]];
}


- (void) windowWillClose:(NSNotification *) notification
{
   NSWindow   *aWindow;
   
   aWindow = [notification object];
   if( ! [aWindow isDocumentEdited])
      return;
   
   // ask user if to saveChanges
   if( askBeforeClosing_)
   {
      // open a sheet and do stuff...
      [editingContext saveChanges:self];
   }
}


- (void) setEditingContext:(EOEditingContext *) context
{
   NSNotificationCenter  *center;
   
   center = [NSNotificationCenter defaultCenter];
   [center removeObserver:self
                     name:EOObjectsChangedInEditingContextNotification
                     object:nil];
  
   editingContext = context;

   if( editingContext)
      [center addObserver:self
                 selector:@selector( editingContextChanged:)
                     name:EOObjectsChangedInEditingContextNotification
                   object:editingContext];
}


- (void) setWindow:(NSWindow *) aWindow
{
   NSNotificationCenter  *center;
   
   center = [NSNotificationCenter defaultCenter];
   [center removeObserver:self
                     name:NSWindowWillCloseNotification
                  object:nil];

   window = aWindow; // no actual use for it

   if( window)
      [center addObserver:self
                 selector:@selector( windowWillClose:)
                     name:NSWindowWillCloseNotification
                   object:window];
}


- (NSWindow *) window
{
   return( window);
}


- (EOEditingContext *) editingContext
{
   return( editingContext);
}


- (id) initWithCoder:(NSCoder *) decoder
{
   int   version;
   
   [self init];
   
   version = [decoder versionForClassName:@"EOWindowStateReflector"];
#if DEBUG_CODER   
   NSLog( @"<<>> %@ %p Version: %d", isa, self, version);
#endif   
   [self setEditingContext:[decoder decodeObject]];
   [self setWindow:[decoder decodeObject]];
#if DEBUG_CODER   
   NSLog( @"Done %@", self);
#endif   
   
   return( self);
}


- (void) encodeWithCoder:(NSCoder *) encoder
{
#if DEBUG_CODER
   NSLog( @"%@ %s Version=%d (%d)", self, __PRETTY_FUNCTION__, [EOWindowStateReflector version], [isa version]);
#endif   
   [encoder encodeObject:editingContext];  // encode "real" data source
   [encoder encodeObject:window];
#if DEBUG_CODER   
   NSLog( @"Done %@", self);
#endif   
}

@end
